#include "quickcam.h"
#include "hdcs.h"

#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)<(b)?(a):(b))                                                                   
/* start grabbing */
static int hdcs_start(struct usb_device * dev)
{
        return(usb_quickcam_i2c_out(dev, HDCS_CONTROL, 0x04));
}
/* stop grabbing */
static int hdcs_stop(struct usb_device * dev)
{
        if (usb_quickcam_i2c_out(dev, HDCS_CONTROL, 0x00)<0)
                return  (-2);
	return(0);
}

/* sensor window to read out */

static int hdcs_set_size(struct usb_device *dev, int mode)
{
    if(mode != 0 && mode != 1)
	goto error;

    // 1: Y-full, 2: y-half
    if (usb_quickcam_set1(dev, STV_Y_CTRL, (mode?2:1)) < 0)
	goto error;
    // 06/0a : Half/Full
    if (usb_quickcam_set1(dev, STV_X_CTRL, (mode?6:0x0a)) < 0)
	goto error;
    
    return(0);
error:
    return(-1);
}


/* 
 * initialise parameters - copy + paste from Georg Acher's user module
 * for hdcs sensor.
 */
static int hdcs_init(struct usb_device *dev, int mode,
int rgain, int bgain, int ggain)
{

        if (usb_quickcam_set1(dev, STV_REG23, 0) < 0)
		goto error;

	/*
	 * reset the Image Sensor. (keeping it to 1 is a problem).
	 */
        if (usb_quickcam_i2c_out(dev, HDCS_CONTROL, 1) < 0) {
		printk(KERN_ERR "usb_quickcam_i2c_out HDCS_CONTROL(0x01) failed\n");
		goto error;
	}
        if (usb_quickcam_i2c_out(dev, HDCS_CONTROL, 0) < 0) {
		printk(KERN_ERR "usb_quickcam_i2c_out HDCS_CONTROL(0x00) failed\n");
		goto error;
	}

/* 
        if (usb_quickcam_i2c_out(dev, HDCS_REG00, 0x00) < 0) {
		/0 I see no reason that the others failed if this one is OK 0/
		printk(KERN_ERR "usb_quickcam_i2c_out HDCS_REG00(0x00) failed\n");
		goto error;
	}
 */
	// clear status.
        if (usb_quickcam_i2c_out(dev, HDCS_STATUS, 0x00) < 0)
		goto error;
	// clear interrupt mask.
        if (usb_quickcam_i2c_out(dev, HDCS_IMASK, 0x00) < 0)
		goto error;

	/* was the place where Georg sets gains */

        if (usb_quickcam_set1(dev, STV_REG00, 0x1d) < 0)
		goto error;
        if (usb_quickcam_set1(dev, STV_REG04, 0x07) < 0)
		goto error;
        if (usb_quickcam_set1(dev, STV_REG03, 0x95) < 0)
		goto error;
 
        if (usb_quickcam_set1(dev, STV_REG23, 0) < 0) 
		goto error;
 
	// larger -> slower 
        if (usb_quickcam_set1(dev, STV_SCAN_RATE, (mode?0x10:0x20)) <0)
		goto error;
	// ISO-Size, 0x34f = 847 ..
        if (usb_quickcam_set2(dev, STV_ISO_SIZE, 0x34f) < 0) {
		printk(KERN_ERR "usb_quickcam_set2 STV_ISO_SIZE(847) failed\n");
		goto error;
	}

	// Set Size 
	if (hdcs_set_size(dev, mode) < 0)
		goto error;
 
        if (usb_quickcam_i2c_out(dev, HDCS_STARTY, 2) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_STARTX, 4) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_STOPY, 0x4e) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_STOPX, 0x5b) < 0)
		goto error;
 
	// 0x7-0x50
        if (usb_quickcam_i2c_out(dev, HDCS_INTEGRATE, 0x9) < 0)
		goto error;
 
        if (usb_quickcam_i2c_out(dev, HDCS_CONTROL, 0) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_REG2A, 0x0) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_REG2C, 0x0) < 0)
		goto error;
 
        if (usb_quickcam_set1(dev, STV_REG01, 0xb5) < 0)
		goto error;
        if (usb_quickcam_set1(dev, STV_REG02, 0xa8) < 0)
		goto error;
 
        if (usb_quickcam_i2c_out(dev, HDCS_REG06, 0x63) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_REG08, 0x00) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_REG0A, 0x20) < 0)
		goto error;
        if (usb_quickcam_i2c_out(dev, HDCS_REG0C, 0x12) < 0)
		goto error;
	/*
         * CONFIG: 0x08 Continous Frame Capture.
	 *         0x04 Stop when Frame Complete.
         */
        if (usb_quickcam_i2c_out(dev, HDCS_CONFIG, 0x08) < 0)
		goto error;
 
        if (usb_quickcam_i2c_out(dev, HDCS_ADC_BITS, 10) < 0)
		goto error;

	return(0);
error:
	return(-1);
}
/* set_shutter */

static int hdcs_set_shutter(struct usb_device *dev, int sval, int xval)
{
char buff[35]; /* why 35 = 23 hex? */

        //buff[0]=HDCS_CONTROL;
        //buff[0x10]=0;
        buff[0]=HDCS_SHUTTERL;
        buff[0x10]=sval&255;
        buff[1]=HDCS_SHUTTERH;
        buff[0x11]=sval>>8;
        buff[2]=HDCS_REG2A;
        buff[0x12]=xval&255;
        buff[3]=HDCS_REG2C;
        buff[0x13]=xval>>8;
        //buff[5]=HDCS_CONTROL;
        //buff[0x15]=4;
 
        buff[0x20]=HDCS_ADDR;
        //buff[0x21]=5; // 6 values
        buff[0x21]=3; // 4 values
        buff[0x22]=1; // Write cmd
 
	return usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buff, 0x23 , HZ);
}

/* set gains */
static int hdcs_set_gains(struct usb_device *dev, int gain,
int white_balance)
{
char buff[35]; /* why 35 = 23 hex? */
int gx = gain/2;

	buff[0]=HDCS_GAIN_GREEN1;
	buff[0x10]=128+max(0,min(127,gx));

	buff[1]=HDCS_GAIN_RED;
	buff[0x11]=128+max(0,min(127,gx+white_balance));

	buff[2]=HDCS_GAIN_BLUE;
	buff[0x12]=128+max(0,min(127,gx-white_balance));

	buff[3]=HDCS_GAIN_GREEN2;
	buff[0x13]=128+max(0,min(127,gx));

	buff[0x20]=HDCS_ADDR;
	buff[0x21]=3; // 4 values
	buff[0x22]=1; // Write cmd

	return usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buff, 0x23 , HZ);
}

/*
 *  Sets the size (scaling) of the capture window.
 */
static int hdcs_set_window(struct usb_device *dev, int x, int y, int width, int height)
{
char buff[35]; /* why 35 = 23 hex? */

        buff[0]=HDCS_STARTY;
        buff[0x10]=2+y;
        buff[1]=HDCS_STARTX;
        buff[0x11]=4+x;
        buff[2]=HDCS_STOPY;
        buff[0x12]=(height+4)/4; // 4
        buff[3]=HDCS_STOPX;
        buff[0x13]=(width+12)/4;  // 12
 
        buff[0x20]=HDCS_ADDR;
        buff[0x21]=3; // 4 values
        buff[0x22]=1; // Write cmd
 
	return usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buff, 0x23 , HZ);
}


/*
 * Fill the pointers so that the routines are called from quickcam.c
 */
void load_hdcs_mod(struct sensorctrl *sensor_ctrl)
{
	sensor_ctrl->init	= hdcs_init;
	sensor_ctrl->set_shutter = hdcs_set_shutter;
	sensor_ctrl->set_gains	= hdcs_set_gains;
	sensor_ctrl->set_window	= hdcs_set_window;
	sensor_ctrl->set_size	= hdcs_set_size;
	sensor_ctrl->start	= hdcs_start;
	sensor_ctrl->stop	= hdcs_stop;
}
