#include "quickcam.h"
#include "pb0100.h"

static int pb0100_set_size(struct usb_device * dev, int mode)
{
    if(mode != 0 && mode != 1)
	goto error;

    // Set screen output size
    if (usb_quickcam_set1(dev, STV_Y_CTRL, (mode?2:1))<0) // 1: Y-full, 2: y-half
	goto error;
    if (usb_quickcam_set1(dev, STV_X_CTRL, (mode?6:0x0a))<0) // 06/0a : Half/Full
	goto error;

	return(0);
error:
	return(-1);
}

/* 
 * initialise parameters of PB100 sensor.
 */
static int pb0100_init(struct usb_device *dev, int mode,
int rgain, int bgain, int ggain)
{

	if (rgain<=0 || rgain>255) rgain=0xc0;
	if (bgain<=0 || bgain>255) bgain=0xc0;
	if (ggain<=0 || ggain>255) ggain=0xb0;
	if (usb_quickcam_set1(dev, STV_REG00, 1)<0)
		goto error;
	if (usb_quickcam_set1(dev, STV_SCAN_RATE, 0)<0)
		goto error;

	// Reset sensor
        if (usb_quickcam_i2c_out2(dev, PB_RESET, 1)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_RESET, 0)<0)
		goto error;

	// Disable chip
	if(usb_quickcam_i2c_out2(dev,PB_CONTROL,0x28)<0)
		goto error;
	
        // Gain stuff...
        if (usb_quickcam_i2c_out2(dev, PB_PREADCTRL, 0x1440)<0)
		goto error;
	if (usb_quickcam_i2c_out2(dev,PB_ADCMAXGAIN, 0x7F)<0)	/* gain max */
		goto error;
	if (usb_quickcam_i2c_out2(dev, PB_ADCMINGAIN, 1)<0)	/* gain min */	
		goto error;

        // Enable auto-exposure...
        if (usb_quickcam_i2c_out2(dev, PB_EXPGAIN, 17)<0)
		goto error;
       
	if (usb_quickcam_i2c_out2(dev, PB_UPDATEINT, 1)<0) 
		goto error;

        if (usb_quickcam_i2c_out2(dev, PB_VOFFSET, 0)<0) /* 0x14 */
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_ADCGAINH, 0x38)<0) /* 0xd */
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_ADCGAINL, 0x0)<0)  /* 0x1 */
		goto error;
        
        // Set individual colour gains
        if (usb_quickcam_i2c_out2(dev, PB_RGAIN, rgain)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_G1GAIN, ggain)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_G2GAIN, ggain)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_BGAIN, bgain)<0)
		goto error;

	// ???
        if (usb_quickcam_set1(dev, STV_REG04, 0x07)<0)
		goto error;
        if (usb_quickcam_set1(dev, STV_REG03, 0x45)<0)
		goto error;
        if (usb_quickcam_set1(dev, STV_REG00, 0x11)<0)
		goto error;

        // Set screen output size
        if (pb0100_set_size(dev, mode) <0)
		goto error;

	// 0x27b: 635... why? - HDCS uses 847 -
        if (usb_quickcam_set2(dev, STV_ISO_SIZE, 847)<0) // ISO-Size 
		goto error;

        // Setup sensor window
        if (usb_quickcam_i2c_out2(dev, PB_RSTART, 0)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_CSTART, 0)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_RWSIZE, 0x11f)<0) // 0xf7: 240
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_CWSIZE, 0x15f)<0) // 0x13f: 320
		goto error;

        // Scan rate?
        if (usb_quickcam_set1(dev, STV_SCAN_RATE, (mode?0x10:0x20))<0) // larger -> slower
		goto error;
                
        // Scan/timing for the sensor
        if (usb_quickcam_i2c_out2(dev, PB_ROWSPEED, 0x1a)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_CFILLIN, 0x2f)<0)
		goto error;
	if (usb_quickcam_i2c_out2(dev, PB_VBL, 0)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_FINTTIME, 0)<0)
		goto error;
        if (usb_quickcam_i2c_out2(dev, PB_RINTTIME, 0x7b)<0)
		goto error;

        if (usb_quickcam_set1(dev, STV_REG01, 0xc2)<0)
		goto error;
        if (usb_quickcam_set1(dev, STV_REG02, 0xb0)<0)
		goto error;
	return(0);
error:
	return(-1);
}

static int pb0100_set_shutter(struct usb_device * dev, int sval, int xval)
{
	return(0);
}

static int pb0100_set_gains(struct usb_device * dev, int gain, int white)
{
	return(0);
}
  /* Window location and size are controlled by R1, R2, R3 and R4.
   * The default size is CIF (352x288) with to right at (4,12)
   * and bottom left at (355, 299)
   *
   * We try to ensure that the captured area is in the center of
   * the camera purely because that's nicer.  It would be better
   * if the PB0100 sensor supported capture scaling!
   *
   * We do it in on step otherwise size changemay take more
   * than one frame (like xawtv who tests 64x48 and uses 352x288)
   * 3072 = 64x48, 16896 = 352x48, 101376 = 352x288.
   */
static int pb0100_set_window(struct usb_device * dev, int x, int y, int w, int h)
{
char buff[35]; /* why 35 = 23 hex? */
int val;
char *ptr=&buff[0x10];

	/* PB_RSTART = 12 + y */
	buff[0]=PB_RSTART;
	val = 12 + y;
	*ptr=val&255;
	ptr++;
        *ptr=(val>>8)&255;
	ptr++;

	/* PB_CSTART = 4 + x */
	buff[1]= PB_CSTART;
	val = 4 + x;
	*ptr=val&255;
	ptr++;
        *ptr=(val>>8)&255;
	ptr++;

	/* PB_RWSIZE = h - 1 */
	buff[2] = PB_RWSIZE;
	val = h - 1;
	*ptr=val&255;
	ptr++;
        *ptr=(val>>8)&255;
	ptr++;

	/* PB_CWSIZE = w - 1 */
	buff[3] = PB_CWSIZE;
	val = w - 1;
	*ptr=val&255;
	ptr++;
        *ptr=(val>>8)&255;
	ptr++;

        buff[0x20]=PB_ADDR;
        buff[0x21]=3; // 4 values.
        buff[0x22]=1; // Write cmd, 03 would be read.

	if (quickcam_usb_control_msg(dev, buff)<0)
		return(-1);
	return(0);
}

/* start grabbing */

static int pb0100_start(struct usb_device * dev)
{
	return(usb_quickcam_i2c_out2(dev,PB_CONTROL, 0x2b));
}
/* stop grabbing */
static int pb0100_stop(struct usb_device * dev)
{
	if (usb_quickcam_i2c_out2(dev,PB_ABORTFRAME, 1)<0)
		return (-3);
	if (usb_quickcam_i2c_out2(dev,PB_CONTROL, 0x28)<0)
		return  (-2);
	return(0);
}

/*
 * Fill the pointers so that the routines are called from quickcam.c
 */
void load_pb0100_mod(struct sensorctrl *sensor_ctrl)
{
	sensor_ctrl->init	= pb0100_init;
	sensor_ctrl->set_shutter = pb0100_set_shutter;
	sensor_ctrl->set_gains	= pb0100_set_gains;
	sensor_ctrl->set_window	= pb0100_set_window;
	sensor_ctrl->set_size	= pb0100_set_size;
	sensor_ctrl->start	= pb0100_start;
	sensor_ctrl->stop	= pb0100_stop;
}
