#include "quickcam.h"
#include "vv6410.h"

#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)<(b)?(a):(b))                                                                   
/* start grabbing */
static int vv6410_start(struct usb_device * dev)
{
	int res;

	printk("vv6410_start\n");
	res = usb_quickcam_i2c_out(dev, VV6410_CONTROL, 0x04);
	if (res <0)
	{
		printk(KERN_ERR "vv6410_start failed\n");
	}
	return res;
}

/* stop grabbing */
static int vv6410_stop(struct usb_device * dev)
{
	printk("vv6410_stop\n");
	if (usb_quickcam_i2c_out(dev, VV6410_CONTROL, 0x00)<0)
	{
		printk(KERN_ERR "vv6410_stop failed\n");
		return  (-2);
	}
	return(0);
}

/* 
 * initialise parameters for vv6410 sensor.
 * Just try to send the same commands as Windows Quickcam soft.
 */
static int vv6410_init(struct usb_device *dev, int mode,
int *rgain, int *bgain, int *ggain)
{
	char buf[35];

	printk("vv6410_init\n");

        if (*rgain<=0 || *rgain>255) *rgain=0xc0;
        if (*bgain<=0 || *bgain>255) *bgain=0xc0;
        if (*ggain<=0 || *ggain>255) *ggain=0xb0;

        if (usb_quickcam_set1(dev, STV_REG23, 4) < 0)
		goto error;

        if (usb_quickcam_set1(dev, 0x1446, 0x00) < 0)
		goto error;

	buf[0] = VV6410_SETUP0;
	buf[1] = 0x8;
	
        buf[0x20]=VV6410_ADDR;
        buf[0x21]=0;
        buf[0x22]=1;
	
	if (usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buf, 0x23 , HZ) < 0)
	{
		printk(KERN_ERR "usb_quickcam_i2c_out VV6410_SETUP0(0x08) failed\n");
		goto error;
	}


	buf[0] = VV6410_GAIN;
	buf[1] = 0x24;
	
        buf[0x20]=VV6410_ADDR;
        buf[0x21]=0;
        buf[0x22]=1;

	if (usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buf, 0x23 , HZ) < 0)
	{
		printk(KERN_ERR "usb_quickcam_i2c_out VV6410_GAIN(0x24) failed\n");
		goto error;
	}


        if (usb_quickcam_set1(dev, 0x1504, 0x07) < 0)
		goto error;

        if (usb_quickcam_set1(dev, 0x1503, 0x45) < 0)
		goto error;



        if (usb_quickcam_set1(dev, STV_Y_CTRL , 0x02) < 0)
		goto error;
        if (usb_quickcam_set2(dev, STV_ISO_SIZE, 0x380) < 0)
		goto error;
        if (usb_quickcam_set1(dev, STV_X_CTRL , 0x00) < 0)
		goto error;
 
	buf[0x00] = 0x52;
	buf[0x10] = 0x01;
	buf[0x01] = 0x53;
	buf[0x11] = 0x9f;
	buf[0x02] = 0x61;
	buf[0x12] = 0x01;
	buf[0x03] = 0x62;
	buf[0x13] = 0x3f;
	buf[0x04] = 0x25;
	buf[0x14] = 0x01;

        buf[0x20]=VV6410_ADDR;
        buf[0x21]=4;
	buf[0x22]=1;

	if (usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buf, 0x23 , HZ) < 0)
	{
		printk(KERN_ERR "usb_control_msg size1 failed");
		goto error;
	}

	//gain
	buf[0x00] = 0x20;
	buf[0x10] = 0x00;
	buf[0x01] = 0x21;
	buf[0x11] = 0x6a;
	buf[0x02] = 0x22;
	buf[0x12] = 0x00;
	buf[0x03] = 0x23;
	buf[0x13] = 0xa0;

        buf[0x20]=VV6410_ADDR;
        buf[0x21]=3;
        buf[0x22]=1;

	if (usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buf, 0x23 , HZ) < 0)
	{
		printk(KERN_ERR "usb_control_msg gain1 failed");
		goto error;
	}

        if (usb_quickcam_set1(dev, 0x1501 , 0xb7) < 0)
		goto error;
        if (usb_quickcam_set1(dev, 0x1502 , 0xa7) < 0)
		goto error;


	// setup
	buf[0x00] = 0x11;
	buf[0x10] = 0x40;
	buf[0x01] = 0x14;
	buf[0x11] = 0x55;
	buf[0x02] = 0x15;
	buf[0x12] = 0x10;
	buf[0x03] = 0x16;
	buf[0x13] = 0x81;
	buf[0x04] = 0x17;
	buf[0x14] = 0x18;
	buf[0x05] = 0x18;
	buf[0x15] = 0x00;
	buf[0x06] = 0x77;
	buf[0x16] = 0x5e;
	buf[0x07] = 0x78;
	buf[0x17] = 0x04;
	
        buf[0x20]=VV6410_ADDR;
        buf[0x21]=7;
        buf[0x22]=1;

	if (usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buf, 0x23 , HZ) < 0)
	{
		printk(KERN_ERR "usb_control_msg setup failed");
		goto error;
	}

	// gain
	buf[0x00] = 0x20;
	buf[0x10] = 0x01;
	buf[0x01] = 0x21;
	buf[0x11] = 0xa0;
	buf[0x02] = 0x22;
	buf[0x12] = 0x00;
	buf[0x03] = 0x23;
	buf[0x13] = 0xa0;

        buf[0x20]=VV6410_ADDR;
        buf[0x21]=3;
        buf[0x22]=1; // Write cmd

	if (usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buf, 0x23 , HZ) < 0)
	{
		printk(KERN_ERR "usb_control_msg gain2 failed");
		goto error;
	}

        if (usb_quickcam_set1(dev, 0x1440 , 0x01) < 0)
		goto error;

	printk("vv6410_init ok\n");
	return 0;

error:
	
	printk(KERN_ERR "vv6410_init failed\n");
	return(-1);
}

static int vv6410_set_shutter(struct usb_device *dev, int sval, int xval)
{
	printk("vv6410_set_shutter\n");
	return 0;
}

static int vv6410_set_gains(struct usb_device *dev,
int rgain, int bgain, int ggain)
{
	printk("vv6410_set_gains\n");
	return 0;
}

static int vv6410_set_window(struct usb_device *dev, int x, int y, int width, int height)
{
	char buf[35];

	printk("vv6410_set_window %d %d %d %d\n",x,y,width,height);

	// x offset
	buf[0x00] = 0x57;
	buf[0x10] = x >> 8;
	buf[0x01] = 0x58;
	buf[0x11] = x & 0xff;

	// y offset
	buf[0x02] = 0x59;
	buf[0x12] = y >> 8;
	buf[0x03] = 0x5a;
	buf[0x13] = y & 0xff;

	// line length
	buf[0x04] = 0x52;
	buf[0x14] = width >> 8;
	buf[0x05] = 0x53;
	buf[0x15] = width & 0xff;

	// field length (num lines)
	buf[0x06] = 0x61;
	buf[0x16] = height >> 8;
	buf[0x07] = 0x62;
	buf[0x17] = height & 0xff;

	buf[0x08] = 0x25;
	buf[0x18] = 0x01;


        buf[0x20]=VV6410_ADDR;
        buf[0x21]=8; 
        buf[0x22]=1; // Write cmd

	if (usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
		0x04,
		0x40,
		0x400, 0,
		buf, 0x23 , HZ) < 0)
	{
		printk(KERN_ERR "set_window failed");
		return -1;
	}
	return 0;
}



static int vv6410_set_size(struct usb_device *dev, int mode)
{
	printk("vv6410_set_size\n");
	return 0;
}

void load_vv6410_mod(struct sensorctrl *sensor_ctrl)
{
	printk("load_vv6410_mod\n");
	sensor_ctrl->init	= vv6410_init;
	sensor_ctrl->set_shutter = vv6410_set_shutter;
	sensor_ctrl->set_gains	= vv6410_set_gains;
	sensor_ctrl->set_window	= vv6410_set_window;
	sensor_ctrl->set_size	= vv6410_set_size;
	sensor_ctrl->start	= vv6410_start;
	sensor_ctrl->stop	= vv6410_stop;
}
