/*
 *  Matt Wu <Matt_Wu@acersoftech.com.cn>
 *  Feb 06, 2001
 *  Routines for control of ALi PCI AUDIO M5451
 *
 *  BUGS:
 *    --
 *
 *  TODO:
 *    --
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public Lcodecnse as published by
 *   the Free Software Foundation; either version 2 of the Lcodecnse, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public Lcodecnse for more details.
 *
 *   You should have received a copy of the GNU General Public Lcodecnse
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define __SND_OSS_COMPAT__
#define SND_MAIN_OBJECT_FILE

#include "../include/driver.h"
#include "../include/pcm.h"
#include "../include/mixer.h"
#include "../include/info.h"
#include "../include/ac97_codec.h"
#include "../include/mpu401.h"
#include "../include/initval.h"
#include <linux/timer.h>
#include <linux/tqueue.h>

EXPORT_NO_SYMBOLS;
MODULE_DESCRIPTION("\
Driver: ALI M5451\n\
PCI: 0x10b9=0x5451\n\
");
MODULE_LICENSE("GPL");

static int snd_index[SND_CARDS] = SND_DEFAULT_IDX;	/* Index 0-MAX */
static char *snd_id[SND_CARDS] = SND_DEFAULT_STR;	/* ID for this card */
static int snd_dac_frame_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 128};
static int snd_adc_frame_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 64};
// static int snd_foldback_frame_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 64};
// static int snd_spdif_frame_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 128};
static int snd_pcm_channels[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 32};
// int snd_wavetable_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 2048};
MODULE_PARM(snd_index, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_index, "Index value for ALI 5451 PCI Audio.");
MODULE_PARM(snd_id, "1-" __MODULE_STRING(SND_CARDS) "s");
MODULE_PARM_DESC(snd_id, "ID string for ALI 5451 PCI Audio.");
MODULE_PARM(snd_dac_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_dac_frame_size, "DAC frame size in kB for ALI 5451 PCI Audio.");
MODULE_PARM(snd_adc_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_adc_frame_size, "ADC frame size in kB for ALI 5451 PCI Audio.");
// MODULE_PARM(snd_spdif_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i");
// MODULE_PARM_DESC(snd_spdif_frame_size, "S/PDIF frame size in kB for ALI 5451 PCI Audio.");
MODULE_PARM(snd_pcm_channels, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_pcm_channels, "Number of hardware channels assigned for PCM. [range=0-31]");

/*
 *  Debug part definations
 */

//#define ALI_DEBUG

#ifdef ALI_DEBUG
#define snd_ali_printk(format, args...) printk(format, ##args);
#else
#define snd_ali_printk(format, args...)
#endif

/*
 *  Constants defination
 */

#ifndef PCI_VENDOR_ID_ALI
#define PCI_VENDOR_ID_ALI	0x10b9
#endif

#ifndef PCI_DEVICE_ID_ALI_5451
#define PCI_DEVICE_ID_ALI_5451	0x5451
#endif

#define DEVICE_ID_ALI5451	((PCI_VENDOR_ID_ALI<<16)|PCI_DEVICE_ID_ALI_5451)


#define ALI_CHANNELS		32

#define ALI_PCM_IN_CHANNEL	31
#define ALI_SPDIF_IN_CHANNEL	19
#define ALI_SPDIF_OUT_CHANNEL	15
#define ALI_CENTER_CHANNEL	24
#define ALI_LEF_CHANNEL		23
#define ALI_SURR_LEFT_CHANNEL	26
#define ALI_SURR_RIGHT_CHANNEL	25

#define	SND_ALI_VOICE_TYPE_PCM	01
#define SND_ALI_VOICE_TYPE_OTH	02

#define	ALI_5451_V02		0x02

/*
 *  Direct Registers
 */

#define ALI_LEGACY_DMAR0        0x00  // ADR0
#define ALI_LEGACY_DMAR4        0x04  // CNT0
#define ALI_LEGACY_DMAR11       0x0b  // MOD 
#define ALI_LEGACY_DMAR15       0x0f  // MMR 
#define ALI_MPUR0		0x20
#define ALI_MPUR1		0x21
#define ALI_MPUR2		0x22
#define ALI_MPUR3		0x23

#define	ALI_AC97_WRITE		0x40
#define ALI_AC97_READ		0x44

#define ALI_SCTRL		0x48
#define ALI_AC97_GPIO		0x4c
#define ALI_SPDIF_CS		0x70
#define ALI_SPDIF_CTRL		0x74
#define ALI_START		0x80
#define ALI_STOP		0x84
#define ALI_CSPF		0x90
#define ALI_AINT		0x98
#define ALI_GC_CIR		0xa0
	#define ENDLP_IE		0x00001000
	#define MIDLP_IE		0x00002000
#define ALI_AINTEN		0xa4
#define ALI_VOLUME		0xa8
#define ALI_SBDELTA_DELTA_R     0xac
#define ALI_MISCINT		0xb0
	#define ADDRESS_IRQ		0x00000020
	#define TARGET_REACHED		0x00008000
	#define MIXER_OVERFLOW		0x00000800
	#define MIXER_UNDERFLOW		0x00000400
#define ALI_SBBL_SBCL           0xc0
#define ALI_SBCTRL_SBE2R_SBDD   0xc4
#define ALI_STIMER		0xc8
#define ALI_GLOBAL_CONTROL	0xd4

#define ALI_CSO_ALPHA_FMS	0xe0
#define ALI_LBA			0xe4
#define ALI_ESO_DELTA		0xe8
#define ALI_GVSEL_PAN_VOC_CTRL_EC	0xf0
#define ALI_EBUF1		0xf4
#define ALI_EBUF2		0xf8

#define ALI_REG(codec, x) ((codec)->port + x)

typedef struct snd_stru_ali ali_t;

typedef struct snd_ali_channel_control {
	// register data
	struct REGDATA {
		unsigned int start;
		unsigned int stop;
		unsigned int aint;
		unsigned int ainten;
	} data;
		
	// register addresses
	struct REGS {
		unsigned int start;
		unsigned int stop;
		unsigned int aint;
		unsigned int ainten;
		unsigned int ac97read;
		unsigned int ac97write;
	} regs;

} snd_ali_channel_control_t;

typedef struct snd_ali_stru_voice {
	unsigned int number;
	int use: 1,
	    pcm: 1,
	    synth:1,
	    midi: 1,
	    mode: 1;

	/* PCM data */
	ali_t *codec;
	snd_pcm_subchn_t *subchn;
	
	int running: 1,
	    ignore_middle: 1;

	int eso;                /* final ESO value for channel */
	int count;              /* count between interrupts */
	int sample;

	/* --- */

	void *private_data;
	void (*private_free)(void *private_data);
} snd_ali_voice_t;


typedef struct snd_stru_alidev {

	snd_ali_voice_t voices[ALI_CHANNELS];	

	unsigned int	chcnt;			/* num of opened channels */
	unsigned int	chmap;			/* bitmap for opened channels */
	unsigned int synthcount;

} alidev_t;


#ifdef CONFIG_PM
#define ALI_GLOBAL_REGS		56
#define ALI_CHANNEL_REGS	8
typedef struct snd_ali_image {
	unsigned long regs[ALI_GLOBAL_REGS];
	unsigned long channel_regs[ALI_CHANNELS][ALI_CHANNEL_REGS];
} ali_image_t;
#endif


struct snd_stru_ali {
	snd_dma_t	*dma_pbk;	/* playback */
	snd_dma_t	*dma_cap;	/* capture */
	snd_irq_t	*irqptr;

	unsigned long	port;
	unsigned char	revision;

	struct pci_dev	*pci;
	struct pci_dev	*pci_m1533;
	struct pci_dev	*pci_m7101;

	snd_card_t	*card;
	snd_pcm_t	*pcm;
	alidev_t	synth;
	snd_ali_channel_control_t chregs;

	/* S/PDIF Mask */
	unsigned int	spdif_mask;

	ac97_t *ac97;
	unsigned short	ac97_ext_id;
	unsigned short	ac97_ext_status;
	snd_kmixer_t	*mixer;
	
	spinlock_t	reg_lock;
	spinlock_t	voice_alloc;
	spinlock_t	multi_synth;
	snd_info_entry_t *proc_entry;

#ifdef CONFIG_PM
	struct pm_dev *pm_dev;
	ali_image_t *image;
#endif
};

typedef struct snd_ali_card {
	struct pci_dev *pci;
	snd_irq_t *irqptr;
	snd_dma_t *dma_pbk;	/* playback */
	snd_dma_t *dma_cap;	/* capture */
	snd_card_t *card;
	ali_t *codec;
	snd_pcm_t *pcm;
	snd_kmixer_t *mixer;
	snd_rawmidi_t *midi_uart;
} ali_card_t;

static ali_card_t *snd_ali_cards[SND_CARDS] = SND_DEFAULT_PTR;

void snd_ali_clear_voices(ali_t *, unsigned int, unsigned int);
static unsigned short snd_ali_codec_peek(ali_t *, int, unsigned short);
static void snd_ali_codec_poke(ali_t *, int, unsigned short, unsigned short);

/*
 *  Debug Part
 */

#ifdef ALI_DEBUG

static void ali_read_regs(ali_t *codec, int channel)
{
	int i,j;
	unsigned int dwVal;

	printk("channel %d registers map:\n", channel);
	outb((unsigned char)(channel & 0x001f), ALI_REG(codec,ALI_GC_CIR));

	printk("    ");
	for(j=0;j<8;j++)
		printk("%2.2x       ", j*4);
	printk("\n");

	for (i=0; i<=0xf8/4;i++) {
		if(i%8 == 0)
			printk("%2.2x  ", (i*4/0x10)*0x10);
		dwVal = inl(ALI_REG(codec,i*4));
		printk("%8.8x ", dwVal);
		if ((i+1)%8 == 0)
			printk("\n");
	}
	printk("\n");
}
static void ali_read_cfg(unsigned int vendor, unsigned deviceid)
{
	unsigned int dwVal;
	struct pci_dev *pci_dev = NULL;
	int i,j;


        pci_dev = pci_find_device(vendor, deviceid, pci_dev);
        if (pci_dev == NULL)
                return ;

	printk("\nM%x PCI CFG\n", deviceid);
	printk("    ");
	for(j=0;j<8;j++)
		printk("%d        ",j);
	printk("\n");

	for(i=0;i<8;i++) {
		printk("%d   ",i);
		for(j=0;j<8;j++)
		{
			pci_read_config_dword(pci_dev, i*0x20+j*4, &dwVal);
			printk("%8.8x ", dwVal);
		}
		printk("\n");
	}
 }
static void ali_read_ac97regs(ali_t *codec, int secondary)
{
	unsigned short i,j;
	unsigned short wVal;

	printk("\ncodec %d registers map:\n", secondary);

	printk("    ");
	for(j=0;j<8;j++)
		printk("%2.2x   ",j*2);
	printk("\n");

	for (i=0; i<64;i++) {
		if(i%8 == 0)
			printk("%2.2x  ", (i/8)*0x10);
		wVal = snd_ali_codec_peek(codec, secondary, i*2);
		printk("%4.4x ", wVal);
		if ((i+1)%8 == 0)
			printk("\n");
	}
	printk("\n");
}

#endif

/*
 *  AC97 ACCESS
 */

static inline unsigned int snd_ali_5451_peek(ali_t *codec,
						unsigned int port )
{
	return (unsigned int)inl(ALI_REG(codec, port)); 
}

static inline void snd_ali_5451_poke(	ali_t *codec,
					unsigned int port,
					unsigned int val )
{
	outl((unsigned int)val, ALI_REG(codec, port));
}

static int snd_ali_codec_ready(	ali_t *codec,
				unsigned int port,
				int sched )
{
	signed long end_time;
	
	end_time = jiffies + 10 * (HZ >> 2);
	do {
		if (!(snd_ali_5451_peek(codec,port) & 0x8000))
			return 0;
		if (sched) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(1);
		}
	} while (end_time - (signed long)jiffies >= 0);
	snd_printk("ali_codec_ready: codec is not ready.\n ");
	return -EIO;
}

static int snd_ali_stimer_ready(ali_t *codec, int sched)
{
	signed long end_time;
	unsigned long dwChk1,dwChk2;
	
	dwChk1 = snd_ali_5451_peek(codec, ALI_STIMER);
	dwChk2 = snd_ali_5451_peek(codec, ALI_STIMER);

	end_time = jiffies + 10 * (HZ >> 2);
	do {
		dwChk2 = snd_ali_5451_peek(codec, ALI_STIMER);
		if (dwChk2 != dwChk1)
			return 0;
		if (sched) {
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(1);
		}
	} while (end_time - (signed long)jiffies >= 0);
	snd_printk("ali_stimer_read: stimer is not ready.\n");
	return -EIO;
}

static void snd_ali_codec_poke(ali_t *codec,int secondary,
				     unsigned short reg,
				     unsigned short val)
{
	unsigned int dwVal = 0;
	unsigned int port = 0;

	if (reg >= 0x80) {
		snd_printk("ali_codec_poke: reg(%xh) invalid.\n", reg);
		return;
	}

	port = codec->chregs.regs.ac97write;

	if (snd_ali_codec_ready(codec, port, 0) < 0)
		return;
	if (snd_ali_stimer_ready(codec, 0) < 0)
		return;

	dwVal  = (unsigned int) (reg & 0xff);
	dwVal |= 0x8000 | (val << 16);
	if (secondary) dwVal |= 0x0080;
	if (codec->revision == ALI_5451_V02) dwVal |= 0x0100;

	snd_ali_5451_poke(codec,port,dwVal);

	return ;
}

static unsigned short snd_ali_codec_peek( ali_t *codec,
					  int secondary,
					  unsigned short reg)
{
	unsigned int dwVal = 0;
	unsigned int port = 0;

	if (reg >= 0x80) {
		snd_printk("ali_codec_peek: reg(%xh) invalid.\n", reg);
		return ~0;
	}

	port = codec->chregs.regs.ac97read;

	if (snd_ali_codec_ready(codec, port, 0) < 0)
		return ~0;
	if (snd_ali_stimer_ready(codec, 0) < 0)
		return ~0;

	dwVal  = (unsigned int) (reg & 0xff);
	dwVal |= 0x8000;				/* bit 15*/
	if (secondary) dwVal |= 0x0080;

	snd_ali_5451_poke(codec, port, dwVal);

	if (snd_ali_stimer_ready(codec, 0) < 0)
		return ~0;
	if (snd_ali_codec_ready(codec, port, 0) < 0)
		return ~0;
	
	return (snd_ali_5451_peek(codec, port) & 0xffff0000)>>16;
}

static void snd_ali_codec_write(void *private_data,
				unsigned short reg,
				unsigned short val )
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, );

	snd_ali_printk("codec_write: reg=%xh data=%xh.\n", reg, val);
	snd_ali_codec_poke(codec, 0, reg, val);
	return ;
}


static unsigned short snd_ali_codec_read(void *private_data, unsigned short reg)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, 0);

	snd_ali_printk("codec_read reg=%xh.\n", reg);
	return (snd_ali_codec_peek(codec, 0, reg));
}

/*
 *	AC97 Reset
 */

static int snd_ali_reset_5451(ali_t *codec)
{
	struct pci_dev *pci_dev = NULL;
	unsigned short wCount, wReg;
	unsigned int   dwVal;
	
	if ((pci_dev = codec->pci_m1533) != NULL) {
		pci_read_config_dword(pci_dev, 0x7c, &dwVal);
		pci_write_config_dword(pci_dev, 0x7c, dwVal | 0x08000000);
		udelay(5000);
		pci_read_config_dword(pci_dev, 0x7c, &dwVal);
		pci_write_config_dword(pci_dev, 0x7c, dwVal & 0xf7ffffff);
		udelay(5000);
	}
	
	pci_dev = codec->pci;
	pci_read_config_dword(pci_dev, 0x44, &dwVal);
	pci_write_config_dword(pci_dev, 0x44, dwVal | 0x000c0000);
	udelay(500);
	pci_read_config_dword(pci_dev, 0x44, &dwVal);
	pci_write_config_dword(pci_dev, 0x44, dwVal & 0xfffbffff);
	udelay(5000);
	
	wCount = 200;
	while(wCount--) {
		wReg = snd_ali_codec_peek(codec, 0, AC97_POWERDOWN);
		if((wReg & 0x000f) == 0x000f)
			return 0;
		udelay(5000);
	}
	return -1;
}

#ifdef CODEC_RESET

static int snd_ali_reset_codec(ali_t *codec)
{
	struct pci_dev *pci_dev = NULL;
	unsigned char bVal = 0;
	unsigned int   dwVal;
	unsigned short wCount, wReg;

	pci_dev = codec->pci_m1533;
	
	pci_read_config_dword(pci_dev, 0x7c, &dwVal);
	pci_write_config_dword(pci_dev, 0x7c, dwVal | 0x08000000);
	udelay(5000);
	pci_read_config_dword(pci_dev, 0x7c, &dwVal);
	pci_write_config_dword(pci_dev, 0x7c, dwVal & 0xf7ffffff);
	udelay(5000);

	bVal = inb(ALI_REG(codec,ALI_SCTRL));
	bVal |= 0x02;
	outb(ALI_REG(codec,ALI_SCTRL),bVal);
	udelay(5000);
	bVal = inb(ALI_REG(codec,ALI_SCTRL));
	bVal &= 0xfd;
	outb(ALI_REG(codec,ALI_SCTRL),bVal);
	udelay(15000);

	wCount = 200;
	while(wCount--) {
		wReg = snd_ali_codec_read(codec, AC97_POWERDOWN);
		if((wReg & 0x000f) == 0x000f)
			return 0;
		udelay(5000);
	}
	return -1;
}

#endif

/*
 *  ALI 5451 Controller
 */

static void snd_ali_enable_special_channel(ali_t *codec, unsigned int channel)
{
	unsigned long dwVal = 0;

	dwVal  = inl(ALI_REG(codec,ALI_GLOBAL_CONTROL));
	dwVal |= 1 << (channel & 0x0000001f);
	outl(dwVal, ALI_REG(codec,ALI_GLOBAL_CONTROL));
}

static void snd_ali_disable_special_channel(ali_t *codec, unsigned int channel)
{
	unsigned long dwVal = 0;

	dwVal  = inl(ALI_REG(codec,ALI_GLOBAL_CONTROL));
	dwVal &= ~(1 << (channel & 0x0000001f));
	outl(dwVal, ALI_REG(codec,ALI_GLOBAL_CONTROL));
}

static void snd_ali_reset_ainten(ali_t * codec, int channel)
{
	unsigned int mask;
	snd_ali_channel_control_t *pchregs = &(codec->chregs);

	snd_ali_printk("reset_ainten: channel=%d\n", channel);

	pchregs->data.ainten =
		inl(ALI_REG(codec,pchregs->regs.ainten));
	mask = 1 << (channel & 0x1f);
	pchregs->data.ainten &= ~mask;

	outl(pchregs->data.ainten,
		ALI_REG(codec,pchregs->regs.ainten));
	// Ack the channel in case the interrupt was set before we disable it.
	outl(mask, ALI_REG(codec, pchregs->regs.aint));
}

static void snd_ali_enable_address_interrupt(ali_t * codec)
{
	unsigned int gc;

	gc  = inl(ALI_REG(codec, ALI_GC_CIR));
	gc |= ENDLP_IE;
	gc |= MIDLP_IE;
	outl( gc, ALI_REG(codec, ALI_GC_CIR));
}

static void snd_ali_disable_address_interrupt(ali_t * codec)
{
	unsigned int gc;

	gc  = inl(ALI_REG(codec, ALI_GC_CIR));
	gc &= ~ENDLP_IE;
	gc &= ~MIDLP_IE;
	outl(gc, ALI_REG(codec, ALI_GC_CIR));
}

void snd_ali_enable_voice_irq(ali_t *codec, unsigned int channel)
{
	unsigned int mask;
	snd_ali_channel_control_t *pchregs = &(codec->chregs);

	snd_ali_printk("enable_voice_irq channel=%d\n",channel);
	
	mask = 1 << (channel & 0x1f);
	pchregs->data.ainten  = inl(ALI_REG(codec,pchregs->regs.ainten));
	pchregs->data.ainten |= mask;
	outl(pchregs->data.ainten,ALI_REG(codec,pchregs->regs.ainten));
}

void snd_ali_disable_voice_irq(ali_t *codec, unsigned int channel)
{
	unsigned int mask;
	snd_ali_channel_control_t *pchregs = &(codec->chregs);

	snd_ali_printk("disable_voice_irq channel=%d\n",channel);

	mask = 1 << (channel & 0x1f);
	pchregs->data.ainten  = inl(ALI_REG(codec,pchregs->regs.ainten));
	pchregs->data.ainten &= ~mask;
	outl(pchregs->data.ainten,ALI_REG(codec,pchregs->regs.ainten));
}

int snd_ali_alloc_pcm_channel(ali_t *codec, int channel)
{
	unsigned int idx =  channel & 0x1f;

	if (codec->synth.chcnt >= ALI_CHANNELS){
		snd_printk("ali_alloc_pcm_channel: no free channels.\n");
		return -1;
	}

	if (!(codec->synth.chmap & (1 << idx))) {
		codec->synth.chmap |= 1 << idx;
		codec->synth.chcnt++;
		snd_ali_printk("alloc_pcm_channel no. %d.\n",idx);
		return idx;
	}
	return -1;
}

static int snd_ali_find_free_channel(ali_t * codec, int rec)
{
	int idx;
	int result = -1;

	snd_ali_printk("find_free_channel: for %s\n",rec ? "rec" : "pcm");

	// recording
	if (rec) {
		if (inl(ALI_REG(codec, ALI_GLOBAL_CONTROL)) & (1<<11) &&
			( codec->revision == ALI_5451_V02 ))
			idx = ALI_SPDIF_IN_CHANNEL;
		else
			idx = ALI_PCM_IN_CHANNEL;

		if ((result = snd_ali_alloc_pcm_channel(codec,idx)) >= 0) {
			return result;
		} else {
			snd_printk("ali_find_free_channel: record channel is busy now.\n");
			return -1;
		}
	}

	//playback...
	if (inl(ALI_REG(codec, ALI_GLOBAL_CONTROL)) & (1<<15)) {
		idx = ALI_SPDIF_OUT_CHANNEL;
		if ((result = snd_ali_alloc_pcm_channel(codec,idx)) >= 0) {
			return result;
		} else {
			snd_printk("ali_find_free_channel: S/PDIF out channel is in busy now.\n");
		}
	}

	for (idx = 0; idx < ALI_CHANNELS; idx++) {
		if ((result = snd_ali_alloc_pcm_channel(codec,idx)) >= 0)
			return result;
	}
	snd_printk("ali_find_free_channel: no free channels.\n");
	return -1;
}

static void snd_ali_free_channel_pcm(ali_t *codec, int channel)
{
	unsigned int idx = channel & 0x0000001f;

	snd_ali_printk("free_channel_pcm channel=%d\n",channel);

	if (channel < 0 || channel >= ALI_CHANNELS)
		return;

	if (!(codec->synth.chmap & (1 << idx))) {
		snd_printk("ali_free_channel_pcm: channel %d is not in use.\n",channel);
		return;
	} else {
		codec->synth.chmap &= ~(1 << idx);
		codec->synth.chcnt--;
	}
}

void snd_ali_start_voice(ali_t * codec, unsigned int channel)
{
	unsigned int mask = 1 << (channel & 0x1f);
	
	snd_ali_printk("start_voice: channel=%d\n",channel);
	outl(mask, ALI_REG(codec,codec->chregs.regs.start));
}

void snd_ali_stop_voice(ali_t * codec, unsigned int channel)
{
	unsigned int mask = 1 << (channel & 0x1f);

	snd_ali_printk("stop_voice: channel=%d\n",channel);

	outl(mask, ALI_REG(codec, codec->chregs.regs.stop));
}

/*
 *    S/PDIF Part
 */

void snd_ali_delay(ali_t *codec,int interval)
{
	unsigned long  begintimer,currenttimer;

	begintimer   = inl(ALI_REG(codec, ALI_STIMER));
	currenttimer = inl(ALI_REG(codec, ALI_STIMER));

	while (currenttimer < begintimer + interval) {
		if(snd_ali_stimer_ready(codec, 1) < 0)
			break;
		currenttimer = inl(ALI_REG(codec,  ALI_STIMER));
	}
}

void snd_ali_detect_spdif_rate(ali_t *codec)
{
	u16 wval  = 0;
	u16 count = 0;
	u8  bval = 0, R1 = 0, R2 = 0;

	bval  = inb(ALI_REG(codec,ALI_SPDIF_CTRL + 1));
	bval |= 0x1F;
	outb(bval,ALI_REG(codec,ALI_SPDIF_CTRL + 1));

	while (((R1 < 0x0B )||(R1 > 0x0E)) && (R1 != 0x12) && count <= 50000) {
		count ++;
		snd_ali_delay(codec, 6);
		bval = inb(ALI_REG(codec,ALI_SPDIF_CTRL + 1));
		R1 = bval & 0x1F;
	}

	if (count > 50000) {
		snd_printk("ali_detect_spdif_rate: timeout!\n");
		return;
	}

	count = 0;
	while (count++ <= 50000) {
		snd_ali_delay(codec, 6);
		bval = inb(ALI_REG(codec,ALI_SPDIF_CTRL + 1));
		R2 = bval & 0x1F;
		if (R2 != R1) R1 = R2; else break;
	}

	if (count > 50000) {
		snd_printk("ali_detect_spdif_rate: timeout!\n");
		return;
	}

	if (R2 >= 0x0b && R2 <= 0x0e) {
		wval  = inw(ALI_REG(codec,ALI_SPDIF_CTRL + 2));
		wval &= 0xE0F0;
		wval |= (u16)0x09 << 8 | (u16)0x05;
		outw(wval,ALI_REG(codec,ALI_SPDIF_CTRL + 2));

		bval  = inb(ALI_REG(codec,ALI_SPDIF_CS +3)) & 0xF0;
		outb(bval|0x02,ALI_REG(codec,ALI_SPDIF_CS + 3));
	} else if (R2 == 0x12) {
		wval  = inw(ALI_REG(codec,ALI_SPDIF_CTRL + 2));
		wval &= 0xE0F0;
		wval |= (u16)0x0E << 8 | (u16)0x08;
		outw(wval,ALI_REG(codec,ALI_SPDIF_CTRL + 2));

		bval  = inb(ALI_REG(codec,ALI_SPDIF_CS +3)) & 0xF0;
		outb(bval|0x03,ALI_REG(codec,ALI_SPDIF_CS + 3));
	}
}

static unsigned int snd_ali_get_spdif_in_rate(ali_t *codec)
{
	u32	dwRate = 0;
	u8	bval = 0;

	bval  = inb(ALI_REG(codec,ALI_SPDIF_CTRL));
	bval &= 0x7F;
	bval |= 0x40;
	outb(bval, ALI_REG(codec,ALI_SPDIF_CTRL));

	snd_ali_detect_spdif_rate(codec);

	bval  = inb(ALI_REG(codec,ALI_SPDIF_CS + 3));
	bval &= 0x0F;

	if (bval == 0) dwRate = 44100;
	if (bval == 1) dwRate = 48000;
	if (bval == 2) dwRate = 32000;

	return dwRate;
}

static void snd_ali_enable_spdif_in(ali_t *codec)
{	
	unsigned int dwVal;

	dwVal = inl(ALI_REG(codec, ALI_GLOBAL_CONTROL));
	dwVal |= 1<<11;
	outl(dwVal, ALI_REG(codec, ALI_GLOBAL_CONTROL));

	dwVal = inb(ALI_REG(codec, ALI_SPDIF_CTRL));
	dwVal |= 0x02;
	outb(dwVal, ALI_REG(codec, ALI_SPDIF_CTRL));

	snd_ali_enable_special_channel(codec, ALI_SPDIF_IN_CHANNEL);
}

static void snd_ali_disable_spdif_in(ali_t *codec)
{
	unsigned int dwVal;
	
	dwVal = inl(ALI_REG(codec, ALI_GLOBAL_CONTROL));
	dwVal &= ~(1<<11);
	outl(dwVal, ALI_REG(codec, ALI_GLOBAL_CONTROL));
	
	snd_ali_disable_special_channel(codec, ALI_SPDIF_IN_CHANNEL);	
}


static void snd_ali_set_spdif_out_rate(ali_t *codec, unsigned int rate)
{
	unsigned char  bVal;
	unsigned int  dwRate = 0;
	
	if (rate == 32000) dwRate = 0x300;
	if (rate == 44100) dwRate = 0;
	if (rate == 48000) dwRate = 0x200;
	
	bVal  = inb(ALI_REG(codec, ALI_SPDIF_CTRL));
	bVal &= (unsigned char)(~(1<<6));
	
	bVal |= 0x80;		//select right
	outb(bVal, ALI_REG(codec, ALI_SPDIF_CTRL));
	outb(dwRate | 0x20, ALI_REG(codec, ALI_SPDIF_CS + 2));
	
	bVal &= (~0x80);	//select left
	outb(bVal, ALI_REG(codec, ALI_SPDIF_CTRL));
	outw(rate | 0x10, ALI_REG(codec, ALI_SPDIF_CS + 2));
}

static void snd_ali_enable_spdif_out(ali_t *codec)
{
	unsigned short wVal;
	unsigned char bVal;

        struct pci_dev *pci_dev = NULL;

        pci_dev = codec->pci_m1533;
        if (pci_dev == NULL)
                return;
        pci_read_config_byte(pci_dev, 0x61, &bVal);
        bVal |= 0x40;
        pci_write_config_byte(pci_dev, 0x61, bVal);
        pci_read_config_byte(pci_dev, 0x7d, &bVal);
        bVal |= 0x01;
        pci_write_config_byte(pci_dev, 0x7d, bVal);

        pci_read_config_byte(pci_dev, 0x7e, &bVal);
        bVal &= (~0x20);
        bVal |= 0x10;
        pci_write_config_byte(pci_dev, 0x7e, bVal);

	bVal = inb(ALI_REG(codec, ALI_SCTRL));
	outb(bVal | 0x20, ALI_REG(codec, ALI_SCTRL));

	bVal = inb(ALI_REG(codec, ALI_SPDIF_CTRL));
	outb(bVal & ~(1<<6), ALI_REG(codec, ALI_SPDIF_CTRL));
   
	{
   		wVal = inw(ALI_REG(codec, ALI_GLOBAL_CONTROL));
   		wVal |= (1<<10);
   		outw(wVal, ALI_REG(codec, ALI_GLOBAL_CONTROL));
		snd_ali_disable_special_channel(codec,ALI_SPDIF_OUT_CHANNEL);
   	}
}

static void snd_ali_enable_spdif_chnout(ali_t *codec)
{
	unsigned short wVal = 0;

	wVal  = inw(ALI_REG(codec, ALI_GLOBAL_CONTROL));
   	wVal &= ~(1<<10);
   	outw(wVal, ALI_REG(codec, ALI_GLOBAL_CONTROL));
/*
	wVal = inw(ALI_REG(codec, ALI_SPDIF_CS));
	if (flag & ALI_SPDIF_OUT_NON_PCM)
   		wVal |= 0x0002;
	else	
		wVal &= (~0x0002);
   	outw(wVal, ALI_REG(codec, ALI_SPDIF_CS));
*/
	snd_ali_enable_special_channel(codec,ALI_SPDIF_OUT_CHANNEL);
}

static void snd_ali_disable_spdif_chnout(ali_t *codec)
{
	unsigned short wVal = 0;
  	wVal  = inw(ALI_REG(codec, ALI_GLOBAL_CONTROL));
   	wVal |= (1<<10);
   	outw(wVal, ALI_REG(codec, ALI_GLOBAL_CONTROL));

	snd_ali_enable_special_channel(codec, ALI_SPDIF_OUT_CHANNEL);
}

static void snd_ali_disable_spdif_out(ali_t *codec)
{
	unsigned char  bVal;

	bVal = inb(ALI_REG(codec, ALI_SCTRL));
	outb(bVal & (~0x20), ALI_REG(codec, ALI_SCTRL));

	snd_ali_disable_spdif_chnout(codec);
}

void snd_ali_update_ptr(ali_t *codec,int channel)
{
	snd_ali_voice_t *pvoice = NULL;
	snd_ali_channel_control_t *pchregs = NULL;
	unsigned int old, mask,temp;
	snd_pcm_runtime_t *runtime = NULL;

	pchregs = &(codec->chregs);

	// check if interrupt occured for channel
	old  = pchregs->data.aint;
	mask = ((unsigned int) 1L) << (channel & 0x1f);

	if (!(old & mask))
		return;
	spin_lock(&codec->reg_lock);
	if (codec->synth.voices[channel].pcm) {
		/* pcm interrupt */
		//ali_read_regs(codec,channel);

		pvoice = &codec->synth.voices[channel];
		runtime = codec->synth.voices[channel].subchn->runtime;

		if (pvoice->running) {
			outb((unsigned char) pvoice->number, ALI_REG(codec, ALI_GC_CIR));
			if (pvoice->ignore_middle) {
				if (!(inl(ALI_REG(codec, ALI_CSPF) & mask))) {
					spin_unlock(&codec->reg_lock);
					snd_pcm_transfer_done(pvoice->subchn);
					spin_lock(&codec->reg_lock);
				}
			} else {
				temp = inw(ALI_REG(codec, ALI_CSO_ALPHA_FMS + 2));
snd_ali_printk("update_ptr: pvoice members: eso=%xh,count=%xh\n",pvoice->eso,pvoice->count);
				spin_unlock(&codec->reg_lock);
				snd_pcm_transfer_done(pvoice->subchn);
				spin_lock(&codec->reg_lock);
			}
		} else {
			snd_ali_stop_voice(codec, channel);
			snd_ali_disable_voice_irq(codec, channel);
		}
	} else if (codec->synth.voices[channel].synth) {
		/* synth interrupt */
	} else if (codec->synth.voices[channel].midi) {
		/* midi interrupt */
	} else {
		/* unknown interrupt */
		snd_ali_stop_voice(codec, channel);
		snd_ali_disable_voice_irq(codec, channel);
	}
	spin_unlock(&codec->reg_lock);
	outl(mask,ALI_REG(codec,pchregs->regs.aint));
	pchregs->data.aint = old & (~mask);
}

void snd_ali_interrupt(ali_t * codec)
{
	int channel;
	unsigned int audio_int;
	snd_ali_channel_control_t *pchregs = NULL;
	pchregs = &(codec->chregs);

	audio_int = inl(ALI_REG(codec, ALI_MISCINT));
	if (audio_int & ADDRESS_IRQ) {
		// get interrupt status for all channels
		pchregs->data.aint = inl(ALI_REG(codec,pchregs->regs.aint));
		for (channel = 0; channel < ALI_CHANNELS; channel++) {
			snd_ali_update_ptr(codec, channel);
		}
	}
	outl((TARGET_REACHED | MIXER_OVERFLOW | MIXER_UNDERFLOW),
		ALI_REG(codec,ALI_MISCINT));
}


static void snd_ali_card_interrupt(int irq,
				   void *dev_id,
				   struct pt_regs *regs)
{
	ali_card_t *scard = (ali_card_t *) dev_id;
	ali_t 	*codec;

	if (scard == NULL || (codec = scard->codec) == NULL)
		return;
	snd_ali_interrupt(codec);
}


snd_ali_voice_t *snd_ali_alloc_voice(ali_t * codec, int type, int rec)
{
	snd_ali_voice_t *pvoice = NULL;
	unsigned long flags;
	int idx;

	snd_ali_printk("alloc_voice: type=%d rec=%d\n",type,rec);

	spin_lock_irqsave(&codec->voice_alloc, flags);
	if (type == SND_ALI_VOICE_TYPE_PCM) {
		idx = snd_ali_find_free_channel(codec,rec);
		if(idx < 0) {
			snd_printk("ali_alloc_voice: err.\n");
			spin_unlock_irqrestore(&codec->voice_alloc, flags);
			return NULL;
		}
		pvoice = &(codec->synth.voices[idx]);
		pvoice->use = 1;
		pvoice->pcm = 1;
		pvoice->mode = rec;
		spin_unlock_irqrestore(&codec->voice_alloc, flags);
		return pvoice;
	}
	spin_unlock_irqrestore(&codec->voice_alloc, flags);
	return NULL;
}


void snd_ali_free_voice(ali_t * codec, snd_ali_voice_t *pvoice)
{
	unsigned long flags;
	void (*private_free)(void *);
	void *private_data;

	snd_ali_printk("free_voice: channel=%d\n",pvoice->number);
	if (pvoice == NULL || !pvoice->use)
		return;
	snd_ali_clear_voices(codec, pvoice->number, pvoice->number);
	spin_lock_irqsave(&codec->voice_alloc, flags);
	private_free = pvoice->private_free;
	private_data = pvoice->private_data;
	pvoice->private_free = NULL;
	pvoice->private_data = NULL;
	if (pvoice->pcm) {
		snd_ali_free_channel_pcm(codec, pvoice->number);
	}
	pvoice->use = pvoice->pcm = pvoice->synth = 0;
	pvoice->subchn = NULL;
	spin_unlock_irqrestore(&codec->voice_alloc, flags);
	if (private_free)
		private_free(private_data);
}


void snd_ali_clear_voices(ali_t * codec,
			  unsigned int v_min,
			  unsigned int v_max)
{
	unsigned int i;

	for (i = v_min; i <= v_max; i++) {
		snd_ali_stop_voice(codec, i);
		snd_ali_disable_voice_irq(codec, i);
	}
}

void snd_ali_write_voice_regs(ali_t * codec,
			 unsigned int Channel,
			 unsigned int LBA,
			 unsigned int CSO,
			 unsigned int ESO,
			 unsigned int DELTA,
			 unsigned int ALPHA_FMS,
			 unsigned int GVSEL,
			 unsigned int PAN,
			 unsigned int VOL,
			 unsigned int CTRL,
			 unsigned int EC)
{
	unsigned int ctlcmds[4];
	
	outb((unsigned char)(Channel & 0x001f),ALI_REG(codec,ALI_GC_CIR));

	ctlcmds[0] =  (CSO << 16) | (ALPHA_FMS & 0x0000ffff);
	ctlcmds[1] =  LBA;
	ctlcmds[2] =  (ESO << 16) | (DELTA & 0x0ffff);
	ctlcmds[3] =  (GVSEL << 31) |
		      ((PAN & 0x0000007f) << 24) |
		      ((VOL & 0x000000ff) << 16) |
		      ((CTRL & 0x0000000f) << 12) |
		      (EC & 0x00000fff);

	outb(Channel, ALI_REG(codec, ALI_GC_CIR));

	outl(0x30000000, ALI_REG(codec, ALI_EBUF1));
	outl(0x30000000, ALI_REG(codec, ALI_EBUF2));

	outl(ctlcmds[0], ALI_REG(codec, ALI_CSO_ALPHA_FMS));
	outl(ctlcmds[1], ALI_REG(codec, ALI_LBA));
	outl(ctlcmds[2], ALI_REG(codec, ALI_ESO_DELTA));
	outl(ctlcmds[3], ALI_REG(codec, ALI_GVSEL_PAN_VOC_CTRL_EC));
}

unsigned int snd_ali_convert_rate(unsigned int rate, int rec)
{
	unsigned int delta;

	if (rate < 4000)  rate = 4000;
	if (rate > 48000) rate = 48000;

	if (rec) {
		if (rate == 44100)
			delta = 0x116a;
		else if (rate == 8000)
			delta = 0x6000;
		else if (rate == 48000)
			delta = 0x1000;
		else
			delta = ((48000 << 12) / rate) & 0x0000ffff;
	} else {
		if (rate == 44100)
			delta = 0xeb3;
		else if (rate == 8000)
			delta = 0x2ab;
		else if (rate == 48000)
			delta = 0x1000;
		else 
			delta = (((rate << 12) + rate) / 48000) & 0x0000ffff;
	}

	return delta;
}

unsigned int snd_ali_control_mode(snd_pcm_subchn_t *subchn)
{
	unsigned int CTRL;
	snd_pcm_runtime_t *runtime = subchn->runtime;

	/* set ctrl mode
	   CTRL default: 8-bit (unsigned) mono, loop mode enabled
	 */
	CTRL = 0x00000001;
	if (snd_pcm_format_width(runtime->format.format) == 16)
		CTRL |= 0x00000008;	// 16-bit data
	if (!snd_pcm_format_unsigned(runtime->format.format))
		CTRL |= 0x00000002;	// signed data
	if (runtime->format.voices > 1)
		CTRL |= 0x00000004;	// stereo data
	return CTRL;
}

/*
 *  PCM part
 */

static int snd_ali_set_buffer_size(ali_t *codec, snd_pcm_subchn_t * subchn)
{
	int shift = 0;
	unsigned int max_size = 0;
	snd_pcm_runtime_t *runtime = subchn->runtime;

	if (snd_pcm_format_width(runtime->format.format) == 16)
		shift++;
	if (runtime->format.voices > 1)
		shift++;
	max_size = 0x8000 << shift;

	if (runtime->buf.block.frag_size > max_size )
		return snd_pcm_lib_set_buffer_size(subchn, max_size);

	if (runtime->mode == SND_PCM_MODE_BLOCK) {
		if (runtime->frags > 2) {
			runtime->buf.block.frag_size *= runtime->frags/2;
			if (runtime->buf.block.frag_size > max_size)
				runtime->buf.block.frag_size = max_size;
			runtime->frags = 2;
		}
	}

	return 0;
}


static int snd_ali_playback_ioctl(void *private_data,
				      snd_pcm_subchn_t * subchn,
				      unsigned int cmd,
				      unsigned long *arg)
{
	int result;
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);
	if (result < 0)
		return result;

	if (cmd == SND_PCM_IOCTL1_PARAMS) {
		snd_ali_set_buffer_size(codec, subchn);
	}
	return 0;
}


static int snd_ali_capture_ioctl(void *private_data,
				      snd_pcm_subchn_t * subchn,
				      unsigned int cmd,
				      unsigned long *arg)
{
	int result;
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);
	if (result < 0)
		return result;
	if (cmd == SND_PCM_IOCTL1_PARAMS) {
		snd_ali_set_buffer_size(codec, subchn);
	}
	return 0;
}

static int snd_ali_sync_trigger(ali_t *codec, snd_pcm_subchn_t *subchn)
{
	unsigned long flags;
	int idx, capture_flag = 0, result = -EINVAL;
	unsigned int mask = 0;
	snd_ali_channel_control_t *pchregs = &(codec->chregs);

	snd_ali_printk("sync_trigger...\n");

	for (idx = 0; idx < ALI_CHANNELS; idx++) {
		snd_ali_voice_t *voice;
		snd_pcm_runtime_t *runtime;
		
		spin_lock_irqsave(&codec->voice_alloc, flags);
		voice = &codec->synth.voices[idx];
		if (voice->subchn == NULL) {
			spin_unlock_irqrestore(&codec->voice_alloc, flags);
			continue;
		}
		runtime = voice->subchn->runtime;
		if (memcmp(&subchn->runtime->sync_group, &runtime->sync_group, sizeof(runtime->sync_group))) {
			spin_unlock_irqrestore(&codec->voice_alloc, flags);
			continue;
		}
		spin_unlock_irqrestore(&codec->voice_alloc, flags);
		if (*runtime->status != SND_PCM_STATUS_PREPARED)
			goto __end;
		if (subchn->channel == SND_PCM_CHANNEL_PLAYBACK) {
			if (!snd_pcm_playback_data(subchn))
				goto __end;
		} else {
			if (!snd_pcm_capture_empty(subchn))
				goto __end;
		}
		if ((result = snd_pcm_channel_go_pre(subchn, SND_PCM_TRIGGER_GO))<0)
			goto __end;
		result = -EINVAL;
		if (subchn->channel == SND_PCM_CHANNEL_CAPTURE)
			capture_flag++;
		mask |= 1 << (idx & 0x1f);
	}
	spin_lock_irqsave(&codec->reg_lock, flags);

	pchregs->data.aint =
		inl(ALI_REG(codec,pchregs->regs.aint));
	pchregs->data.ainten =
		inl(ALI_REG(codec,pchregs->regs.ainten));

	for (idx = 0; idx < ALI_CHANNELS; idx++)
		if (mask & (1 << (idx & 0x1f)))
			codec->synth.voices[idx].running = 1;
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	result = 0;
      __end:
	for (idx = ALI_CHANNELS - 1; idx >= 0; idx--) {
		snd_ali_voice_t *voice;
		snd_pcm_runtime_t *runtime;
			
		spin_lock_irqsave(&codec->voice_alloc, flags);
		voice = &codec->synth.voices[idx];
		if (voice->subchn == NULL) {
			spin_unlock_irqrestore(&codec->voice_alloc, flags);
			continue;
		}
		runtime = voice->subchn->runtime;
		if (memcmp(&subchn->runtime->sync_group, &runtime->sync_group, sizeof(runtime->sync_group))) {
			spin_unlock_irqrestore(&codec->voice_alloc, flags);
			continue;
		}
		spin_unlock_irqrestore(&codec->voice_alloc, flags);
		*runtime->status = SND_PCM_STATUS_PREPARED;
		snd_pcm_channel_go_post(subchn, SND_PCM_TRIGGER_GO, result);
	}
	return result;
}


static int snd_ali_trigger(ali_t *codec,
			   snd_pcm_subchn_t * subchn,
			   snd_ali_voice_t *pvoice,
			   int cmd)
{
	unsigned long flags;

	if (cmd == SND_PCM_TRIGGER_SYNC_GO) {
		return snd_ali_sync_trigger(codec, subchn);
	}
	spin_lock_irqsave(&codec->reg_lock, flags);
	if (cmd == SND_PCM_TRIGGER_GO) {

#ifdef ALI_DEBUG
		ali_read_cfg(0x10b9, 0x1533);
		ali_read_cfg(0x10b9, 0x7101);
		ali_read_cfg(0x10b9, 0x5451);

		ali_read_ac97regs(codec, 0);
		//ali_read_ac97regs(codec, 1);
#endif
		snd_ali_enable_voice_irq(codec, pvoice->number);
		snd_ali_start_voice(codec, pvoice->number);
		pvoice->running = 1;

	} else if (cmd == SND_PCM_TRIGGER_STOP) {

		snd_ali_stop_voice(codec, pvoice->number);

		snd_ali_disable_voice_irq(codec, pvoice->number);
		snd_ali_reset_ainten(codec, pvoice->number);
		pvoice->running = 0;
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return 0;
}

static int snd_ali_playback_trigger(void *private_data,
					snd_pcm_subchn_t * subchn,
					int cmd)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;

	return snd_ali_trigger(codec, subchn, pvoice, cmd);
}


static int snd_ali_capture_trigger(void *private_data,
				       snd_pcm_subchn_t * subchn,
				       int cmd)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;

	return snd_ali_trigger(codec, subchn, pvoice, cmd);
}

static int snd_ali_playback_prepare(void *private_data,
				    snd_pcm_subchn_t * subchn)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;
	unsigned long flags;
	unsigned int LBA;
	unsigned int Delta;
	unsigned int ESO;
	unsigned int CTRL;
	unsigned int GVSEL;
	unsigned int PAN;
	unsigned int VOL;
	unsigned int EC;
	
	snd_ali_printk("playback_prepare ...\n");

	spin_lock_irqsave(&codec->reg_lock, flags);	
	
	/* set Delta (rate) value */
	Delta = snd_ali_convert_rate(runtime->format.rate, 0);

	if ((pvoice->number == ALI_SPDIF_IN_CHANNEL) || 
	    (pvoice->number == ALI_PCM_IN_CHANNEL))
		snd_ali_disable_special_channel(codec, pvoice->number);
	else if ((inl(ALI_REG(codec, ALI_GLOBAL_CONTROL)) & (1<<15)) 
		&& (pvoice->number == ALI_SPDIF_OUT_CHANNEL)) {
		if (codec->revision == ALI_5451_V02) {
			snd_ali_set_spdif_out_rate(codec, runtime->format.rate);
			Delta = 0x1000;
		}
	}
	
	/* set Loop Back Address */
	LBA = virt_to_bus(runtime->dma_area->buf);

	/* set interrupt count size */
	pvoice->count = snd_pcm_lib_transfer_fragment(subchn);

	/* set target ESO for channel */
	pvoice->eso = snd_pcm_lib_transfer_size(subchn); 

	pvoice->ignore_middle = pvoice->count == pvoice->eso;

	snd_ali_printk("playback_prepare: eso(%xh) == count(%xh)=%d\n",pvoice->eso,pvoice->count,pvoice->ignore_middle);

	/* convert to sample counters */
	if (snd_pcm_format_width(runtime->format.format) == 16) {
		pvoice->eso >>= 1;
		pvoice->count >>= 1;
	}
	if (runtime->format.voices > 1) {
		pvoice->eso >>= 1;
		pvoice->count >>= 1;
	}

	/* set ESO to capture first MIDLP interrupt */
	ESO = pvoice->ignore_middle ? pvoice->eso -1 : (pvoice->count << 1) - 1;

	/* set ctrl mode */
	CTRL = snd_ali_control_mode(subchn);

	GVSEL = 1;
	PAN = 0;
	VOL = 0;
	EC = 0;
	snd_ali_printk("playback_prepare:\n    ch=%d, Rate=%d Delta=%xh,GVSEL=%xh,PAN=%xh,CTRL=%xh\n",pvoice->number,runtime->format.rate,Delta,GVSEL,PAN,CTRL);
	snd_ali_printk("    pvoice members: ESO=%xh eso=%xh,count=%xh\n",ESO,pvoice->eso,pvoice->count);
	
	snd_ali_write_voice_regs(    codec,
				     pvoice->number,
				     LBA,
				     0,	/* cso */
				     ESO,
				     Delta,
				     0,	/* alpha */
				     GVSEL,
				     PAN,
				     VOL,
				     CTRL,
				     EC);
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	return 0;
}


static int snd_ali_capture_prepare(void *private_data,
				       snd_pcm_subchn_t * subchn)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;
	unsigned long flags;
	unsigned int LBA;
	unsigned int Delta;
	unsigned int ESO;
	unsigned int CTRL;
	unsigned int GVSEL;
	unsigned int PAN;
	unsigned int VOL;
	unsigned int EC;
	u8	 bValue;

	spin_lock_irqsave(&codec->reg_lock, flags);

	snd_ali_printk("capture_prepare...\n");

	snd_ali_enable_special_channel(codec,pvoice->number);

	Delta = snd_ali_convert_rate(runtime->format.rate, 1);

	// Prepare capture intr channel
	if (pvoice->number == ALI_SPDIF_IN_CHANNEL) {

		unsigned int rate;
		
		if (codec->revision != ALI_5451_V02)
			return -1;
		rate = snd_ali_get_spdif_in_rate(codec);
		if (rate == 0) {
			snd_printk("ali_capture_preapre: spdif rate detect err!\n");
			rate = 48000;
		}
		snd_ali_printk("capture_prepare: format.rate=%xh(%d),detect rate=%xh(%d)\n",runtime->format.rate,runtime->format.rate,rate,rate);
		bValue = inb(ALI_REG(codec,ALI_SPDIF_CTRL));
		if (bValue & 0x10) {
			outb(bValue,ALI_REG(codec,ALI_SPDIF_CTRL));
			snd_ali_printk("clear SPDIF parity error flag.\n");
		}

		if (rate != 48000)
			Delta = ((rate << 12)/runtime->format.rate)&0x00ffff;
	}
	{
		memset(runtime->dma_area->buf,0,runtime->dma_area->size);
	}
	// set target ESO for channel 
	pvoice->eso = snd_pcm_lib_transfer_size(subchn); 
	// set interrupt count size 
	pvoice->count = snd_pcm_lib_transfer_fragment(subchn);

	pvoice->ignore_middle = pvoice->count == pvoice->eso;

	snd_ali_printk("playback_prepare: eso(%xh) == count(%xh)=%d\n",pvoice->eso,pvoice->count,pvoice->ignore_middle);

	// convert to sample counters 
	if (snd_pcm_format_width(runtime->format.format) == 16) {
		pvoice->eso >>= 1;
		pvoice->count >>= 1;
	}
	if (runtime->format.voices > 1) {
		pvoice->eso >>= 1;
		pvoice->count >>= 1;
	}

	// set Loop Back Address 
	LBA = virt_to_bus(runtime->dma_area->buf);

	// set ESO to capture first MIDLP interrupt 
	ESO = pvoice->ignore_middle ? pvoice->eso -1 : (pvoice->count << 1) - 1;
	CTRL = snd_ali_control_mode(subchn);
	GVSEL = 0;
	PAN = 0x00;
	VOL = 0x00;
	EC = 0;

	snd_ali_printk("ali: --capture prepare --\n      ch=%d, Delta=%xh,GVSEL=%xh,PAN=%xh,CTRL=%xh\n",pvoice->number,Delta,GVSEL,PAN,CTRL);
	snd_ali_printk("pvoice members: eso=%xh,count=%xh\n",pvoice->eso,pvoice->count);

	snd_ali_write_voice_regs(    codec,
				     pvoice->number,
				     LBA,
				     0,	/* cso */
				     ESO,
				     Delta,
				     0,	/* alpha */
				     GVSEL,
				     PAN,
				     VOL,
				     CTRL,
				     EC);


	spin_unlock_irqrestore(&codec->reg_lock, flags);

	return 0;
}


static unsigned int snd_ali_playback_pointer(void *private_data,
						  snd_pcm_subchn_t * subchn)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;
	unsigned int cso;
	unsigned long flags;

	spin_lock_irqsave(&codec->reg_lock, flags);
	if (!pvoice->running) {
		spin_unlock_irqrestore(&codec->reg_lock, flags);
		return 0;
	}
	outb(pvoice->number, ALI_REG(codec, ALI_GC_CIR));
	cso = inw(ALI_REG(codec, ALI_CSO_ALPHA_FMS + 2));
	if (snd_pcm_format_width(runtime->format.format) == 16)
		cso <<= 1;
	if (runtime->format.voices > 1)
		cso <<= 1;
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	return cso;
}


static unsigned int snd_ali_capture_pointer(void *private_data,
						snd_pcm_subchn_t * subchn)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;
	unsigned int cso;
	unsigned long flags;

	spin_lock_irqsave(&codec->reg_lock, flags);
	if (!pvoice->running) {
		spin_unlock_irqrestore(&codec->reg_lock, flags);
		return 0;
	}
	outb(pvoice->number, ALI_REG(codec, ALI_GC_CIR));
	cso = inw(ALI_REG(codec, ALI_CSO_ALPHA_FMS + 2));
	if (snd_pcm_format_width(runtime->format.format) == 16)
		cso <<= 1;
	if (runtime->format.voices > 1)
		cso <<= 1;
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	return cso;
}

/*
 *  Playback support device description
 */

static snd_pcm_hardware_t snd_ali_playback =
{
	SND_PCM_CHNINFO_MMAP | SND_PCM_CHNINFO_STREAM |
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE |
	SND_PCM_CHNINFO_BLOCK_TRANSFER |
	SND_PCM_CHNINFO_MMAP_VALID,		/* flags */
	SND_PCM_FMT_U8 | SND_PCM_FMT_S16_LE |
	SND_PCM_FMT_S8 | SND_PCM_FMT_U16_LE,	/* formats */
	SND_PCM_RATE_CONTINUOUS | SND_PCM_RATE_8000_48000, /* supported RATES */
	4000,			/* min. rate */
	48000,			/* max. rate */
	1,			/* min. voices */
	2,			/* max. voices */
	64,			/* min. fragment size */
	(128*1024),		/* max. fragment size */
	3,			/* fragment align */
	0,			/* FIFO size (unknown) */
	32,			/* transfer block size */
	snd_ali_playback_ioctl,
	snd_ali_playback_prepare,
	snd_ali_playback_trigger,
	snd_ali_playback_pointer
};

/*
 *  Capture support device description
 */

static snd_pcm_hardware_t snd_ali_capture =
{
	SND_PCM_CHNINFO_MMAP | SND_PCM_CHNINFO_STREAM |
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE |
	SND_PCM_CHNINFO_BLOCK_TRANSFER |
	SND_PCM_CHNINFO_MMAP_VALID,		/* flags */
	SND_PCM_FMT_U8 | SND_PCM_FMT_S16_LE |
	SND_PCM_FMT_S8 | SND_PCM_FMT_U16_LE,	/* formats */
	SND_PCM_RATE_CONTINUOUS | SND_PCM_RATE_8000_48000, /* supported RATES */
	4000,			/* min. rate */
	48000,			/* max. rate */
	1,			/* min. voices */
	2,			/* max. voices */
	64,       		/* min. fragment size */
	(128*1024),		/* max. fragment size */
	3,			/* fragment align */
	0,			/* FIFO size (unknown) */
	16,			/* transfer block size */
	snd_ali_capture_ioctl,
	snd_ali_capture_prepare,
	snd_ali_capture_trigger,
	snd_ali_capture_pointer
};

static void snd_ali_pcm_free_subchn(void *private_data)
{
	unsigned long flags;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) private_data;
	ali_t *codec;

	if (pvoice) {
		codec = pvoice->codec;
		spin_lock_irqsave(&codec->reg_lock, flags);
		snd_ali_free_voice(pvoice->codec, pvoice);
		spin_unlock_irqrestore(&codec->reg_lock, flags);
	}
}


static int snd_ali_playback_open(void *private_data,
			             snd_pcm_subchn_t * subchn)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_hardware_t *hw;
	snd_ali_voice_t *pvoice;
	unsigned long flags = 0;
	int err;

	hw = (snd_pcm_hardware_t *)snd_kmalloc(sizeof(snd_pcm_hardware_t), GFP_KERNEL);
	if (hw == NULL)
		return -ENOMEM;

	if ((err = snd_pcm_dma_alloc(subchn, codec->dma_pbk,
				     "ALI 5451 - DAC")) < 0) {
		snd_kfree(hw);
		return err;
	}

	spin_lock_irqsave(&codec->reg_lock, flags);
	pvoice = snd_ali_alloc_voice(codec, SND_ALI_VOICE_TYPE_PCM, 0);
	if (pvoice == NULL) {
		spin_unlock_irqrestore(&codec->reg_lock, flags);
		snd_pcm_dma_free(subchn);
		return -EAGAIN;
	}
	pvoice->codec = codec;
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	
	pvoice->subchn = subchn;
	subchn->runtime->private_data = pvoice;
	subchn->runtime->private_free = snd_ali_pcm_free_subchn;

	memcpy(hw, &snd_ali_playback, sizeof(*hw));
	subchn->runtime->hw = hw;
	subchn->runtime->hw_free = (snd_kfree_type)_snd_kfree;
	snd_pcm_set_sync(subchn);
	snd_pcm_set_mixer(subchn, codec->mixer->device, codec->ac97->me_playback);
	return 0;
}


static int snd_ali_capture_open(void *private_data,
				    snd_pcm_subchn_t * subchn)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_hardware_t *hw;
	snd_ali_voice_t *pvoice;
	int err;
	unsigned long flags;

	hw = (snd_pcm_hardware_t *)snd_kmalloc(sizeof(snd_pcm_hardware_t), GFP_KERNEL);
	if (hw == NULL)
		return -ENOMEM;
	if ((err = snd_pcm_dma_alloc(subchn, codec->dma_cap,
				     "ALI 5451 - ADC")) < 0) {
		snd_kfree(hw);
		return err;
	}

	spin_lock_irqsave(&codec->reg_lock, flags);
	pvoice = snd_ali_alloc_voice(codec, SND_ALI_VOICE_TYPE_PCM, 1);
	if (pvoice == NULL) {
		spin_unlock_irqrestore(&codec->reg_lock, flags);
		snd_pcm_dma_free(subchn);
		return -EAGAIN;
	}
	pvoice->codec = codec;
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	pvoice->subchn = subchn;
	subchn->runtime->private_data = pvoice;
	subchn->runtime->private_free = snd_ali_pcm_free_subchn;
	
	memcpy(hw, &snd_ali_capture, sizeof(*hw));
	subchn->runtime->hw = hw;
	subchn->runtime->hw_free = (snd_kfree_type)_snd_kfree;

	snd_pcm_set_sync(subchn);
	snd_pcm_set_mixer(subchn, codec->mixer->device, codec->ac97->me_capture);
	return 0;
}


static int snd_ali_playback_close(void *private_data,
				      snd_pcm_subchn_t * subchn)
{
//	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
//	snd_pcm_runtime_t *runtime = subchn->runtime;
//	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;

	snd_pcm_dma_free(subchn);
	return 0;
}

static int snd_ali_capture_close(void *private_data,
				     snd_pcm_subchn_t * subchn)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;
	snd_ali_voice_t *pvoice = (snd_ali_voice_t *) runtime->private_data;

	snd_ali_disable_special_channel(codec,pvoice->number);

	snd_pcm_dma_free(subchn);

	return 0;
}


static void snd_ali_pcm_free(void *private_data)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, );
	codec->pcm = NULL;
}

int snd_ali_pcm(ali_t * codec, int device, snd_pcm_t ** rpcm)
{
	snd_pcm_t *pcm;
	int err;

	*rpcm = NULL;
	err = snd_pcm_new(codec->card, "ALI 5451", device, ALI_CHANNELS, 1, &pcm);
	if (err < 0) {
		snd_printk("snd_ali_pcm: err called snd_pcm_new.\n");
		return err;
	}

	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].private_data = codec;
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].open = snd_ali_playback_open;
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].close = snd_ali_playback_close;

	pcm->chn[SND_PCM_CHANNEL_CAPTURE].private_data = codec;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].open = snd_ali_capture_open;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].close = snd_ali_capture_close;

	pcm->private_data = codec;
	pcm->private_free = snd_ali_pcm_free;
	pcm->info_flags = SND_PCM_INFO_PLAYBACK | SND_PCM_INFO_CAPTURE |
			  SND_PCM_INFO_DUPLEX;
	strcpy(pcm->name, "ALI 5451 - pcm");
	*rpcm = codec->pcm = pcm;
	return 0;
}

static int snd_ali_spdif_get_switch(snd_kmixer_t * mixer,
					snd_kswitch_t * kswitch,
					snd_switch_t * uswitch)
{
	unsigned long flags;
	ali_t *codec = snd_magic_cast(ali_t, kswitch->private_data, -ENXIO);

	uswitch->type = SND_SW_TYPE_BOOLEAN;
	uswitch->subtype = SND_SW_SUBTYPE_HEXA;

	spin_lock_irqsave(&codec->reg_lock, flags);
	switch(kswitch->private_value) {
	case 0:
		uswitch->value.enable = (codec->spdif_mask & 0x02) ? 1 : 0;
		break;
	case 1:
		uswitch->value.enable = ((codec->spdif_mask & 0x02) && (codec->spdif_mask & 0x04)) ? 1 : 0;
		break;
	case 2:
		uswitch->value.enable = (codec->spdif_mask & 0x01) ? 1 : 0;
		break;
	default:
		break;
	}

	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return 0;
}

static int snd_ali_spdif_set_switch(snd_kmixer_t * mixer,
				        snd_kswitch_t * kswitch,
				        snd_switch_t * uswitch)
{
	ali_t *codec = snd_magic_cast(ali_t, kswitch->private_data, -ENXIO);
	unsigned int enable = 0, change = 0;
	unsigned long flags;

	if (uswitch->type != SND_SW_TYPE_BOOLEAN)
		return -EINVAL;
	enable = uswitch->value.enable ? 1 : 0;

	spin_lock_irqsave(&codec->reg_lock, flags);
	switch (kswitch->private_value) {
	case 0:
		change = (codec->spdif_mask & 0x02) ? 1 : 0;
		change = change ^ enable;
		if (change) {
			if (enable) {
				codec->spdif_mask |= 0x02;
				snd_ali_enable_spdif_out(codec);
			} else {
				codec->spdif_mask &= ~(0x02);
				codec->spdif_mask &= ~(0x04);
				snd_ali_disable_spdif_out(codec);
			}
		}
		break;
	case 1: 
		change = (codec->spdif_mask & 0x04) ? 1 : 0;
		change = change ^ enable;
		if (change && (codec->spdif_mask & 0x02)) {
			if (enable) {
				codec->spdif_mask |= 0x04;
				snd_ali_enable_spdif_chnout(codec);
			} else {
				codec->spdif_mask &= ~(0x04);
				snd_ali_disable_spdif_chnout(codec);
			}
		}
		break;
	case 2:
		change = (codec->spdif_mask & 0x01) ? 1 : 0;
		change = change ^ enable;
		if (change) {
			if (enable) {
				codec->spdif_mask |= 0x01;
				snd_ali_enable_spdif_in(codec);
			} else {
				codec->spdif_mask &= ~(0x01);
				snd_ali_disable_spdif_in(codec);
			}
		}
		break;
	default:
		break;
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	return change;
}

static snd_kswitch_t snd_ali_spdif_out_switch =
{
	name:           "S/PDIF Out",
	get:            (snd_get_switch_t *)snd_ali_spdif_get_switch,
	set:            (snd_set_switch_t *)snd_ali_spdif_set_switch,
	private_value:  0,
};
static snd_kswitch_t snd_ali_spdif_chnout_switch =
{
	name:           "S/PDIF Out to S/PDIF Channel",
	get:            (snd_get_switch_t *)snd_ali_spdif_get_switch,
	set:            (snd_set_switch_t *)snd_ali_spdif_set_switch,
	private_value:  1,
};
static snd_kswitch_t snd_ali_spdif_chnin_switch =
{
	name:           "S/PDIF In from S/PDIF Channel",
	get:            (snd_get_switch_t *)snd_ali_spdif_get_switch,
	set:            (snd_set_switch_t *)snd_ali_spdif_set_switch,
	private_value:  2,
};

static void snd_ali_mixer_free_ac97(void *private_data)
{
	ali_t *codec = snd_magic_cast(ali_t, private_data, );
	codec->mixer = NULL;
}

int snd_ali_mixer(ali_t * codec, int device, snd_pcm_t *pcm, snd_kmixer_t ** rmixer)
{
	snd_kmixer_t *mixer;
	ac97_t ac97;
	int err;

	*rmixer = NULL;
	memset(&ac97, 0, sizeof(ac97));
	ac97.write = snd_ali_codec_write;
	ac97.read = snd_ali_codec_read;
	ac97.private_data = codec;
	ac97.private_free = snd_ali_mixer_free_ac97;
	if ((err = snd_ac97_mixer(codec->card, device, &ac97, 1, &(pcm->device), &mixer)) < 0) {
		snd_printk("ALI mixer creating error.\n");
		return err;
	}

	if (codec->revision == ALI_5451_V02) {
		snd_mixer_switch_new(mixer, &snd_ali_spdif_chnin_switch, codec);
		snd_mixer_switch_new(mixer, &snd_ali_spdif_out_switch, codec);
		snd_mixer_switch_new(mixer, &snd_ali_spdif_chnout_switch, codec);
	}

	codec->ac97 = snd_magic_cast(ac97_t, mixer->private_data, -ENXIO);
	*rmixer = codec->mixer = mixer;
	return 0;
}

static int snd_ali_free(ali_t * codec)
{

	snd_ali_disable_address_interrupt(codec);
	synchronize_irq();
#ifdef CONFIG_PM
	if (codec->pm_dev)
		pm_unregister(codec->pm_dev);
	if (codec->image)
		snd_kfree(codec->image);
#endif
	snd_magic_kfree(codec);
	return 0;
}


static int snd_ali_chip_init(ali_t *codec)
{
	unsigned int legacy;
	unsigned char temp;
	struct pci_dev *pci_dev = NULL;

	if (snd_ali_reset_5451(codec)) {
		snd_printk("ali_chip_init: reset 5451 error.\n");
		return -1;
	}

	if (codec->revision == ALI_5451_V02) {
		pci_dev = pci_find_device(0x10b9,0x7101, pci_dev);
		if (pci_dev == NULL)
                	return -1;

		pci_read_config_byte(pci_dev,0xb8,&temp);
		temp |= 1 << 6;
		pci_write_config_byte(pci_dev, 0xB8, ~(0x20));
	}

	pci_read_config_dword(codec->pci, 0x44, &legacy);
	legacy &= 0xff00ff00;
	legacy |= 0x000800aa;
	pci_write_config_dword(codec->pci, 0x44, legacy);

	outl(0x80000001, ALI_REG(codec, ALI_GLOBAL_CONTROL));
	outl(0x00000000, ALI_REG(codec, ALI_AINTEN));
	outl(0xffffffff, ALI_REG(codec, ALI_AINT));
	outl(0x00000000, ALI_REG(codec, ALI_VOLUME));
	outb(0x10, 	 ALI_REG(codec, ALI_MPUR2));

	codec->ac97_ext_id = snd_ali_codec_read(codec, AC97_EXTENDED_ID);
	codec->ac97_ext_status = snd_ali_codec_read(codec, AC97_EXTENDED_STATUS);
	if (codec->revision == ALI_5451_V02)
		snd_ali_enable_spdif_out(codec);
	return 0;

}

#ifdef CONFIG_PM
static void snd_ali_suspend(ali_t *chip)
{
	ali_image_t *im;
	unsigned long flags;
	int i, j;

	im = chip->image;
	if (! im)
		return;

	save_flags(flags); 
	cli();
	
	im->regs[ALI_MISCINT >> 2] = inl(ALI_REG(chip, ALI_MISCINT));
	// im->regs[ALI_START >> 2] = inl(ALI_REG(chip, ALI_START));
	im->regs[ALI_STOP >> 2] = inl(ALI_REG(chip, ALI_STOP));
	
	// disable all IRQ bits
	outl(0, ALI_REG(chip, ALI_MISCINT));
	
	for (i = 0; i < ALI_GLOBAL_REGS; i++) {	
		if ((i*4 == ALI_MISCINT) || (i*4 == ALI_STOP))
			continue;
		im->regs[i] = inl(ALI_REG(chip, i*4));
	}
	
	for (i = 0; i < ALI_CHANNELS; i++) {
		outb(i, ALI_REG(chip, ALI_GC_CIR));
		for (j = 0; j < ALI_CHANNEL_REGS; j++) 
			im->channel_regs[i][j] = inl(ALI_REG(chip, j*4 + 0xe0));
	}

	// stop all HW channel
	outl(0xffffffff, ALI_REG(chip, ALI_STOP));

	restore_flags(flags);
}

static void snd_ali_resume(ali_t *chip)
{
	ali_image_t *im;
	unsigned long flags;
	int i, j;

	im = chip->image;
	if (! im)
		return;

	save_flags(flags); 
	cli();
	
	for (i = 0; i < ALI_CHANNELS; i++) {
		outb(i, ALI_REG(chip, ALI_GC_CIR));
		for (j = 0; j < ALI_CHANNEL_REGS; j++) 
			outl(im->channel_regs[i][j], ALI_REG(chip, j*4 + 0xe0));
	}
	
	for (i = 0; i < ALI_GLOBAL_REGS; i++) {	
		if ((i*4 == ALI_MISCINT) || (i*4 == ALI_STOP) || (i*4 == ALI_START))
			continue;
		outl(im->regs[i], ALI_REG(chip, i*4));
	}
	
	snd_ac97_resume(chip->ac97);
	
	// start HW channel
	outl(im->regs[ALI_START >> 2], ALI_REG(chip, ALI_START));
	// restore IRQ enable bits
	outl(im->regs[ALI_MISCINT >> 2], ALI_REG(chip, ALI_MISCINT));
	
	restore_flags(flags);
}

static int
snd_ali5451_pm_callback(struct pm_dev *dev, pm_request_t request, void *data)
{
	ali_t *chip = snd_magic_cast(ali_t, dev->data, 0);

	switch (request) {
	case PM_SUSPEND:
		snd_ali_suspend(chip);
		break;
	case PM_RESUME:
		snd_ali_resume(chip);
		break;
	}
	return 0;
}
#endif

static int __init snd_ali_create(snd_card_t * card,
				     struct pci_dev *pci,
				     snd_dma_t * dma_pbk,
				     snd_dma_t * dma_cap,
				     snd_irq_t * irqptr,
				     int pcm_channels,
				     ali_t ** r_ali)
{
	ali_t *codec;
	int i, err;
	unsigned short cmdw = 0;
	struct pci_dev *pci_dev = NULL;
        static snd_device_ops_t ops = {
		(snd_dev_free_t *)snd_ali_free,
		NULL,
		NULL
        };

	if ((codec = snd_magic_kcalloc(ali_t, 0, GFP_KERNEL)) == NULL)
		return -ENOMEM;

	spin_lock_init(&codec->reg_lock);
	spin_lock_init(&codec->multi_synth);
	spin_lock_init(&codec->voice_alloc);

	codec->card = card;
	codec->pci = pci;
	codec->dma_pbk = dma_pbk;
	codec->dma_cap = dma_cap;
	codec->irqptr = irqptr;
	codec->port = pci_resource_start(pci, 0);

	pci_read_config_byte(pci, PCI_REVISION_ID, &codec->revision);
	
	pci_set_master(pci);
	pci_read_config_word(pci, PCI_COMMAND, &cmdw);
	if ((cmdw & PCI_COMMAND_IO) != PCI_COMMAND_IO) {
		cmdw |= PCI_COMMAND_IO;
		pci_write_config_word(pci, PCI_COMMAND, cmdw);
	}
	pci_set_master(pci);
	
	synchronize_irq();

	codec->synth.chmap = 0;
	codec->synth.chcnt = 0;
	codec->synth.synthcount = 0;
	codec->spdif_mask = 0x00000002;

	if (codec->revision == ALI_5451_V02)
		codec->chregs.regs.ac97read = ALI_AC97_WRITE;
	else
		codec->chregs.regs.ac97read = ALI_AC97_READ;
	codec->chregs.regs.ac97write = ALI_AC97_WRITE;

	codec->chregs.regs.start  = ALI_START;
	codec->chregs.regs.stop   = ALI_STOP;
	codec->chregs.regs.aint   = ALI_AINT;
	codec->chregs.regs.ainten = ALI_AINTEN;

	codec->chregs.data.start  = 0x00;
	codec->chregs.data.stop   = 0x00;
	codec->chregs.data.aint   = 0x00;
	codec->chregs.data.ainten = 0x00;

       	pci_dev = pci_find_device(0x10b9,0x1533, pci_dev);
	codec->pci_m1533 = pci_dev;
       	pci_dev = pci_find_device(0x10b9,0x7101, pci_dev);
	codec->pci_m7101 = pci_dev;

	if ((err = snd_device_new(card, SND_DEV_LOWLEVEL, codec, 0, &ops, NULL)) < 0) {
		snd_ali_free(codec);
		return err;
	}

	/* initialise synth voices*/
	for (i = 0; i < ALI_CHANNELS; i++ ) {
		codec->synth.voices[i].number = i;
	}

	if ((err = snd_ali_chip_init(codec)) < 0) {
		snd_printk("ali create: chip init error.\n");
		snd_ali_free(codec);
		return err;
	}

#ifdef CONFIG_PM
	codec->image = snd_kmalloc(sizeof(*codec->image), GFP_KERNEL);
	if (! codec->image) {
		snd_printk("ali5451: can't allocate apm buffer\n");
	} else {
		codec->pm_dev = pm_register(PM_PCI_DEV, PM_PCI_ID(pci),
					    snd_ali5451_pm_callback);
		if (codec->pm_dev)
			codec->pm_dev->data = codec;
	}
#endif

	snd_ali_enable_address_interrupt(codec);

	*r_ali = codec;
	return 0;
}

static void snd_ali_use_inc(snd_card_t * card)
{
	MOD_INC_USE_COUNT;
}

static void snd_ali_use_dec(snd_card_t * card)
{
	MOD_DEC_USE_COUNT;
}

static int __init snd_ali_detect(snd_card_t * card, ali_card_t *scard,
				      unsigned short vendor, unsigned short device)
{
	if ((scard->pci = pci_find_device(vendor, device, scard->pci)) == NULL)
		return -ENODEV;
	if (snd_register_ioport(card, pci_resource_start(scard->pci, 0), 256, "ALI 5451", NULL) < 0)
		goto __nodev;
	return 0;
      __nodev:
	snd_unregister_ioports(card);
	return -ENODEV;
}

static int __init snd_ali_resources(snd_card_t * card,
					ali_card_t *scard,
					int dev)
{
	int err;

	if ((err = snd_register_interrupt(card,
			"ALI 5451", scard->pci->irq,
			SND_IRQ_TYPE_PCI, snd_ali_card_interrupt,
			scard, NULL, &scard->irqptr)) < 0)
		return err;
	if ((err = snd_register_dma_channel(card,
			"ALI 5451 - playback", 0,
			SND_DMA_TYPE_PCI, snd_dac_frame_size[dev],
			NULL, &scard->dma_pbk)) < 0)
		return err;
	scard->dma_pbk->multi = 1;

	if ((err = snd_register_dma_channel(card,
			"ALI 5451 - capture", 1,
			SND_DMA_TYPE_PCI, snd_adc_frame_size[dev],
			NULL, &scard->dma_cap)) < 0)
		return err;
	return 0;
}

static int __init snd_ali_probe(int dev, ali_card_t *scard)
{
	snd_card_t *card;
	snd_pcm_t *pcm = NULL;
	snd_kmixer_t *mixer = NULL;
	ali_t *codec = NULL;

	card = snd_card_new(snd_index[dev], snd_id[dev],
			    snd_ali_use_inc, snd_ali_use_dec);
	if (card == NULL)
		return -ENOMEM;
	do {
		if (!snd_ali_detect(card, scard, PCI_VENDOR_ID_ALI, PCI_DEVICE_ID_ALI_5451))
			break;
	} while (scard->pci);
	
	card->type = SND_CARD_TYPE_ALI_5451;

	if (scard->pci == NULL) {
		snd_card_free(card);
		return -ENODEV;
	}

	if (pci_enable_device(scard->pci)) {
		snd_card_free(card);
		return -ENODEV;
	}
	if (snd_ali_resources(card, scard, dev) < 0) {
		snd_card_free(card);
		return -ENODEV;
	}

	if (snd_ali_create(    card,
			       scard->pci,
			       scard->dma_pbk,
			       scard->dma_cap,
			       scard->irqptr,
			       snd_pcm_channels[dev],
			       &scard->codec) < 0)
		goto __nodev;
	codec = scard->codec;

	if (snd_ali_pcm(codec, 0, &pcm) < 0)
		goto __nodev;
	
	if (snd_ali_mixer(codec, 0, pcm, &mixer) < 0)
		goto __nodev;

	strcpy(card->abbreviation, "ALI5451");
	strcpy(card->shortname, "ALI 5451");
	
	sprintf(card->longname, "%s at 0x%lx, irq %li",
		card->shortname, codec->port, scard->irqptr->irq);

	if (!snd_card_register(card)) {
		scard->card = card;
		scard->mixer = mixer;
		scard->pcm = pcm;
		return 0;
	}
	goto __nodev;

      __nodev:
	snd_printd("ali __nodev error.\n");
	snd_card_free(card);
	return -ENXIO;
}

#ifdef MODULE

static int __exit snd_ali_card_free(int dev)
{
	ali_card_t *scard;

	scard = snd_ali_cards[dev];
	snd_ali_cards[dev] = NULL;
	if (scard) {
		snd_card_unregister(scard->card);
		snd_kfree(scard);
	}
	return 0;
}

#endif

int __init alsa_card_ali_init(void)
{
	int dev, cards;
	ali_card_t *scard;

	for (dev = cards = 0; dev < SND_CARDS; dev++) {
		scard = (ali_card_t *)
				snd_kcalloc(sizeof(ali_card_t), GFP_KERNEL);
		if (scard == NULL)
			continue;
		if (snd_ali_probe(dev, scard) < 0) {
			snd_kfree(scard);
			break;
		}
		snd_ali_cards[dev] = scard;
		cards++;
	}
	if (!cards) {
		return -ENODEV;
	}
	return 0;
}

void __exit alsa_card_ali_exit(void)
{
	int dev;

	for (dev = 0; dev < SND_CARDS; dev++)
		snd_ali_card_free(dev);
}

module_init(alsa_card_ali_init)
module_exit(alsa_card_ali_exit)
