/*
 *   ALSA driver for Intel ICH (i8x0) chipsets
 *
 *	Copyright (c) 2000 Jaroslav Kysela <perex@suse.cz>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */      

#define SND_MAIN_OBJECT_FILE

#include "../include/driver.h"
#include "../include/pcm.h"
#include "../include/mixer.h"
#include "../include/ac97_codec.h"
#include "../include/info.h"
#include "../include/initval.h"

EXPORT_NO_SYMBOLS;
MODULE_DESCRIPTION("\
Driver: Intel 82801AA,82801BA,82901AB,i810,i815,i820,i830,i840,MX440\n\
PCI: 0x8086=0x2415\n\
PCI: 0x8086=0x2425\n\
PCI: 0x8086=0x2445\n\
PCI: 0x8086=0x2485\n\
PCI: 0x8086=0x7195\n\
");
MODULE_LICENSE("GPL");

static int snd_index[SND_CARDS] = SND_DEFAULT_IDX;	/* Index 0-MAX */
static char *snd_id[SND_CARDS] = SND_DEFAULT_STR;	/* ID for this card */
static int snd_enable[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 1}; /* all enabled */
static int snd_pbk_frame_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 128};
static int snd_cap_frame_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 128};
static int snd_mic_frame_size[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 128};
static int snd_ac97_clock[SND_CARDS] = {[0 ... (SND_CARDS - 1)] = 48000};
MODULE_AUTHOR("Jaroslav Kysela <perex@suse.cz>");
MODULE_PARM(snd_index, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_index, "Index value for Intel i8x0 soundcard.");
MODULE_PARM(snd_id, "1-" __MODULE_STRING(SND_CARDS) "s");
MODULE_PARM_DESC(snd_id, "ID string for Intel i8x0 soundcard.");
MODULE_PARM(snd_enable, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_enable, "Enable this soundcard. [BOOL]");
MODULE_PARM(snd_pbk_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_pbk_frame_size, "Playback frame size in kB.");
MODULE_PARM(snd_cap_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_cap_frame_size, "Capture frame size in kB.");
MODULE_PARM(snd_mic_frame_size, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_mic_frame_size, "MIC capture frame size in kB.");
MODULE_PARM(snd_ac97_clock, "1-" __MODULE_STRING(SND_CARDS) "i");
MODULE_PARM_DESC(snd_ac97_clock, "AC'97 codec clock (default 48000Hz).");

#ifndef PCI_DEVICE_ID_INTEL_82801
#define PCI_DEVICE_ID_INTEL_82801	0x2415
#endif
#ifndef PCI_DEVICE_ID_INTEL_82801BA
#define PCI_DEVICE_ID_INTEL_82801BA	0x2445
#endif
#ifndef PCI_DEVICE_ID_INTEL_82901
#define PCI_DEVICE_ID_INTEL_82901	0x2425
#endif
#ifndef PCI_DEVICE_ID_INTEL_440MX
#define PCI_DEVICE_ID_INTEL_440MX	0x7195
#endif
#ifndef PCI_DEVICE_ID_INTEL_ICH3
#define PCI_DEVICE_ID_INTEL_ICH3	0x2485
#endif

/*
 *  Direct registers
 */

#define ICHREG(ice, x) ((ice)->bmport + ICH_REG_##x)

/* capture block */
#define ICH_REG_PI_BDBAR		0x00	/* dword - buffer descriptor list base address */
#define ICH_REG_PI_CIV			0x04	/* byte - current index value */
#define ICH_REG_PI_LVI			0x05	/* byte - last valid index */
#define   ICH_REG_LVI_MASK		0x1f
#define ICH_REG_PI_SR			0x06	/* byte - status register */
#define   ICH_FIFOE			0x10	/* FIFO error */
#define   ICH_BCIS			0x08	/* buffer completion interrupt status */
#define   ICH_LVBCI			0x04	/* last valid buffer completion interrupt */
#define   ICH_CELV			0x02	/* current equals last valid */
#define   ICH_DCH			0x01	/* DMA controller halted */
#define ICH_REG_PI_PICB			0x08	/* word - position in current buffer */
#define ICH_REG_PI_PIV			0x0a	/* byte - prefetched index value */
#define   ICH_REG_PIV_MASK		0x1f	/* mask */
#define ICH_REG_PI_CR			0x0b	/* byte - control register */
#define   ICH_IOCE			0x10	/* interrupt on completion enable */
#define   ICH_FEIE			0x08	/* fifo error interrupt enable */
#define   ICH_LVBIE			0x04	/* last valid buffer interrupt enable */
#define   ICH_RESETREGS			0x02	/* reset busmaster registers */
#define   ICH_STARTBM			0x01	/* start busmaster operation */
/* playback block */
#define ICH_REG_PO_BDBAR		0x10	/* dword - buffer descriptor list base address */
#define ICH_REG_PO_CIV			0x14	/* byte - current index value */
#define ICH_REG_PO_LVI			0x15	/* byte - last valid command */
#define ICH_REG_PO_SR			0x16	/* byte - status register */
#define ICH_REG_PO_PICB			0x18	/* word - position in current buffer */
#define ICH_REG_PO_PIV			0x1a	/* byte - prefetched index value */
#define ICH_REG_PO_CR			0x1b	/* byte - control register */
/* mic capture block */
#define ICH_REG_MC_BDBAR		0x20	/* dword - buffer descriptor list base address */
#define ICH_REG_MC_CIV			0x24	/* byte - current index value */
#define ICH_REG_MC_LVI			0x25	/* byte - last valid command */
#define ICH_REG_MC_SR			0x26	/* byte - status register */
#define ICH_REG_MC_PICB			0x28	/* word - position in current buffer */
#define ICH_REG_MC_PIV			0x2a	/* byte - prefetched index value */
#define ICH_REG_MC_CR			0x2b	/* byte - control register */
/* global block */
#define ICH_REG_GLOB_CNT		0x2c	/* dword - global control */
#define   ICH_PCM_246_MASK	0x00300000	/* 6 channels (not all chips) */
#define   ICH_PCM_6		0x00200000	/* 6 channels (not all chips) */
#define   ICH_PCM_4		0x00100000	/* 4 channels (not all chips) */
#define   ICH_PCM_2		0x00000000	/* 2 channels (stereo) */
#define   ICH_SRIE		0x00000020	/* secondary resume interrupt enable */
#define   ICH_PRIE		0x00000010	/* primary resume interrupt enable */
#define   ICH_ACLINK		0x00000008	/* AClink shut off */
#define   ICH_AC97WARM		0x00000004	/* AC'97 warm reset */
#define   ICH_AC97COLD		0x00000002	/* AC'97 cold reset */
#define   ICH_GIE		0x00000001	/* GPI interrupt enable */
#define ICH_REG_GLOB_STA		0x30	/* dword - global status */
#define   ICH_MD3		0x00020000	/* modem power down semaphore */
#define   ICH_AD3		0x00010000	/* audio power down semaphore */
#define   ICH_RCS		0x00008000	/* read completion status */
#define   ICH_BIT3		0x00004000	/* bit 3 slot 12 */
#define   ICH_BIT2		0x00002000	/* bit 2 slot 12 */
#define   ICH_BIT1		0x00001000	/* bit 1 slot 12 */
#define   ICH_SRI		0x00000800	/* secondary resume interrupt */
#define   ICH_PRI		0x00000400	/* primary resume interrupt */
#define   ICH_SCR		0x00000200	/* secondary codec ready */
#define   ICH_PCR		0x00000100	/* primary codec ready */
#define   ICH_MCINT		0x00000080	/* MIC capture interrupt */
#define   ICH_POINT		0x00000040	/* playback interrupt */
#define   ICH_PIINT		0x00000020	/* capture interrupt */
#define   ICH_MOINT		0x00000004	/* modem playback interrupt */
#define   ICH_MIINT		0x00000002	/* modem capture interrupt */
#define   ICH_GSCI		0x00000001	/* GPI status change interrupt */
#define ICH_REG_ACC_SEMA		0x34	/* byte - codec write semaphore */
#define   ICH_CAS			0x01	/* codec access semaphore */

/*
 *  
 */

typedef struct {
	unsigned int reg_offset;
	unsigned int *bdbar;
	unsigned int rates;
	int min_rate, max_rate;
        snd_pcm_subchn_t *subchn;
        unsigned long physbuf;
        unsigned int size;
        unsigned int fragsize;
        unsigned int fragsize1;
        unsigned int position;
        int frags;
        int lvi;
        int lvi_frag;
	int ack;
	int ack_reload;
#ifdef CONFIG_PM
	unsigned char civ_saved;
	unsigned char piv_saved;
	unsigned short picb_saved;
#endif
} ichdev_t;

typedef struct snd_stru_intel8x0 intel8x0_t;

struct snd_stru_intel8x0 {
	snd_dma_t * dma_pbk;	/* playback */
	snd_dma_t * dma_cap;	/* capture */
	snd_dma_t * dma_mic;	/* mic capture */
	snd_irq_t * irqptr;

	unsigned int port;
	unsigned int bmport;

	struct pci_dev *pci;
	snd_card_t *card;

	snd_pcm_t *pcm;
	snd_pcm_t *pcm_mic;
	ichdev_t playback;
	ichdev_t capture;
	ichdev_t capture_mic;

	ac97_t *ac97;
	unsigned int ac97_clock;
	unsigned short ac97_ext_id;
	snd_kmixer_t *mixer;

	spinlock_t reg_lock;
	snd_info_entry_t *proc_entry;

	unsigned int *bdbars;
#ifdef CONFIG_PM
	struct pm_dev *pm_dev;
	int in_suspend;
#endif
};


/*
 *
 */

static struct pci_device_id snd_intel8x0_ids[] __devinitdata = {
	{ 0x8086, PCI_DEVICE_ID_INTEL_82801, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0x8086, PCI_DEVICE_ID_INTEL_82801BA, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0x8086, PCI_DEVICE_ID_INTEL_82901, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0x8086, PCI_DEVICE_ID_INTEL_440MX, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0x8086, PCI_DEVICE_ID_INTEL_ICH3, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
	{ 0, }
};

MODULE_DEVICE_TABLE(pci, snd_intel8x0_ids);

static void snd_intel8x0_use_inc(snd_card_t * card)
{
	MOD_INC_USE_COUNT;
}

static void snd_intel8x0_use_dec(snd_card_t * card)
{
	MOD_DEC_USE_COUNT;
}

/*
 *  Basic I/O
 */
 
static int snd_intel8x0_codec_semaphore(intel8x0_t *codec, int secondary)
{
	int time;
	
	/* codec ready ? */
	if ((inl(ICHREG(codec, GLOB_STA)) & (secondary ? ICH_SCR : ICH_PCR)) == 0)
		return -EIO;

	/* Anyone holding a semaphore for 1 msec should be shot... */
	time = 100;
      	do {
      		if (!(inb(ICHREG(codec, ACC_SEMA)) & ICH_CAS))
      			return 0;
		udelay(10);
	} while (time--);
	snd_printk("snd_intel8x0_codec_semaphore: codec %i semaphore is not ready [0x%x][0x%x]\n", secondary, inb(ICHREG(codec, ACC_SEMA)), inl(ICHREG(codec, GLOB_STA)));
	inw(codec->port);	/* clear semaphore flag */
	return -EBUSY;
}

static void snd_intel8x0_codec_write(void *private_data,
				     unsigned short reg,
				     unsigned short val)
{
	intel8x0_t *codec = (intel8x0_t *)private_data;
	
	snd_intel8x0_codec_semaphore(codec, 0);
	return outw(val, codec->port + reg);
}

static unsigned short snd_intel8x0_codec_read(void *private_data,
					      unsigned short reg)
{
	intel8x0_t *codec = (intel8x0_t *)private_data;

	if (snd_intel8x0_codec_semaphore(codec, 0) < 0)
		return ~0;
	return inw(codec->port + reg);
}

static int snd_intel8x0_trigger(intel8x0_t *codec, ichdev_t *ichdev, int cmd)
{
	unsigned char val = 0;
	unsigned int port = codec->bmport + ichdev->reg_offset;
	
	switch (cmd) {
	case SND_PCM_TRIGGER_GO:
		val = ICH_IOCE | ICH_STARTBM;
		break;
	case SND_PCM_TRIGGER_STOP:
		val = 0;
		break;
	case SND_PCM_TRIGGER_PAUSE_PUSH:
		val = ICH_IOCE;
		break;
	case SND_PCM_TRIGGER_PAUSE_RELEASE:
		val = ICH_IOCE | ICH_STARTBM;
		break;
	default:
		return -EINVAL;
	}
	outb(val, port + ICH_REG_PI_CR);
	if (cmd == SND_PCM_TRIGGER_STOP) {
		/* reset whole DMA things */
		while (!(inb(port + ICH_REG_PI_SR) & ICH_DCH)) ;
		outb(ICH_RESETREGS, port + ICH_REG_PI_CR);
	}
	return 0;
}

static void snd_intel8x0_setup_fragments(intel8x0_t *codec, ichdev_t *ichdev) 
{
	int idx;
	unsigned int *bdbar = ichdev->bdbar;
	unsigned int port = codec->bmport + ichdev->reg_offset;

	outl(virt_to_bus(ichdev->bdbar), port + ICH_REG_PI_BDBAR);
	if (ichdev->size == ichdev->fragsize) {
		ichdev->ack_reload = ichdev->ack = 2;
		ichdev->fragsize1 = ichdev->fragsize >> 1;
		for (idx = 0; idx < (ICH_REG_LVI_MASK + 1) * 2; idx += 4) {
			bdbar[idx + 0] = ichdev->physbuf;
			bdbar[idx + 1] = 0x80000000 | /* interrupt on completion */
					 ichdev->fragsize1 >> 1;
			bdbar[idx + 2] = ichdev->physbuf + (ichdev->size >> 1);
			bdbar[idx + 3] = 0x80000000 | /* interrupt on completion */
					 ichdev->fragsize1 >> 1;
		}
		ichdev->frags = 2;
	} else {
		ichdev->ack_reload = ichdev->ack = 1;
		ichdev->fragsize1 = ichdev->fragsize;
		for (idx = 0; idx < (ICH_REG_LVI_MASK + 1) * 2; idx += 2) {
			bdbar[idx + 0] = ichdev->physbuf + (((idx >> 1) * ichdev->fragsize) % ichdev->size);
			bdbar[idx + 1] = 0x80000000 | /* interrupt on completion */
					 ichdev->fragsize >> 1;
			// printk("bdbar[%i] = 0x%x [0x%x]\n", idx + 0, bdbar[idx + 0], bdbar[idx + 1]);
		}
		ichdev->frags = ichdev->size / ichdev->fragsize;
	}
	outb(0, port + ICH_REG_PI_CIV);
	outb(ichdev->lvi = ICH_REG_LVI_MASK, port + ICH_REG_PI_LVI);
	ichdev->lvi_frag = ICH_REG_LVI_MASK % ichdev->frags;
	ichdev->position = 0;
#if 0
	printk("lvi_frag = %i, frags = %i, frag_size = 0x%x, frag_size1 = 0x%x\n",
			ichdev->lvi_frag, ichdev->frags, ichdev->fragsize, ichdev->fragsize1);
#endif
	/* clear interrupts */
	outb(ICH_FIFOE | ICH_BCIS | ICH_LVBCI, port + ICH_REG_PI_SR);
}

/*
 *  Interrupt handler
 */

static inline void snd_intel8x0_update(intel8x0_t *codec, ichdev_t *ichdev)
{
	unsigned int port = codec->bmport + ichdev->reg_offset;
	int ack = 0;

	spin_lock(&codec->reg_lock);
	ichdev->position += ichdev->fragsize1;
	ichdev->position %= ichdev->size;
	ichdev->lvi++;
	ichdev->lvi &= ICH_REG_LVI_MASK;
	outb(ichdev->lvi, port + ICH_REG_PI_LVI);
	ichdev->lvi_frag++;
	ichdev->lvi_frag %= ichdev->frags;
	ichdev->bdbar[ichdev->lvi * 2] = ichdev->physbuf + ichdev->lvi_frag * ichdev->fragsize1;
	// printk("new: bdbar[%i] = 0x%x [0x%x], prefetch = %i, all = 0x%x, 0x%x\n", ichdev->lvi * 2, ichdev->bdbar[ichdev->lvi * 2], ichdev->bdbar[ichdev->lvi * 2 + 1], inb(ICH_REG_PI_PIV + port), inl(port + 4), inb(port + ICH_REG_PI_CR));
	if ((ack = (--ichdev->ack == 0)) != 0)
		ichdev->ack = ichdev->ack_reload;
	spin_unlock(&codec->reg_lock);
	if (ack && ichdev->subchn)
		snd_pcm_transfer_done(ichdev->subchn);
	outb(ICH_FIFOE | ICH_BCIS | ICH_LVBCI, port + ICH_REG_PI_SR);
}

static void snd_intel8x0_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	intel8x0_t *codec = (intel8x0_t *) dev_id;
	unsigned int status;
	
	status = inl(ICHREG(codec, GLOB_STA));
	if ((status & (ICH_MCINT | ICH_POINT | ICH_PIINT)) == 0)
		return;
	if (status & ICH_POINT)
		snd_intel8x0_update(codec, &codec->playback);
	if (status & ICH_PIINT)
		snd_intel8x0_update(codec, &codec->capture);
	if (status & ICH_MCINT)
		snd_intel8x0_update(codec, &codec->capture_mic);
}

/*
 *  PCM part
 */

static int snd_intel8x0_playback_ioctl(void *private_data,
				      snd_pcm_subchn_t * subchn,
				      unsigned int cmd,
				      unsigned long *arg)
{
	int result;
	// intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);
	if (result < 0)
		return result;
	return 0;
}

static int snd_intel8x0_capture_ioctl(void *private_data,
				      snd_pcm_subchn_t * subchn,
				      unsigned int cmd,
				      unsigned long *arg)
{
	int result;
	// intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);
	if (result < 0)
		return result;
	return 0;
}

static int snd_intel8x0_playback_trigger(void *private_data,
					snd_pcm_subchn_t * pcm,
					int cmd)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	return snd_intel8x0_trigger(codec, &codec->playback, cmd);
}

static int snd_intel8x0_capture_trigger(void *private_data,
				       snd_pcm_subchn_t * subchn,
				       int cmd)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	return snd_intel8x0_trigger(codec, &codec->capture, cmd);
}

static int snd_intel8x0_playback_prepare(void *private_data,
					 snd_pcm_subchn_t * subchn)
{
	unsigned long flags;
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;

	codec->playback.physbuf = virt_to_bus(runtime->dma_area->buf);
	codec->playback.size = snd_pcm_lib_transfer_size(subchn);
	codec->playback.fragsize = snd_pcm_lib_transfer_fragment(subchn);
	/* disable DAC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_POWERDOWN, ~0x0200, 0x0200);
	snd_ac97_write_lock(codec->ac97, AC97_PCM_FRONT_DAC_RATE,
			    (codec->ac97_clock * runtime->format.rate) / 48000);
	/* enable DAC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_POWERDOWN, ~0x0200, 0x0000);
#if 0
	printk("prepare: AC97_PCM_FRONT_DAC_RATE=0x%x (%i), AC97_EXTENDED_STATUS=0x%x\n",
		snd_ac97_read_lock(codec->ac97, AC97_PCM_FRONT_DAC_RATE),
		snd_ac97_read_lock(codec->ac97, AC97_PCM_FRONT_DAC_RATE),
		snd_ac97_read_lock(codec->ac97, AC97_EXTENDED_STATUS));
#endif
	spin_lock_irqsave(&codec->reg_lock, flags);
	snd_intel8x0_setup_fragments(codec, &codec->playback);
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return 0;
}

static int snd_intel8x0_capture_prepare(void *private_data,
				       snd_pcm_subchn_t * subchn)
{
	unsigned long flags;
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;

	codec->capture.physbuf = virt_to_bus(runtime->dma_area->buf);
	codec->capture.size = snd_pcm_lib_transfer_size(subchn);
	codec->capture.fragsize = snd_pcm_lib_transfer_fragment(subchn);
	/* disable ADC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_POWERDOWN, ~0x0100, 0x0100);
	snd_ac97_write_lock(codec->ac97, AC97_PCM_LR_ADC_RATE,
			    (codec->ac97_clock * runtime->format.rate) / 48000);
	/* enable ADC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_POWERDOWN, ~0x0100, 0x0000);
	spin_lock_irqsave(&codec->reg_lock, flags);
	snd_intel8x0_setup_fragments(codec, &codec->capture);
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return 0;
}

static unsigned int snd_intel8x0_playback_pointer(void *private_data,
						  snd_pcm_subchn_t * subchn)
{
	unsigned long flags;
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	unsigned int result;

	spin_lock_irqsave(&codec->reg_lock, flags);
	result = codec->playback.fragsize1 - (inw(ICHREG(codec, PO_PICB)) << 1);
	result += codec->playback.position;
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return result;
}

static unsigned int snd_intel8x0_capture_pointer(void *private_data,
						snd_pcm_subchn_t * subchn)
{
	unsigned long flags;
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	unsigned int result;

	spin_lock_irqsave(&codec->reg_lock, flags);
	result = codec->capture.fragsize1 - (inw(ICHREG(codec, PI_PICB)) << 1);
	result += codec->capture.position;
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return result;
}

static snd_pcm_hardware_t snd_intel8x0_playback =
{
	SND_PCM_CHNINFO_MMAP | SND_PCM_CHNINFO_STREAM |
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE |
	SND_PCM_CHNINFO_BLOCK_TRANSFER |
	SND_PCM_CHNINFO_MMAP_VALID |
	SND_PCM_CHNINFO_PAUSE,	/* flags */
	SND_PCM_FMT_S16_LE,	/* hardware formats */
	0,			/* overwritten */
	8000,			/* min. rate */
	48000,			/* max. rate */
	2,			/* min. voices */
	2,			/* max. voices */
	32,			/* min. fragment size */
	128 * 1024,		/* max. fragment size */
	3,			/* fragment align */
	0,			/* FIFO size (unknown) */
	16,			/* transfer block size */
	snd_intel8x0_playback_ioctl,
	snd_intel8x0_playback_prepare,
	snd_intel8x0_playback_trigger,
	snd_intel8x0_playback_pointer
};

static snd_pcm_hardware_t snd_intel8x0_capture =
{
	SND_PCM_CHNINFO_MMAP | SND_PCM_CHNINFO_STREAM |
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE |
	SND_PCM_CHNINFO_BLOCK_TRANSFER |
	SND_PCM_CHNINFO_MMAP_VALID,	/* flags */
	SND_PCM_FMT_S16_LE,	/* hardware formats */
	0,			/* overwritten */
	8000,			/* min. rate */
	48000,			/* max. rate */
	2,			/* min. voices */
	2,			/* max. voices */
	32,			/* min. fragment size */
	128 * 1024,		/* max. fragment size */
	3,			/* fragment align */
	0,			/* FIFO size (unknown) */
	16,			/* transfer block size */
	snd_intel8x0_capture_ioctl,
	snd_intel8x0_capture_prepare,
	snd_intel8x0_capture_trigger,
	snd_intel8x0_capture_pointer
};

static int snd_intel8x0_playback_open(void *private_data,
			             snd_pcm_subchn_t * subchn)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	snd_pcm_hardware_t *hw;
	int err;

	hw = (snd_pcm_hardware_t *)snd_kmalloc(sizeof(snd_pcm_hardware_t), GFP_KERNEL);
	if (hw == NULL)
		return -ENOMEM;
	if ((err = snd_pcm_dma_alloc(subchn, codec->dma_pbk,
				     "Intel ICH - DAC")) < 0) {
		snd_kfree(hw);
		return err;
	}
	codec->playback.subchn = subchn;
	memcpy(hw, &snd_intel8x0_playback, sizeof(*hw));
	hw->rates = codec->playback.rates;
	hw->min_rate = codec->playback.min_rate;
	hw->max_rate = codec->playback.max_rate;
	subchn->runtime->hw = hw;
	subchn->runtime->hw_free = (snd_kfree_type)_snd_kfree;
	snd_pcm_set_mixer(subchn, codec->mixer->device, codec->ac97->me_playback);
	return 0;
}

static int snd_intel8x0_capture_open(void *private_data,
				    snd_pcm_subchn_t * subchn)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	snd_pcm_hardware_t *hw;
	int err;

	hw = (snd_pcm_hardware_t *)snd_kmalloc(sizeof(snd_pcm_hardware_t), GFP_KERNEL);
	if (hw == NULL)
		return -ENOMEM;
	if ((err = snd_pcm_dma_alloc(subchn, codec->dma_cap,
				     "Intel ICH - ADC")) < 0) {
		snd_kfree(hw);
		return err;
	}
	codec->capture.subchn = subchn;
	memcpy(hw, &snd_intel8x0_capture, sizeof(*hw));
	hw->rates = codec->capture.rates;
	hw->min_rate = codec->capture.min_rate;
	hw->max_rate = codec->capture.max_rate;
	subchn->runtime->hw = hw;
	subchn->runtime->hw_free = (snd_kfree_type)_snd_kfree;
	snd_pcm_set_mixer(subchn, codec->mixer->device, codec->ac97->me_capture);
	return 0;
}

static int snd_intel8x0_playback_close(void *private_data,
				      snd_pcm_subchn_t * subchn)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	codec->playback.subchn = NULL;
	snd_pcm_dma_free(subchn);
	/* disable DAC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_POWERDOWN, ~0x0200, 0x0200);
	return 0;
}

static int snd_intel8x0_capture_close(void *private_data,
				     snd_pcm_subchn_t * subchn)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	codec->capture.subchn = NULL;
	snd_pcm_dma_free(subchn);
	/* disable ADC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_POWERDOWN, ~0x0100, 0x0100);
	return 0;
}

static void snd_intel8x0_pcm_free(void *private_data)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, );
	codec->pcm = NULL;
}

static int snd_intel8x0_pcm(intel8x0_t * codec, int device, snd_pcm_t ** rpcm)
{
	snd_pcm_t *pcm;
	int err;

	*rpcm = NULL;
	err = snd_pcm_new(codec->card, "Intel ICH", device, 1, 1, &pcm);
	if (err < 0)
		return err;

	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].private_data = codec;
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].open = snd_intel8x0_playback_open;
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].close = snd_intel8x0_playback_close;

	pcm->chn[SND_PCM_CHANNEL_CAPTURE].private_data = codec;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].open = snd_intel8x0_capture_open;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].close = snd_intel8x0_capture_close;

	pcm->private_data = codec;
	pcm->private_free = snd_intel8x0_pcm_free;
	pcm->info_flags = SND_PCM_INFO_PLAYBACK | SND_PCM_INFO_CAPTURE |
			  SND_PCM_INFO_DUPLEX;
	strcpy(pcm->name, "Intel ICH");
	*rpcm = codec->pcm = pcm;
	return 0;
}

/*
 *  PCM code - MIC
 */

static int snd_intel8x0_capture_mic_ioctl(void *private_data,
					  snd_pcm_subchn_t * subchn,
					  unsigned int cmd,
					  unsigned long *arg)
{
	int result;
	// intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);
	if (result < 0)
		return result;
	return 0;
}

static int snd_intel8x0_capture_mic_trigger(void *private_data,
					    snd_pcm_subchn_t * subchn,
					    int cmd)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	return snd_intel8x0_trigger(codec, &codec->capture_mic, cmd);
}

static int snd_intel8x0_capture_mic_prepare(void *private_data,
					    snd_pcm_subchn_t * subchn)
{
	unsigned long flags;
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;

	codec->capture_mic.physbuf = virt_to_bus(runtime->dma_area->buf);
	codec->capture_mic.size = snd_pcm_lib_transfer_size(subchn);
	codec->capture_mic.fragsize = snd_pcm_lib_transfer_fragment(subchn);
	/* disable ADC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_EXTENDED_STATUS, ~0x4000, 0x4000);
	snd_ac97_write_lock(codec->ac97, AC97_PCM_MIC_ADC_RATE,
			    (codec->ac97_clock * runtime->format.rate) / 48000);
	/* enable ADC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_EXTENDED_STATUS, ~0x4000, 0x0000);
	spin_lock_irqsave(&codec->reg_lock, flags);
	snd_intel8x0_setup_fragments(codec, &codec->capture_mic);
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return 0;
}

static unsigned int snd_intel8x0_capture_mic_pointer(void *private_data,
						     snd_pcm_subchn_t * subchn)
{
	unsigned long flags;
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	unsigned int result;

	spin_lock_irqsave(&codec->reg_lock, flags);
	result = codec->capture_mic.fragsize1 - (inw(ICHREG(codec, MC_PICB)) << 1);
	result += codec->capture_mic.position;
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return result;
}

static snd_pcm_hardware_t snd_intel8x0_capture_mic =
{
	SND_PCM_CHNINFO_MMAP | SND_PCM_CHNINFO_STREAM |
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE |
	SND_PCM_CHNINFO_BLOCK_TRANSFER |
	SND_PCM_CHNINFO_MMAP_VALID |
	SND_PCM_CHNINFO_PAUSE,	/* flags */
	SND_PCM_FMT_S16_LE,	/* hardware formats */
	0,			/* overwritten */
	8000,			/* min. rate */
	48000,			/* max. rate */
	1,			/* min. voices */
	1,			/* max. voices */
	32,			/* min. fragment size */
	128 * 1024,		/* max. fragment size */
	31,			/* fragment align */
	0,			/* FIFO size (unknown) */
	16,			/* transfer block size */
	snd_intel8x0_capture_mic_ioctl,
	snd_intel8x0_capture_mic_prepare,
	snd_intel8x0_capture_mic_trigger,
	snd_intel8x0_capture_mic_pointer
};

static int snd_intel8x0_capture_mic_open(void *private_data,
					 snd_pcm_subchn_t * subchn)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	snd_pcm_hardware_t *hw;
	int err;

	hw = (snd_pcm_hardware_t *)snd_kmalloc(sizeof(snd_pcm_hardware_t), GFP_KERNEL);
	if (hw == NULL)
		return -ENOMEM;
	if ((err = snd_pcm_dma_alloc(subchn, codec->dma_mic,
				     "Intel ICH - MIC ADC")) < 0) {
		snd_kfree(hw);
		return err;
	}
	codec->capture_mic.subchn = subchn;
	memcpy(hw, &snd_intel8x0_capture_mic, sizeof(*hw));
	hw->rates = codec->capture_mic.rates;
	hw->min_rate = codec->capture_mic.min_rate;
	hw->max_rate = codec->capture_mic.max_rate;
	subchn->runtime->hw = hw;
	subchn->runtime->hw_free = (snd_kfree_type)_snd_kfree;
	return 0;
}

static int snd_intel8x0_capture_mic_close(void *private_data,
					  snd_pcm_subchn_t * subchn)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, -ENXIO);

	codec->capture_mic.subchn = NULL;
	snd_pcm_dma_free(subchn);
	/* disable ADC power */
	snd_ac97_write_bitmask_lock(codec->ac97, AC97_EXTENDED_STATUS, ~0x4000, 0x4000);
	return 0;
}

static void snd_intel8x0_pcm_mic_free(void *private_data)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, );
	codec->pcm_mic = NULL;
}

static int snd_intel8x0_pcm_mic(intel8x0_t * codec, int device, snd_pcm_t ** rpcm)
{
	snd_pcm_t *pcm;
	int err;

	*rpcm = NULL;
	err = snd_pcm_new(codec->card, "Intel ICH - MIC ADC", device, 1, 1, &pcm);
	if (err < 0)
		return err;

	pcm->chn[SND_PCM_CHANNEL_CAPTURE].private_data = codec;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].open = snd_intel8x0_capture_mic_open;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].close = snd_intel8x0_capture_mic_close;

	pcm->private_data = codec;
	pcm->private_free = snd_intel8x0_pcm_mic_free;
	pcm->info_flags = SND_PCM_INFO_CAPTURE;
	strcpy(pcm->name, "Intel ICH - MIC ADC");

	*rpcm = codec->pcm_mic = pcm;	
	return 0;
}

/*
 *  Mixer part
 */

static void snd_intel8x0_codec_init(void *private_data, ac97_t *ac97)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, );
	int idx;

	for (idx=10; idx--;) {
		if ((snd_ac97_read(ac97, AC97_POWERDOWN) & 0x0f) == 0x0f)
			goto __ready_ok;
#ifdef CONFIG_PM
		if (codec->in_suspend) {
			mdelay(10);
			continue;
		}
#endif
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ / 10);
	}
	snd_printk("snd_intel8x0_chip_init: Analog subsections not ready.\n");

__ready_ok:
	if (codec->ac97_ext_id & 0x0001)	/* VRA */
		snd_ac97_write_lock(ac97, AC97_EXTENDED_STATUS, 0x0009);	
	/* disable DAC & ADC power */
	snd_ac97_write_bitmask_lock(ac97, AC97_POWERDOWN, ~0x0300, 0x0300);
	/* disable center DAC/surround DAC/LFE DAC/MIC ADC */
	snd_ac97_write_bitmask_lock(ac97, AC97_EXTENDED_STATUS, (short)~0xe800, 0xe800);
}

static void snd_intel8x0_mixer_free_ac97(void *private_data)
{
	intel8x0_t *codec = snd_magic_cast(intel8x0_t, private_data, );
	codec->mixer = NULL;
}

static int snd_intel8x0_mixer(intel8x0_t * codec, int device, int pcm_count, int *pcm_devs, snd_kmixer_t ** rmixer)
{
	snd_kmixer_t *mixer;
	ac97_t ac97;
	int err;

	*rmixer = NULL;
	memset(&ac97, 0, sizeof(ac97));
	ac97.write = snd_intel8x0_codec_write;
	ac97.read = snd_intel8x0_codec_read;
	ac97.init = snd_intel8x0_codec_init;
	ac97.private_data = codec;
	ac97.private_free = snd_intel8x0_mixer_free_ac97;
	if ((err = snd_ac97_mixer(codec->card, device, &ac97, pcm_count, pcm_devs, &mixer)) < 0) {
		return err;
	}
	codec->ac97 = snd_magic_cast(ac97_t, mixer->private_data, -ENXIO);
	*rmixer = codec->mixer = mixer;
	return 0;
}

/*
 *
 */

static int snd_intel8x0_test_rate(intel8x0_t *codec, int reg, int rate)
{
	unsigned short val;
	unsigned int tmp;

	tmp = ((unsigned int)rate * codec->ac97_clock) / 48000;
	snd_intel8x0_codec_write(codec, reg, tmp & 0xffff);
	val = snd_intel8x0_codec_read(codec, reg);
	return val == (tmp & 0xffff);
}

static void snd_intel8x0_determine_rates(intel8x0_t *codec, int reg, ichdev_t *dev)
{
	static int test_rates[] = {
		8000, 11025, 16000, 22050, 32000, 44100, 48000, 0,
	};
	static unsigned int test_rate_bits[] = {
		SND_PCM_RATE_8000, SND_PCM_RATE_11025,
		SND_PCM_RATE_16000, SND_PCM_RATE_22050,
		SND_PCM_RATE_32000, SND_PCM_RATE_44100,
		SND_PCM_RATE_48000, 0,
	};
	int i;

	dev->rates = 0;
	dev->min_rate = dev->max_rate = 0;
	/* test a non-standard frequency */
	if (snd_intel8x0_test_rate(codec, reg, 11000))
		dev->rates |= SND_PCM_RATE_CONTINUOUS;
	/* let's try to obtain standard frequencies */
	for (i = 0; test_rates[i]; i++) {
		if (snd_intel8x0_test_rate(codec, reg, test_rates[i])) {
			dev->rates |= test_rate_bits[i];
			if (! dev->min_rate)
				dev->min_rate = test_rates[i];
			dev->max_rate = test_rates[i];
		}
	}
}


static int snd_intel8x0_chip_init(intel8x0_t *codec)
{
	unsigned short ext_id;
	signed long end_time;
	unsigned int cnt;

	/* put logic to right state */
	/* first clear status bits */
	cnt =  inl(ICHREG(codec, GLOB_STA));
	outl(cnt & (ICH_RCS | ICH_MCINT | ICH_POINT | ICH_PIINT), ICHREG(codec, GLOB_STA));

	/* ACLink on, 2 channels */
	cnt = inl(ICHREG(codec, GLOB_CNT));
	cnt &= ~(ICH_ACLINK | ICH_PCM_246_MASK);
	/* finish cold or do warm reset */
	cnt |= (cnt & ICH_AC97COLD) == 0 ? ICH_AC97COLD : ICH_AC97WARM;
	outl(cnt, ICHREG(codec, GLOB_CNT));
	end_time = jiffies + (HZ >> 2);
	do {
		if ((inl(ICHREG(codec, GLOB_CNT)) & ICH_AC97WARM) == 0)
			goto __ok;
#ifdef CONFIG_PM
		if (codec->in_suspend) {
			mdelay(10);
			continue;
		}
#endif
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(1);
	} while (end_time - (signed long)jiffies >= 0);
	snd_printk("AC'97 warm reset still in progress? [0x%x]\n", inl(ICHREG(codec, GLOB_CNT)));

      __ok:	
	/* try to wait to the ready state */
	inw(codec->port);	/* clear semaphore flag */
	snd_intel8x0_codec_semaphore(codec, 0);
	inw(codec->port);	/* clear semaphore flag */

	codec->ac97_ext_id = 
	ext_id = snd_intel8x0_codec_read(codec, AC97_EXTENDED_ID);
	if (ext_id == 0xffff) {
		snd_printk("snd_intel8x0_chip_init: invalid AC'97 codec\n");
		return -EIO;
	}
	if (ext_id & 0x0001) {
		snd_intel8x0_codec_write(codec, AC97_EXTENDED_STATUS, 0x0009);
		snd_intel8x0_determine_rates(codec, AC97_PCM_FRONT_DAC_RATE, &codec->playback);
		snd_intel8x0_determine_rates(codec, AC97_PCM_LR_ADC_RATE, &codec->capture);
	} else {
		codec->playback.rates = SND_PCM_RATE_48000;
		codec->capture.rates = SND_PCM_RATE_48000;
		codec->playback.min_rate = codec->playback.max_rate = 48000;
		codec->capture.min_rate = codec->capture.max_rate = 48000;
	}
	if (ext_id & 0x0008) {
		snd_intel8x0_determine_rates(codec, AC97_PCM_MIC_ADC_RATE, &codec->capture_mic);
	} else {
		codec->capture_mic.rates = SND_PCM_RATE_48000;
		codec->capture_mic.min_rate = codec->capture_mic.max_rate = 48000;
	}
	/* disable interrupts */
	outb(0x00, ICHREG(codec, PI_CR));
	outb(0x00, ICHREG(codec, PO_CR));
	outb(0x00, ICHREG(codec, MC_CR));
	/* reset channels */
	outb(ICH_RESETREGS, ICHREG(codec, PI_CR));
	outb(ICH_RESETREGS, ICHREG(codec, PO_CR));
	outb(ICH_RESETREGS, ICHREG(codec, MC_CR));
	/* initialize Buffer Descriptor Lists */
	outl(virt_to_bus(codec->playback.bdbar), ICHREG(codec, PO_BDBAR));
	outl(virt_to_bus(codec->capture.bdbar), ICHREG(codec, PI_BDBAR));
	outl(virt_to_bus(codec->capture_mic.bdbar), ICHREG(codec, MC_BDBAR));
	return 0;
}

#ifdef CONFIG_PM
static void suspend_ichdev(intel8x0_t *chip, ichdev_t *ichdev)
{
	unsigned long port;

	if (! ichdev->subchn)
		return;
	port = chip->bmport + ichdev->reg_offset;
	/* stop */
	outb(0, port + ICH_REG_PI_CR);
	ichdev->civ_saved = inb(port + ICH_REG_PI_CIV);
	ichdev->piv_saved = inb(port + ICH_REG_PI_PIV);
	ichdev->picb_saved = inw(port + ICH_REG_PI_PICB);
	/* reset whole DMA things */
	while (!(inb(port + ICH_REG_PI_SR) & ICH_DCH))
		;
	outb(ICH_RESETREGS, port + ICH_REG_PI_CR);
}

static void snd_intel8x0_suspend(intel8x0_t *chip)
{
	chip->in_suspend = 1;
	suspend_ichdev(chip, &chip->playback);
	suspend_ichdev(chip, &chip->capture);
	suspend_ichdev(chip, &chip->capture_mic);
}

static void resume_ichdev(intel8x0_t *chip, ichdev_t *ichdev, int ac97_reg, int power_mask)
{
	snd_pcm_runtime_t *runtime;
	unsigned long port;

	if (! ichdev->subchn)
		return;
	runtime = ichdev->subchn->runtime;
	if (! runtime)
		return;

	port = chip->bmport + ichdev->reg_offset;
	snd_ac97_write_bitmask_lock(chip->ac97, AC97_POWERDOWN, ~power_mask, power_mask);
	snd_ac97_write_lock(chip->ac97, ac97_reg,
			    (chip->ac97_clock * runtime->format.rate) / 48000);
	snd_ac97_write_bitmask_lock(chip->ac97, AC97_POWERDOWN, ~power_mask, 0);
	outl(virt_to_bus(ichdev->bdbar), port + ICH_REG_PI_BDBAR);
	outb(ichdev->lvi, port + ICH_REG_PI_LVI);
	outb(ichdev->civ_saved, port + ICH_REG_PI_CIV);
	outb(ichdev->piv_saved, port + ICH_REG_PI_PIV);
	outw(ichdev->picb_saved, port + ICH_REG_PI_PICB);
	outb(ICH_FIFOE | ICH_BCIS | ICH_LVBCI, port + ICH_REG_PI_SR);
	/* trigger */
	outb(ICH_IOCE | ICH_STARTBM, port + ICH_REG_PI_CR);
}

static void snd_intel8x0_resume(intel8x0_t *chip)
{
	unsigned long flags;

	snd_intel8x0_chip_init(chip);

	snd_ac97_resume(chip->ac97);

	/* resume streams */
	spin_lock_irqsave(&chip->reg_lock, flags);
	resume_ichdev(chip, &chip->playback, AC97_PCM_FRONT_DAC_RATE, 0x0200);
	resume_ichdev(chip, &chip->capture, AC97_PCM_LR_ADC_RATE, 0x0100);
	resume_ichdev(chip, &chip->capture_mic, AC97_PCM_MIC_ADC_RATE, 0x0000);
	spin_unlock_irqrestore(&chip->reg_lock, flags);

	chip->in_suspend = 0;
}

static int 
snd_intel8x0_pm_callback(struct pm_dev *dev, pm_request_t request, void *data)
{
	intel8x0_t *chip = snd_magic_cast(intel8x0_t, dev->data, 0);

	switch (request) {
	case PM_SUSPEND:
		snd_intel8x0_suspend(chip);
		break;
	case PM_RESUME:
		snd_intel8x0_resume(chip);
		break;
        default: 
		break;
	}
	return 0;
}

#endif

static int snd_intel8x0_free(intel8x0_t * codec)
{
	/* disable interrupts */
	outb(0x00, ICHREG(codec, PI_CR));
	outb(0x00, ICHREG(codec, PO_CR));
	outb(0x00, ICHREG(codec, MC_CR));
	/* reset channels */
	outb(ICH_RESETREGS, ICHREG(codec, PI_CR));
	outb(ICH_RESETREGS, ICHREG(codec, PO_CR));
	outb(ICH_RESETREGS, ICHREG(codec, MC_CR));
	/* --- */
	synchronize_irq();
	if (codec->bdbars)
		snd_kfree(codec->bdbars);
#ifdef CONFIG_PM
	if (codec->pm_dev)
		pm_unregister(codec->pm_dev);
#endif
	snd_magic_kfree(codec);
	return 0;
}

static int snd_intel8x0_dev_free(void *private_data)
{
	intel8x0_t *chip = snd_magic_cast(intel8x0_t, private_data, -ENXIO);
	return snd_intel8x0_free(chip);
}

static int __init snd_intel8x0_create(snd_card_t * card,
				      struct pci_dev *pci,
				      int pbk_frame_size,
				      int cap_frame_size,
				      int mic_frame_size,
				      unsigned int ac97_clock,
				      intel8x0_t ** r_intel8x0)
{
	intel8x0_t *codec;
	unsigned short cmdw;
	unsigned char cmdb;
	int err;
	static snd_device_ops_t ops = {
		dev_free: snd_intel8x0_dev_free,
	};

	*r_intel8x0 = NULL;
	codec = snd_magic_kcalloc(intel8x0_t, 0, GFP_KERNEL);
	if (codec == NULL)
		return -ENOMEM;
	spin_lock_init(&codec->reg_lock);
	codec->card = card;
	codec->pci = pci;
	codec->port = pci_resource_start(pci, 0);
	codec->bmport = pci_resource_start(pci, 1);

	if ((err = snd_register_ioport(card, pci_resource_start(pci, 0), 256, "Intel ICH - AC'97", NULL)) < 0)
		goto __error;
	if ((err = snd_register_ioport(card, pci_resource_start(pci, 1), 64, "Intel ICH - Controller", NULL)) < 0)
		goto __error;
	if ((err = snd_register_interrupt(card,
			"Intel ICH", pci->irq,
			SND_IRQ_TYPE_PCI, snd_intel8x0_interrupt,
			codec, NULL, &codec->irqptr)) < 0)
		goto __error;
	if ((err = snd_register_dma_channel(card,
			"Intel ICH - playback", 0,
			SND_DMA_TYPE_PCI, pbk_frame_size,
			NULL, &codec->dma_pbk)) < 0)
		goto __error;
	if ((err = snd_register_dma_channel(card,
			"Intel ICH - capture", 1,
			SND_DMA_TYPE_PCI, cap_frame_size,
			NULL, &codec->dma_cap)) < 0)
		goto __error;
	if ((err = snd_register_dma_channel(card,
			"Intel ICH - MIC capture", 2,
			SND_DMA_TYPE_PCI, mic_frame_size,
			NULL, &codec->dma_mic)) < 0)
		goto __error;

	pci_read_config_word(pci, PCI_COMMAND, &cmdw);
	if ((cmdw & PCI_COMMAND_IO) != PCI_COMMAND_IO) {
		cmdw |= PCI_COMMAND_IO;
		pci_write_config_word(pci, PCI_COMMAND, cmdw);
	}
	pci_set_master(pci);
	pci_read_config_byte(pci, PCI_LATENCY_TIMER, &cmdb);
	if (cmdb < 64)
		cmdb = 64;
	pci_write_config_byte(pci, PCI_LATENCY_TIMER, cmdb);
	synchronize_irq();

	/* initialize offsets */
	codec->playback.reg_offset = 0x10;
	codec->capture.reg_offset = 0;
	codec->capture_mic.reg_offset = 0x20;

	/* allocate buffer descriptor lists */
	/* the start of each lists must be aligned to 8 bytes */
	codec->bdbars = (unsigned int *)snd_kcalloc(8 + sizeof(unsigned int) * 32 * 2, GFP_KERNEL);
	if (codec->bdbars == NULL) {
		err = -ENOMEM;
		goto __error;
	}
	codec->playback.bdbar = (unsigned int *)(((unsigned long)codec->bdbars + 7) & ~7);
	codec->capture.bdbar = codec->playback.bdbar + 32 * 2;
	codec->capture_mic.bdbar = codec->capture.bdbar + 32 * 2;
	if (ac97_clock >= 8000 && ac97_clock <= 48000)
		codec->ac97_clock = ac97_clock;
	else
		codec->ac97_clock = 48000;

	if ((err = snd_intel8x0_chip_init(codec)) < 0)
		goto __error;

#ifdef CONFIG_PM
	codec->pm_dev = pm_register(PM_PCI_DEV, PM_PCI_ID(pci), snd_intel8x0_pm_callback);
	if (codec->pm_dev)
		codec->pm_dev->data = codec;
#endif

	if ((err = snd_device_new(card, SND_DEV_LOWLEVEL, codec, 0, &ops, NULL)) < 0)
		goto __error;

	*r_intel8x0 = codec;
	return 0;

 __error:
	snd_intel8x0_free(codec);
	return err;
}

#define INTEL8X0_TESTBUF_SIZE	32768	/* enough large for one shot */

static void __init intel8x0_measure_ac97_clock(intel8x0_t *chip)
{
	void *buf;
	unsigned long port;
	unsigned long pos, t;
	unsigned long flags;
	struct timeval start_time, stop_time;

	if (chip->ac97_clock != 48000)
		return; /* specified in module option */

	buf = snd_malloc_pages(INTEL8X0_TESTBUF_SIZE, NULL, GFP_KERNEL);
	if (buf == NULL) {
		snd_printk("intel8x0: cannot allocate test buffer.  abort measuring ac97 clock\n");
		return;
	}
	chip->playback.physbuf = virt_to_bus(buf);
	chip->playback.size = chip->playback.fragsize = INTEL8X0_TESTBUF_SIZE;
	chip->playback.subchn = NULL; /* to be sure */

	/* set rate */
	snd_ac97_write_bitmask_lock(chip->ac97, AC97_POWERDOWN, ~0x0200, 0x0200);
	snd_ac97_write_lock(chip->ac97, AC97_PCM_FRONT_DAC_RATE, chip->ac97_clock);
	snd_ac97_write_bitmask_lock(chip->ac97, AC97_POWERDOWN, ~0x0200, 0x0000);

	spin_lock_irqsave(&chip->reg_lock, flags);
	snd_intel8x0_setup_fragments(chip, &chip->playback);
	port = chip->bmport + chip->playback.reg_offset;
	outb(ICH_IOCE | ICH_STARTBM, port + ICH_REG_PI_CR); /* trigger */
	do_gettimeofday(&start_time);
	spin_unlock_irqrestore(&chip->reg_lock, flags);
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(HZ / 20);
	spin_lock_irqsave(&chip->reg_lock, flags);
	/* check the position */
	pos = chip->playback.fragsize1 - (inw(ICHREG(chip, PO_PICB)) << 1);
	pos += chip->playback.position;
	do_gettimeofday(&stop_time);
	outb(0, port + ICH_REG_PI_CR); /* stop */
	/* reset whole DMA things */
	while (!(inb(port + ICH_REG_PI_SR) & ICH_DCH))
		;
	outb(ICH_RESETREGS, port + ICH_REG_PI_CR);
	spin_unlock_irqrestore(&chip->reg_lock, flags);

	t = stop_time.tv_sec - start_time.tv_sec;
	t *= 1000000;
	if (stop_time.tv_usec < start_time.tv_usec)
		t -= start_time.tv_usec - stop_time.tv_usec;
	else
		t += stop_time.tv_usec - start_time.tv_usec;
	if (t == 0) {
		snd_printk("intel8x0: ?? calculation error..\n");
	} else {
		pos = (pos / 4) * 1000;
		pos = (pos / t) * 1000 + ((pos % t) * 1000) / t;
		if (pos && (pos > 48500 || pos < 47500)) {
			chip->ac97_clock = (chip->ac97_clock * 48000) / pos;
			snd_printd("intel8x0: setting clocking to %d\n", chip->ac97_clock);
		}
	}
	snd_free_pages(buf, INTEL8X0_TESTBUF_SIZE);
}

static int __init snd_card_intel8x0_probe(struct pci_dev *pci,
					  const struct pci_device_id *id)
{
	static int dev = 0;
	snd_card_t *card;
	snd_pcm_t *pcm = NULL, *pcm_mic = NULL;
	snd_kmixer_t *mixer = NULL;
	intel8x0_t *chip;
	int pcm_dev = 0, mixer_dev = 0;
	int err;

	for ( ; dev < SND_CARDS; dev++) {
		if (!snd_enable[dev]) {
			dev++;
			return -ENOENT;
		}
		break;
	}
	if (dev >= SND_CARDS)
		return -ENODEV;

	if (pci_enable_device(pci))
		return -ENODEV;

	card = snd_card_new(snd_index[dev], snd_id[dev],
			    snd_intel8x0_use_inc, snd_intel8x0_use_dec);
	if (card == NULL)
		return -ENOMEM;
	card->type = SND_CARD_TYPE_INTEL8X0;

	if ((err = snd_intel8x0_create(card, pci,
				       snd_pbk_frame_size[dev],
				       snd_cap_frame_size[dev],
				       snd_mic_frame_size[dev],
				       snd_ac97_clock[dev],
				       &chip)) < 0)
		goto __nodev;

	if ((err = snd_intel8x0_mixer(chip, mixer_dev++, 1, &pcm_dev, &mixer)) < 0)
		goto __nodev;
	if ((err = snd_intel8x0_pcm(chip, pcm_dev++, &pcm)) < 0)
		goto __nodev;
	if (chip->ac97_ext_id & 0x0008) {	/* MIC VRM */
		if ((err = snd_intel8x0_pcm_mic(chip, pcm_dev++, &pcm_mic)) < 0)
			goto __nodev;
	}
	
	strcpy(card->abbreviation, "ICH");
	strcpy(card->shortname, "Intel ICH");
	
	switch (pci->device) {
	case PCI_DEVICE_ID_INTEL_82801:
		strcpy(card->shortname, "Intel ICH 82801AA");
		break;
	case PCI_DEVICE_ID_INTEL_82801BA:
		strcpy(card->shortname, "Intel ICH 82801BA");
		break;
	case PCI_DEVICE_ID_INTEL_82901:
		strcpy(card->shortname, "Intel ICH 82901AB");
		break;
	case PCI_DEVICE_ID_INTEL_ICH3:
		strcpy(card->shortname, "Intel ICH3");
		break;
	case PCI_DEVICE_ID_INTEL_440MX:
		strcpy(card->shortname, "Intel 440MX");
		break;
	}

	sprintf(card->longname, "%s at 0x%x, irq %i",
		card->shortname, chip->port, pci->irq);

	if (chip->ac97_clock == 48000)
		intel8x0_measure_ac97_clock(chip); /* auto-probe */

	if ((err = snd_card_register(card)) < 0)
		goto __nodev;

	PCI_SET_DRIVER_DATA(pci, card);
	dev++;
	return 0;

      __nodev:
	snd_card_free(card);
	return err;
}

static void __exit snd_card_intel8x0_remove(struct pci_dev *pci)
{
	snd_card_free(PCI_GET_DRIVER_DATA(pci));
	PCI_SET_DRIVER_DATA(pci, NULL);
}

static struct pci_driver driver = {
	name: "Intel ICH audio",
	id_table: snd_intel8x0_ids,
	probe: snd_card_intel8x0_probe,
	remove: snd_card_intel8x0_remove,
};

static int __init alsa_card_intel8x0_init(void)
{
	int err;

	if ((err = pci_module_init(&driver)) < 0) {
#ifdef MODULE
		snd_printk("Intel ICH soundcard not found or device busy\n");
#endif
		return err;
	}
	return 0;
}

static void __exit alsa_card_intel8x0_exit(void)
{
	pci_unregister_driver(&driver);
}

module_init(alsa_card_intel8x0_init)
module_exit(alsa_card_intel8x0_exit)
