#ifndef __HD64572_H
#define __HD64572_H

/* SCA-II HD64572 register definitions - all addresses for mode 0
   and 2 (Intel MPU). For modes 2 and 3, XOR the address with 0x03.

   Source: HD64572 SCA User's Manual
*/

#define	ILAR	0x00		/* Illegal Access */

/* Wait Controller Registers */
#define PABR0L	0x20		/* Physical Address Boundary 0 L */
#define PABR0H	0x21		/* Physical Address Boundary 0 H */
#define PABR1L	0x22		/* Physical Address Boundary 1 L */
#define PABR1H	0x23		/* Physical Address Boundary 1 H */
#define WCRL	0x24		/* Wait Control L */
#define WCRM	0x25		/* Wait Control M */
#define WCRH	0x26		/* Wait Control H */

/* Interrupt Registers */
#define IVR	0x60		/* Interrupt Vector */
#define IMVR	0x64		/* Interrupt Modified Vector */
#define ITCR	0x68		/* Interrupt Control */
#define ISR0	0x6c		/* Interrupt Status 0 */
#define ISR1	0x70		/* Interrupt Status 1 */
#define IER0	0x74		/* Interrupt Enable 0 */
#define IER1	0x78		/* Interrupt Enable 1 */

/* MSCI Channel Registers */
#define MSCI0_OFFSET 0x00
#define MSCI1_OFFSET 0x80

#define MD0	0x138		/* Mode reg 0 */
#define MD1	0x139		/* Mode reg 1 */
#define MD2	0x13A		/* Mode reg 2 */
#define MD3	0x13B		/* Mode reg 3 */
#define CTL	0x130		/* Control reg */
#define RXS	0x13C		/* RX clock source */
#define TXS	0x13D		/* TX clock source */
#define EXS	0x13E		/* External clock input selection */
#define TMCT	0x144		/* Time constant (Tx) */
#define TMCR	0x145		/* Time constant (Rx) */
#define CMD	0x128		/* Command reg */
#define ST0	0x118		/* Status reg 0 */
#define ST1	0x119		/* Status reg 1 */
#define ST2	0x11A		/* Status reg 2 */
#define ST3	0x11B		/* Status reg 3 */
#define ST4	0x11C		/* Status reg 4 */
#define FST	0x11D		/* frame Status reg  */
#define IE0	0x120		/* Interrupt enable reg 0 */
#define IE1	0x121		/* Interrupt enable reg 1 */
#define IE2	0x122		/* Interrupt enable reg 2 */
#define IE4	0x124		/* Interrupt enable reg 4 */
#define FIE	0x125		/* Frame Interrupt enable reg  */
#define SA0	0x140		/* Syn Address reg 0 */
#define SA1	0x141		/* Syn Address reg 1 */
#define IDL	0x142		/* Idle register */
#define TRBL	0x100		/* TX/RX buffer reg L */ 
#define TRBK	0x101		/* TX/RX buffer reg K */ 
#define TRBJ	0x102		/* TX/RX buffer reg J */ 
#define TRBH	0x103		/* TX/RX buffer reg H */ 
#define TRC0	0x148		/* TX Ready control reg 0 */ 
#define TRC1	0x149		/* TX Ready control reg 1 */ 
#define RRC	0x14a		/* RX Ready control reg */ 
#define CST0	0x108		/* Current Status 0 */ 
#define CST1	0x109		/* Current Status 1 */ 
#define CST2	0x10A		/* Current Status 2 */ 
#define CST3	0x10B		/* Current Status 3 */ 
#define GPO	0x131		/* General Purpose Output Pin Ctl Reg */
#define TFS	0x14B		/* Tx Start Threshold Ctl Reg */
#define TFN	0x143		/* Inter-transmit-frame Time Fill Ctl Reg */
#define TBN	0x110		/* Tx Buffer Number Reg */
#define RBN	0x111		/* Rx Buffer Number Reg */
#define TNR0	0x150		/* Tx DMA Request Ctl Reg 0 */
#define TNR1	0x151		/* Tx DMA Request Ctl Reg 1 */
#define TCR	0x152		/* Tx DMA Critical Request Reg */
#define RNR	0x154		/* Rx DMA Request Ctl Reg */
#define RCR	0x156		/* Rx DMA Critical Request Reg */

/* Timer Registers */
#define TIMER0RX_OFFSET 0x00
#define TIMER0TX_OFFSET 0x10
#define TIMER1RX_OFFSET 0x20
#define TIMER1TX_OFFSET 0x30

#define TCNTL	0x200		/* Timer Upcounter L */
#define TCNTH	0x201		/* Timer Upcounter H */
#define TCONRL	0x204		/* Timer Constant L */
#define TCONRH	0x205		/* Timer Constant H */
#define TCSR	0x206		/* Timer Control/Status */
#define TEPR	0x207		/* Timer Expand Prescale */

/* DMA registers */
#define PCR		0x40	/* DMA priority control reg */
#define DRR		0x44	/* DMA reset reg */
#define DMER		0x07	/* DMA Master Enable reg */
#define BTCR		0x08	/* Burst Tx Ctl Reg */
#define BOLR		0x0c	/* Back-off Length Reg */
#define DSR_RX(node) (0x48 + node*2) /* DMA Status Reg (Rx) */
#define DSR_TX(node) (0x49 + node*2) /* DMA Status Reg (Tx) */
#define DIR_RX(node) (0x4c + node*2) /* DMA IRQ Enable Reg (Rx) */
#define DIR_TX(node) (0x4d + node*2) /* DMA IRQ Enable Reg (Tx) */
#define FCT_RX(node) (0x50 + node*2) /* Frame End IRQ Counter (Rx) */
#define FCT_TX(node) (0x51 + node*2) /* Frame End IRQ Counter (Tx) */
#define DMR_RX(node) (0x54 + node*2) /* DMA Mode Reg (Rx) */
#define DMR_TX(node) (0x55 + node*2) /* DMA Mode Reg (Tx) */
#define DCR_RX(node) (0x58 + node*2) /* DMA Command Reg (Rx) */
#define DCR_TX(node) (0x59 + node*2) /* DMA Command Reg (Tx) */

/* DMA Channel Registers */
#define DMAC0RX_OFFSET 0x00
#define DMAC0TX_OFFSET 0x20
#define DMAC1RX_OFFSET 0x40
#define DMAC1TX_OFFSET 0x60

#define DARL	0x80		/* Dest Addr L (single-block, RX only) */
#define DARH	0x81		/* Dest Addr H (single-block, RX only) */
#define DARB	0x82		/* Dest Addr B (single-block, RX only) */
#define DARBH	0x83		/* Dest Addr BH (single-block, RX only) */
#define SARL	0x80		/* Source Addr L (single-block, TX only) */
#define SARH	0x81		/* Source Addr H (single-block, TX only) */
#define SARB	0x82		/* Source Addr B (single-block, TX only) */
#define DARBH	0x83		/* Source Addr BH (single-block, TX only) */
#define BARL	0x80		/* Buffer Addr L (chained-block) */
#define BARH	0x81		/* Buffer Addr H (chained-block) */
#define BARB	0x82		/* Buffer Addr B (chained-block) */
#define BARBH	0x83		/* Buffer Addr BH (chained-block) */
#define CDAL	0x84		/* Current Descriptor Addr L */
#define CDAH	0x85		/* Current Descriptor Addr H */
#define CDAB	0x86		/* Current Descriptor Addr B */
#define CDABH	0x87		/* Current Descriptor Addr BH */
#define EDAL	0x88		/* Error Descriptor Addr L */
#define EDAH	0x89		/* Error Descriptor Addr H */
#define EDAB	0x8a		/* Error Descriptor Addr B */
#define EDABH	0x8b		/* Error Descriptor Addr BH */
#define BFLL	0x90		/* RX Buffer Length L (only RX) */
#define BFLH	0x91		/* RX Buffer Length H (only RX) */
#define BCRL	0x8c		/* Byte Count L */
#define BCRH	0x8d		/* Byte Count H */

/* Block Descriptor Structure */
typedef struct {
	u32 cp;			/* pointer to next block descriptor */
	u32 bp;			/* buffer pointer */
	u16 len;		/* data length */
	u8 stat;		/* status */
	u8 unused;		/* pads to 4-byte boundary */
}pkt_desc;


/* Packet Descriptor Status bits */

#define ST_TX_EOM     0x80	/* End of frame */
#define ST_TX_UNDRRUN 0x08
#define ST_TX_OWNRSHP 0x02
#define ST_TX_EOT     0x01	/* End of transmition */

#define ST_RX_EOM     0x80	/* End of frame */
#define ST_RX_SHORT   0x40	/* Short frame */
#define ST_RX_ABORT   0x20	/* Abort */
#define ST_RX_RESBIT  0x10	/* Residual bit */
#define ST_RX_OVERRUN 0x08	/* Overrun */
#define ST_RX_CRC     0x04	/* CRC */
#define ST_RX_OWNRSHP 0x02
                   
#define ST_ERROR_MASK 0x7C

#define MD0_HDLC        0x80	/* Bit-sync HDLC mode */

#define MD0_CRC_NONE	0x00
#define MD0_CRC_16_0	0x04
#define MD0_CRC_16	0x05
#define MD0_CRC_ITU32	0x06
#define MD0_CRC_ITU	0x07

#define MD2_NRZ		0x00
#define MD2_NRZI	0x20
#define MD2_NRZI_IEEE	0x40
#define MD2_MANCHESTER	0x80
#define MD2_FM_MARK	0xA0
#define MD2_FM_SPACE	0xC0
#define MD2_LOOPBACK    0x03	/* Local data Loopback */

#define CTL_NORTS	0x01
#define CTL_NODTR	0x02
#define CTL_IDLE	0x10
#define CTL_UDRNC	0x20
#define CTL_URSKP	0x40
#define CTL_URCT	0x80

#define ST1_UDRN      	0x80	/* MSCI TX underrun */

#define IE0_TXINT     	0x00000080 /* TX INT MSCI interrupt enable */
#define IE0_UDRN      	0x00008000 /* TX underrun MSCI interrupt enable */

#define CLK_BRG_MASK	0x0F
#define CLK_PIN_OUT	0x80
#define CLK_LINE    	0x00	/* clock line input */
#define CLK_BRG     	0x40	/* internal baud rate generator */
#define CLK_TX_RXCLK	0x60	/* TX clock from RX clock */

#define EXS_TES1	0x20
#define EXS_RES1	0x02

#define CMD_RESET	0x21
#define CMD_TX_ENABLE	0x02
#define CMD_RX_ENABLE	0x12

#define DSR_DWE		0x01
#define DSR_DE		0x02
#define DSR_REF		0x04
#define DSR_UDRF	0x04
#define DSR_COA		0x08
#define DSR_COF		0x10
#define DSR_BOF		0x20
#define DSR_EOM		0x40
#define DSR_EOT		0x80

#define DIR_REFE	0x04
#define DIR_UDRFE	0x04
#define DIR_COAE	0x08
#define DIR_COFE	0x10
#define DIR_BOFE	0x20
#define DIR_EOME	0x40
#define DIR_EOTE	0x80

#define DMR_CNTE	0x02
#define DMR_NF		0x04
#define DMR_SEOME	0x08
#define DMR_TMOD	0x10

#define DMER_DME        0x80	/* DMA Master Enable */

#define DCR_ABORT	0x01
#define DCR_CLEAR_EOF	0x02

#define PCR_COTE	0x80

#endif /* (__HD64572_H) */
