#ifndef __LINUX_USBVISION_H
#define __LINUX_USBVISION_H

#include <linux/list.h>

#define USBVISION_PWR_REG				0x00
	#define USBVISION_SSPND_EN			(1 << 1)
	#define USBVISION_RES2			(1 << 2)
	#define USBVISION_PWR_VID			(1 << 5)
        #define USBVISION_E2_EN			(1 << 7)
#define USBVISION_CONFIG_REG			0x01
#define USBVISION_ADRS_REG				0x02

#define USBVISION_STATUS_REG			0x05
#define USBVISION_IOPIN_REG			0x06
	#define USBVISION_IO_1			(1 << 0)
	#define USBVISION_IO_2			(1 << 1)
#define USBVISION_SER_MODE				0x07
#define USBVISION_SER_ADRS				0x08
#define USBVISION_SER_CONT				0x09
#define USBVISION_SER_DAT1				0x0A
#define USBVISION_SER_DAT2				0x0B
#define USBVISION_SER_DAT3				0x0C
#define USBVISION_SER_DAT4				0x0D
#define USBVISION_EE_DATA				0x0E
#define USBVISION_EE_LSBAD				0x0F
#define USBVISION_EE_CONT				0x10

#define USBVISION_DRM_CONT				0x12
	#define USBVISION_REF			(1 << 0)
	#define USBVISION_RES_UR			(1 << 2)
	#define USBVISION_RES_FDL			(1 << 3)
	#define USBVISION_RES_VDW			(1 << 4)
#define USBVISION_DRM_PRM1				0x13
#define USBVISION_DRM_PRM2				0x14
#define USBVISION_DRM_PRM3				0x15
#define USBVISION_DRM_PRM4				0x16
#define USBVISION_DRM_PRM5				0x17
#define USBVISION_DRM_PRM6				0x18
#define USBVISION_DRM_PRM7				0x19
#define USBVISION_DRM_PRM8				0x1A
#define USBVISION_VIN_REG1				0x1B
	#define USBVISION_16_422_SYNC		0x02
	#define USBVISION_VSNC_POL			(1 << 3)
	#define USBVISION_HSNC_POL			(1 << 4)
	#define USBVISION_FID_POL			(1 << 5)
	#define USBVISION_HVALID_PO		(1 << 6)
	#define USBVISION_VCLK_POL			(1 << 7)
#define USBVISION_VIN_REG2				0x1C
	#define USBVISION_AUTO_FID			(1 << 0)
	#define USBVISION_NONE_INTER		(1 << 1)
	#define USBVISION_NOHVALID			(1 << 2)
	#define USBVISION_UV_ID			(1 << 3)
	#define USBVISION_FIX_2C			(1 << 4)
	#define USBVISION_SEND_FID			(1 << 5)
	#define USBVISION_KEEP_BLANK		(1 << 7)

#define USBVISION_LXSIZE_I				0x1D
#define USBVISION_MXSIZE_I				0x1E
#define USBVISION_LYSIZE_I				0x1F
#define USBVISION_MYSIZE_I				0x20
#define USBVISION_LX_OFFST				0x21
#define USBVISION_MX_OFFST				0x22
#define USBVISION_LY_OFFST				0x23
#define USBVISION_MY_OFFST				0x24
#define USBVISION_FRM_RATE				0x25
#define USBVISION_LXSIZE_O				0x26
#define USBVISION_MXSIZE_O				0x27
#define USBVISION_LYSIZE_O				0x28
#define USBVISION_MYSIZE_O				0x29
#define USBVISION_FILT_CONT			0x2A
#define USBVISION_VO_MODE				0x2B
#define USBVISION_INTRA_CYC			0x2C
#define USBVISION_STRIP_SZ				0x2D
#define USBVISION_FORCE_INTRA			0x2E
#define USBVISION_FORCE_UP				0x2F
#define USBVISION_BUF_THR				0x30
#define USBVISION_DVI_YUV				0x31

#define USBVISION_PCM_THR1				0x38
#define USBVISION_PCM_THR2				0x39
#define USBVISION_DIST_THR_L			0x3A
#define USBVISION_DIST_THR_H			0x3B
#define USBVISION_MAX_DIST_L			0x3C
#define USBVISION_MAX_DIST_H			0x3D

#define USBVISION_OP_CODE				0x33

#define MAX_BYTES_PER_PIXEL     4
#define MIN_FRAME_WIDTH		64
#define MAX_FRAME_WIDTH		320
#define MIN_FRAME_HEIGHT	48
#define MAX_FRAME_HEIGHT	240
#define MAX_FRAME_SIZE          (MAX_FRAME_WIDTH * MAX_FRAME_HEIGHT * \
		                 MAX_BYTES_PER_PIXEL)
#define USBVISION_CLIPMASK_SIZE	(MAX_FRAME_WIDTH * MAX_FRAME_HEIGHT / 8) //bytesize of clipmask

#define USBVISION_URB_FRAMES       32
#define USBVISION_MAX_ISO_PACKET   959	// Doc say 1023 ????

#define USBVISION_NUM_HEADERMARKER	20
#define USBVISION_NUMFRAMES		2
#define USBVISION_NUMSBUF		2

#define FRAMES_PER_DESC		(USBVISION_URB_FRAMES)
#define FRAME_SIZE_PER_DESC	(USBVISION_MAX_ISO_PACKET)

/* This macro restricts an int variable to an inclusive range */
#define RESTRICT_TO_RANGE(v,mi,ma) { if ((v) < (mi)) (v) = (mi); else if ((v) > (ma)) (v) = (ma); }

#if 0
/*
 * This macro performs bounds checking - use it when working with
 * new formats, or else you may get oopses all over the place.
 * If pixel falls out of bounds then it gets shoved back (as close
 * to place of offence as possible) and is painted bright red.
 */
#define USBVISION_PUTPIXEL(fr, ix, iy, vr, vg, vb) { \
	register unsigned char *pf; \
	int limiter = 0, mx, my; \
	mx = ix; \
	my = iy; \
	if (mx < 0) { \
		mx=0; \
		limiter++; \
	} else if (mx >= 352) { \
		mx=351; \
		limiter++; \
	} \
	if (my < 0) { \
		my = 0; \
		limiter++; \
	} else if (my >= V4L_FRAME_HEIGHT) { \
		my = V4L_FRAME_HEIGHT - 1; \
		limiter++; \
	} \
	pf = (fr)->data + V4L_BYTES_PER_PIXEL*((iy)*352 + (ix)); \
	if (limiter) { \
		*pf++ = 0; \
		*pf++ = 0; \
		*pf++ = 0xFF; \
	} else { \
		*pf++ = (vb); \
		*pf++ = (vg); \
		*pf++ = (vr); \
	} \
}
#endif

/*
 * We use macros to do YUV -> RGB conversion because this is
 * very important for speed and totally unimportant for size.
 *
 * YUV -> RGB Conversion
 * ---------------------
 *
 * B = 1.164*(Y-16)		    + 2.018*(V-128)
 * G = 1.164*(Y-16) - 0.813*(U-128) - 0.391*(V-128)
 * R = 1.164*(Y-16) + 1.596*(U-128)
 *
 * If you fancy integer arithmetics (as you should), hear this:
 *
 * 65536*B = 76284*(Y-16)		  + 132252*(V-128)
 * 65536*G = 76284*(Y-16) -  53281*(U-128) -  25625*(V-128)
 * 65536*R = 76284*(Y-16) + 104595*(U-128)
 *
 * Make sure the output values are within [0..255] range.
 */
#define LIMIT_RGB(x) (((x) <= 0) ? 1 : (((x) > 255) ? 255 : (x)))
#define YUV_TO_RGB_BY_THE_BOOK(my,mu,mv,mr,mg,mb) { \
    int mm_y, mm_yc, mm_u, mm_v, mm_r, mm_g, mm_b; \
    mm_y = (my) - 16;  \
    mm_u = (mu) - 128; \
    mm_v = (mv) - 128; \
    mm_yc= mm_y * 76284; \
    mm_b = (mm_yc		+ 132252*mm_v	) >> 16; \
    mm_g = (mm_yc -  53281*mm_u -  25625*mm_v	) >> 16; \
    mm_r = (mm_yc + 104595*mm_u			) >> 16; \
    mb = LIMIT_RGB(mm_b); \
    mg = LIMIT_RGB(mm_g); \
    mr = LIMIT_RGB(mm_r); \
}

/* Debugging aid */
#define USBVISION_SAY_AND_WAIT(what) { \
	wait_queue_head_t wq; \
	init_waitqueue_head(&wq); \
	printk(KERN_INFO "Say: %s\n", what); \
	interruptible_sleep_on_timeout (&wq, HZ*3); \
}

/*
 * This macro checks if usbvision is still operational. The 'usbvision'
 * pointer must be valid, usbvision->dev must be valid, we are not
 * removing the device and the device has not erred on us.
 */
#define USBVISION_IS_OPERATIONAL(udevice) (\
	(udevice != NULL) && \
	((udevice)->dev != NULL) && \
	((udevice)->last_error == 0) && \
	(!(udevice)->remove_pending))

enum {
	STATE_SCANNING,		/* Scanning for header */
	STATE_LINES,		/* Parsing lines */
};

enum {
	FRAME_UNUSED,		/* Unused (no MCAPTURE) */
	FRAME_READY,		/* Ready to start grabbing */
	FRAME_GRABBING,		/* In the process of being grabbed into */
	FRAME_DONE,		/* Finished grabbing, but not been synced yet */
	FRAME_ERROR,		/* Something bad happened while processing */
};

enum {
	ISOC_INFRAME,		/* Isoc packet is member of frame */
	ISOC_NOFRAME,		/* Isoc packet is not member of any frame */
};

struct usb_device;

struct usbvision_sbuf {
	char *data;
	urb_t *urb;
};

#define USBVISION_MAGIC_1      0xAA
#define USBVISION_MAGIC_2      0x55
#define USBVISION_MAGIC		0xAA55
#define USBVISION_HEADER_LENGTH	0x0c
#define USBVISION_SAA7111_ADDR	0x48
#define USBVISION_IIC_LRACK	0x20
#define USBVISION_IIC_LRNACK	0x30

#define USBVISION_FRAME_FORMAT_PARAM_INTRA	(1<<7)

struct usbvision_frame_header {
	unsigned short vid_frm_patt;	/* 0 - 1  magic */
	unsigned char header_length;	/* 2 */
	unsigned char frame_numb;	/* 3 */
	unsigned char frame_phase;	/* 4 */
	unsigned char frame_latency;	/* 5 */
	unsigned char data_format;	/* 6 */
	unsigned char format_param;	/* 7 */
	__u16 frame_width;	/* 8 - 9 */
	__u16 frame_height;	/* 10 - 11 */
};

struct usbvision_regvals {
	unsigned char reg;
	unsigned char value;
};

struct usbvision_frame {
	char *data;		/* Frame buffer */
	struct usbvision_frame_header header;	/* Header from stream */

	int width;		/* Width application is expecting */
	int height;		/* Height */

	int frmwidth;		/* Width the frame actually is */
	int frmheight;		/* Height */

	volatile int grabstate;	/* State of grabbing */
	int scanstate;		/* State of scanning */

	int curline;		/* Line of frame we're working on */

	long scanlength;	/* uncompressed, raw data length of frame */
	long bytes_read;	/* amount of scanlength that has been read 
				   from *data */
	int v4l_format;		// format the user needs
	int v4l_linesize;	// bytes for one videoline

	wait_queue_head_t wq;	/* Processes waiting */
};

#define	USBVISION_MIRO		1	// MIRO PCTV USB
#define USBVISION_HAUPPAUGE	2	// HAUPPAUGE WINTV-USB
#define USBVISION_HAUPPAUGE_FM	3       // HAUPPAUGE WINTV-USB with FM radio
#define USBVISION_HAUPPAUGE_UK	4	// HAUPPAUGE WINTV_USB UK
#define USBVISION_HAUPPAUGE_F	5	// HAUPPAUGE WINTV_USB France
#define USBVISION_HAUPPAUGE_US  6	// HAUPPAUGE WINTV_USB USA
#define USBVISION_BELKIN	7	// BELKIN ???
#define USBVISION_X10		8	// X10 ???


/*
 * Supported Devices: A table for usbvision.c
 */
static struct usbvision_device_data_st {
	int idVendor;
	int idProduct;
	int MaxIsoPacketLen;
	int Interface; 	//to handle special interface number like BELKIN
	int VideoChannels;
	int AudioChannels;
	int Tuner;
	char *ModelString;
} usbvision_device_data[] = {
	{0x050D, 0x0208, 895,  0, 2, 1, 0, "Belkin USB"},
	{0x0573, 0x2000, 959, -1, 2, 1, 0, "X10 USB"},
	{0x0573, 0x4D00, 959, -1, 3, 1, 1, "HAUPPAUGE WINTV-USB USA"},
	{0x0573, 0x4D01, 959, -1, 3, 1, 1, "HAUPPAUGE WINTV-USB"},
	{0x0573, 0x4D02, 959, -1, 3, 1, 1, "HAUPPAUGE WINTV-USB UK"},
	{0x0573, 0x4D03, 959, -1, 3, 1, 1, "HAUPPAUGE WINTV-USB France"},
	{0x0573, 0x4D11, 959, -1, 3, 1, 1, "HAUPPAUGE WINTV-USB FM Germany"},
	{0x2304, 0x0111, 959, -1, 3, 1, 1, "Miro PCTV USB"},
	{ }
};
		
/*
 * Supported Devices: A table for the usb.c
 */
static struct usb_device_id usbvision_table [] = {
       	{ USB_DEVICE(0x050d, 0x0208) }, /* Belkin ??? */
        { USB_DEVICE(0x0573, 0x2000) }, /* X10 ??? */
	{ USB_DEVICE(0x0573, 0x4d00) }, /* Hauppauge WINTV-USB USA */
	{ USB_DEVICE(0x0573, 0x4d01) }, /* Hauppauge WINTV-USB */ 
	{ USB_DEVICE(0x0573, 0x4d02) }, /* Hauppauge WINTV-USB UK */
	{ USB_DEVICE(0x0573, 0x4d03) }, /* Hauppauge WINTV-USB France */
        { USB_DEVICE(0x0573, 0x4d11) }, /* Hauppauge WINTV-USB with FM radio */
	{ USB_DEVICE(0x2304, 0x0111) }, /* Miro PCTV USB */
        { }  /* Terminating entry */
};


#define I2C_CLIENTS_MAX		8

struct usb_usbvision {
	struct video_device vdev;
//      struct video_device rdev;               /* Radio dev */
	struct video_audio audio_dev;	/* Current audio params */
	
	/* i2c */
	struct i2c_adapter i2c_adap;
	struct i2c_algo_usb_data i2c_algo;
	struct i2c_client i2c_client;
	int i2c_state, i2c_ok;
	struct i2c_client *i2c_clients[I2C_CLIENTS_MAX];

	int have_tuner;
	int tuner_type;
	int channel;
	int video_inputs;	/* # of inputs */
	unsigned long freq;

	unsigned int nr;	/* Number of the device < MAX_USBVISION */

	/* Device structure */
	struct usb_device *dev;

	unsigned char iface;	/* Video interface number */
	unsigned char ifaceAltActive, ifaceAltInactive;	/* Alt settings */

	struct semaphore lock;
	int user;		/* user count for exclusive use */

	int usbvision_used;	/* Is this structure in use? */
	int initialized;	/* Had we already sent init sequence? */
	int vdev_model;		/* What type of USBVISION video dev we got? */
	int streaming;		/* Are we streaming Isochronous? */
	int grabbing;		/* Are we grabbing? */
	int last_error;		/* What calamity struck us? */

	int compress;		/* Should the next frame be compressed? */

	char *fbuf;		/* Videodev buffer area for mmap*/
	int fbuf_size;		/* Videodev buffer size */

	int curframe;
	struct usbvision_frame frame[USBVISION_NUMFRAMES];	/* Double buffering */

	int cursbuf;		/* Current receiving sbuf */
	struct usbvision_sbuf sbuf[USBVISION_NUMSBUF];	/* Double buffering */
	volatile int remove_pending;	/* If set then about to exit */

	/*
	 * Scratch space from the Isochronous pipe.
	 */
	unsigned char *scratch;
	int scratch_read_ptr;
	int scratch_write_ptr;
	int scratch_headermarker[USBVISION_NUM_HEADERMARKER];
	int scratch_headermarker_read_ptr;
	int scratch_headermarker_write_ptr;
	int isocstate;
	
	struct video_picture vpic;	/* Picture settings */
	struct video_capability vcap;	/* Video capabilities */
	struct video_channel vchan;	/* May be used for tuner support */
	unsigned char video_endp;	/* 0x82 for USBVISION devices based */
	int has_hdr;
	int frame_num;
	int iso_packet_len;	/* Videomode-dependent, saves bus bandwidth */

	/* OVERLAY stuff */
	struct video_buffer vid_buf;
	struct video_window vid_win;
	int	overlay;		//status: are we overlaying?
	unsigned int 	clipmask[USBVISION_CLIPMASK_SIZE / 4];
	//rwlock_t	lock;
	unsigned char   *overlay_base;	// virt base address of video buffer
	unsigned char   *overlay_win;	// virt start address of overlay window
	struct usbvision_frame overlay_frame;
	
	/* Statistics that can be overlayed on screen */
	unsigned long urb_count;	/* How many URBs we received so far */
	unsigned long urb_length;	/* Length of last URB */
	unsigned long data_count;	/* How many bytes we received */
	unsigned long header_count;	/* How many frame headers we found */
	unsigned long scratch_ovf_count;	/* How many times we overflowed
						   scratch */
	unsigned long iso_skip_count;	/* How many empty ISO packets
					   received */
	unsigned long iso_err_count;	/* How many bad ISO packets received */
};

static struct usbvision_v4l_format_st {
    int           supported;
    int	          bytes_per_pixel;
    int           depth;
    unsigned int  red;
    unsigned int  green;
    unsigned int  blue;
    int           format;
} usbvision_v4l_format[] = {
    { 0 },
    { 0, 1,  8,     0x00FF,     0x00FF,     0x00FF, VIDEO_PALETTE_GREY   },
    { 0, 0,  0,          0,          0,          0, VIDEO_PALETTE_HI240  },
    { 1, 2, 16,     0xF800,     0x07E0,     0x001F, VIDEO_PALETTE_RGB565 },
    { 1, 3, 24,     0xf800,     0x07e0,     0x001f, VIDEO_PALETTE_RGB24  },
    { 1, 4, 32, 0x00ff0000, 0x0000ff00, 0x000000ff, VIDEO_PALETTE_RGB32  },
    { 1, 2, 15,     0x7C00,     0x03E0,     0x001F, VIDEO_PALETTE_RGB555 },
    { 0 /* END OF LIST */ },
};


#endif				/* __LINUX_USBVISION_H */
