/*
 *  linux/fs/supermount/super.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/super.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2, 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewriten for kernel 2.4.6 (C) 2001, 2002 MandrakeSoft Inc
 *                                Juan Quintela (quintela@mandrakesoft.com)
 *
 */

#include <linux/kernel.h>
#include <linux/fs.h>

#include "supermount_i.h"

static char error_buf[1024];

void
supermount_error(struct super_block *sb, const char *function,
		 const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vsprintf(error_buf, fmt, args);
	va_end(args);
	printk(KERN_CRIT "SUPERMOUNT error (device %d/%d): %s: %s\n",
	       MAJOR(sb->s_dev), MINOR(sb->s_dev), function, error_buf);
	/* @@ Do we want to do any special error handling here? */
}

NORET_TYPE void
supermount_panic(struct super_block *sb,
		 const char *function, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vsprintf(error_buf, fmt, args);
	va_end(args);
	panic("SUPERMOUNT panic (device %d/%d): %s: %s\n",
	      MAJOR(sb->s_dev), MINOR(sb->s_dev), function, error_buf);
}

void
supermount_warning(struct super_block *sb, const char *function,
		   const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vsprintf(error_buf, fmt, args);
	va_end(args);
	printk(KERN_WARNING "SUPERMOUNT warning (device %d/%d): %s: %s\n",
	       MAJOR(sb->s_dev), MINOR(sb->s_dev), function, error_buf);
}
