#ifndef GR_ACL_H
#define GR_ACL_H
#endif
#include <linux/grsecurity.h>
#include <linux/grdefs.h>


/* * * * * * * * * * * * * * * * * * * * *
 * Oblivion kernel Modification
 * Main header file
 * Purpose: define most Oblivion data structures 
 * * * * * * * * * * * * * * * * * * * * */

#define GR_VERSION "grsecurity 2.0"

#define SHUTDOWN  0
#define ENABLE    1
#define CHCAPS    2
#define AUTH      3
#define GOD       4

#define GR_SET_INHERIT 0
#define GR_SET_AUTH 1
#define GR_SET_PROC 2
/* required headers: <linux/kdev_t.h>, <linux/types.h> */

/* I need to include linux/fs.h in oblivion.c and it would be
 * overkill to include kdev_t.h twice(once here and once in fs.h)
 * so I just did the other necessary header inclusion before including oblivion.h*/



#define GR_PW_LEN 129
#define GR_SHA_SIZE 20
#define GR_EXTRA_LEN 1025

/* Begin Data Structures */
struct admin_pw {
	char pw[GR_PW_LEN];
	char extra[GR_EXTRA_LEN];
	__u8 sum[GR_SHA_SIZE];  /* 160-bit SHA hash of the password*/
	__u8 mode; /* On/Off/Restart(Restart is not ready yet..*/
};

struct file_acl {
	ino_t inode;
	kdev_t device;
	__u32 mode; 
};

/* When sorting/searching, we can effectively search obv_file_acl and obv_acl structures since their identifying information(inodes/devs) are the same name and their size, due to the above padding, is also the same*/

struct file_db {
	struct file_acl *db;
	int max_ref;
};


/* information about r/w/x info for a file by a process */
/* Uses a (balanced )binary tree for its implementation...ideas are welcome for 
 * a more efficient way however=)*/

struct proc_acl {
	ino_t inode;
	kdev_t device; 
	/* to do rules like /tmp * set the dir /tmp rw .. best solution I can find*/
	__u16 mode;
	/* you may consider the file inode, file_dev, and mode to(combined) be the key */
	struct proc_acl *left;
	struct proc_acl *right;
	__u16 height; /* used for avl rebalancing*/
};



/* information about a process */
struct acl_subject {
	ino_t inode;
	kdev_t device;
	__u8 mode;
	kernel_cap_t file_caps; /* capability set*/
	struct proc_acl *tree;
};

/* database - contains process ACLs */
struct proc_db {
	struct acl_subject *db;
	int max_ref;
};

/* Temporary storage--see kernel/oblivion.c */
struct acl_temp {
	ino_t inode;
	kdev_t device;
	__u16 mode;
};
/* Begin Capabilities subsection */

#define CONFIG_GR_MAX_CAP 30

struct cap_entry {
	char * capname;
	kernel_cap_t capval;
};

struct capset {
	struct cap_entry table[CONFIG_GR_MAX_CAP];
};
/* End Capabilities subsection */
/* End Data Structures */

/* Begin SHA hash section */
#define SHA1HANDSOFF /* Copies data before messing with it. */

struct SHA1_CTX {
	unsigned long state[5];
	unsigned long count[2];
	unsigned char buffer[64];
};

#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
#ifdef __LITTLE_ENDIAN
#define blk0(i) (block->l[i] = (rol(block->l[i],24)&0xFF00FF00) \
		|(rol(block->l[i],8)&0x00FF00FF))
#else
#define blk0(i) block->l[i]
#endif /* __LITTLE_ENDIAN */
#define blk(i) (block->l[i&15] = rol(block->l[(i+13)&15]^block->l[(i+8)&15] \
			^block->l[(i+2)&15]^block->l[i&15],1))

	/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk0(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R1(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R2(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0x6ED9EBA1+rol(v,5);w=rol(w,30);
#define R3(v,w,x,y,z,i) z+=(((w|x)&y)|(w&x))+blk(i)+0x8F1BBCDC+rol(v,5);w=rol(w,30);
#define R4(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0xCA62C1D6+rol(v,5);w=rol(w,30);
/* End SHA hash section */		       
