/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Oblivion: A Security-Enhancing Linux Kernel Modification
 * Author: Michael Dalton
 * E-Mail: michael@linux.com
 * Information: This modification is designed to give 
 * fine-tuned control over processes, to prevent
 * the damage of buffer overflows by disallowing them to run or
 * write to anything that is abnormal.  It also protects files from
 * access by unauthorized programs(read only/append only/hidden modes)
 * and has capability management features.
 *
 * Modifications
 * 4-20-2001 - Began development of OPM .3 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Access Layers */

/* Files -
 * file database -> file info
 */

/* Process
 * process database -> process entry (and capabilities) -> process file access rights 
 */

#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/ctype.h>
#include <asm/uaccess.h>
#include <asm/errno.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/capability.h>
#include <linux/sched.h>
#include <linux/gracl.h>
#include <linux/grhash.h>
#include <linux/sysctl.h>

#define SET_AUTH           1
#define SET_INHERIT        2
#define SET_PROC           3

#define PROC_CONF           "/etc/grsec/proc.acl"
#define FILE_CONF           "/etc/grsec/file.acl"
#define PW_CONF             "/etc/grsec/pw"
#define GRADM_ACL           CONFIG_GRADM_PATH ## " fx"
#define GODMODE_ACL         "/ frwxoi"
/* i.e. this would become "/sbin/gradm x" for executable access only to
 * /sbin/gradm (and find it if its hidden) */

#define empty_tree (struct proc_acl *) NULL

#define heightof(tree) ( (tree == empty_tree) ? 0 : (tree->height) )

#define INIT_CAP_TABLE(table) \
table[0].capname = "CAP_CHOWN";\
table[0].capval = CAP_CHOWN; \
table[1].capname = "CAP_DAC_OVERRIDE"; \
table[1].capval = CAP_DAC_OVERRIDE; \
table[2].capname = "CAP_DAC_READ_SEARCH"; \
table[2].capval = CAP_DAC_READ_SEARCH; \
table[3].capname = "CAP_FOWNER"; \
table[3].capval = CAP_FOWNER; \
table[4].capname = "CAP_FSETID"; \
table[4].capval = CAP_FSETID; \
table[5].capname = "CAP_FS_MASK"; \
table[5].capval = CAP_FS_MASK; \
table[6].capname = "CAP_KILL"; \
table[6].capval = CAP_KILL; \
table[7].capname = "CAP_SETGID"; \
table[7].capval = CAP_SETGID; \
table[8].capname = "CAP_SETUID"; \
table[8].capval = CAP_SETUID; \
table[9].capname = "CAP_SETPCAP"; \
table[9].capval = CAP_SETPCAP; \
table[10].capname = "CAP_LINUX_IMMUTABLE"; \
table[10].capval = CAP_LINUX_IMMUTABLE; \
table[11].capname = "CAP_NET_BIND_SERVICE"; \
table[11].capval = CAP_NET_BIND_SERVICE; \
table[12].capname = "CAP_NET_BROADCAST"; \
table[12].capval = CAP_NET_BROADCAST;\
table[13].capname = "CAP_NET_ADMIN"; \
table[13].capval = CAP_NET_ADMIN; \
table[14].capname = "CAP_NET_RAW"; \
table[14].capval = CAP_NET_RAW; \
table[15].capname = "CAP_IPC_LOCK"; \
table[15].capval = CAP_IPC_LOCK; \
table[16].capname = "CAP_IPC_OWNER"; \
table[16].capval = CAP_IPC_OWNER; \
table[17].capname = "CAP_SYS_MODULE"; \
table[17].capval = CAP_SYS_MODULE; \
table[18].capname = "CAP_SYS_RAWIO"; \
table[18].capval = CAP_SYS_RAWIO; \
table[19].capname = "CAP_SYS_CHROOT"; \
table[19].capval = CAP_SYS_CHROOT; \
table[20].capname = "CAP_SYS_PTRACE"; \
table[20].capval = CAP_SYS_PTRACE; \
table[21].capname = "CAP_SYS_PACCT"; \
table[21].capval = CAP_SYS_PACCT; \
table[22].capname = "CAP_SYS_ADMIN"; \
table[22].capval = CAP_SYS_ADMIN; \
table[23].capname = "CAP_SYS_BOOT"; \
table[23].capval = CAP_SYS_BOOT; \
table[24].capname = "CAP_SYS_NICE"; \
table[24].capval = CAP_SYS_NICE; \
table[25].capname = "CAP_SYS_RESOURCE"; \
table[25].capval = CAP_SYS_RESOURCE; \
table[26].capname = "CAP_SYS_TIME"; \
table[26].capval = CAP_SYS_TIME; \
table[27].capname = "CAP_SYS_TTY_CONFIG"; \
table[27].capval = CAP_SYS_TTY_CONFIG; \
table[28].capname = "CAP_MKNOD"; \
table[28].capval = CAP_MKNOD; \
table[29].capname = "CAP_LEASE"; \
table[29].capval = CAP_LEASE;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Global variables
 * * * * * * * * ** * * * * * * * * * * * * * * * * * * * * */

int _enable_acl_load = 0;
int gr_disable = 0; /* handles boot time disabling of oblivion*/
static struct proc_db procdb;
static struct file_db filedb;
/* Main variables for holding the rulesets*/
static spinlock_t gr_lock = SPIN_LOCK_UNLOCKED;
__u8 gr_status = (GR_1ST | GR_DISABLED);
struct admin_pw pwent;
struct capset * caplist;    
struct acl_subject god = { 0,0,0,~0,empty_tree};
/* iddqd style god mode;) Full caps/access rights*/
extern kdev_t ROOT_DEV;
struct acl_subject auth = { 0,0,(__u8)GR_PROC_AUTH,CAP_INIT_EFF_SET,empty_tree };
/* acl for authentication mode */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Sorting Routines
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



static __inline__ int procacl_less(const struct acl_temp *a, const struct proc_acl *b) 
{
	return ( (a->inode < b->inode) || ( (a->inode == b->inode) && (a->device < b->device) ) ) ? 1 : 0;
}

static __inline__ int procacl_eq(const struct acl_temp *a, const struct proc_acl *b)
{
	return ( (a->inode == b->inode) && (a->device == b->device) ) ? 1:0;
}


static __inline__ int acl_eq(const struct acl_subject *a, const struct acl_subject *b) 
{
	return ( (a->inode == b->inode) && (a->device == b->device)) ? 1 : 0;
}

extern void qsort(void *const pbase,size_t total_elems,size_t size,int (*cmp) (const void *, const void *));

static __inline__ int file_acl_cmp(const void *va, const void *vb) 
{
	struct file_acl *a = (struct file_acl *) va;
	struct file_acl *b = (struct file_acl *) vb;
	if (a->inode < b->inode) 
		return -1;
	else if(a->inode == b->inode)  {
		if(a->device < b->device) 
			return -1;
		else if(a->device == b->device)
			return 0;
	}
	return 1;
}

static __inline__ int fileacl_less(const struct file_acl *f1, const struct acl_temp *f2) {
	return ( ((f1->inode < f2->inode)) || (f1->inode == f2->inode && f1->device < f2->device) ) ? 1 : 0;
}

static __inline__ int fileacl_eq(const struct file_acl *f1, const struct acl_temp *f2) {
	return ( (f1->inode == f2->inode) && (f1->device == f2->device)) ? 1 : 0;
}
static __inline__ int acl_cmp(const void *va, const void *vb) 
{
	struct acl_subject *a = (struct acl_subject *) va;
	struct acl_subject *b = (struct acl_subject *) vb;
	if (a->inode < b->inode)  
		return -1;

	else if(a->inode == b->inode)  {
		if(a->device < b->device) 
			return -1;
		else if(a->device == b->device) 
			return 0;
	}
	return 1;
}
#define acl_less(a,b) (((acl_cmp((const void *)a,(const void *)b)) < 0))

static struct proc_acl * avl_rotl(struct proc_acl *tree) 
{
	struct proc_acl *tmp = tree->right;
	tree->right = tmp->left;
	tmp->left = tree;
	if(heightof(tree->left) >= heightof(tree->right))
		tree->height = heightof(tree->left) +1;
	else
		tree->height = heightof(tree->right) + 1;
	if(heightof(tmp->left) >= heightof(tmp->right))
		tmp->height = heightof(tmp->left) + 1;
	else
		tmp->height = heightof(tmp->right) + 1;
	return tmp;
}

static struct proc_acl * avl_rotr(struct proc_acl *tree)
{
	struct proc_acl *tmp = tree->left;
	tree->left = tmp->right;
	tmp->right = tree;
	if(heightof(tree->left) >= heightof(tree->right))
		tree->height=heightof(tree->left) + 1;
	else
		tree->height = heightof(tree->right) + 1;
	if(heightof(tmp->left) >= heightof(tmp->right))
		tmp->height = heightof(tmp->left) + 1;
	else
		tmp->height = heightof(tmp->right) + 1;  
	return tmp;
}

static struct proc_acl * avl_drlr(struct proc_acl *tree) {
	tree->right = avl_rotr(tree->right);
	return avl_rotl(tree);
}

static struct proc_acl * avl_dlrr(struct proc_acl *tree) {
	tree->left = avl_rotl(tree->left);
	return avl_rotr(tree);
}


static void acl_sort(void) 
{
	qsort((void *)procdb.db,procdb.max_ref,sizeof(struct acl_subject),acl_cmp);
	qsort((void *)filedb.db,filedb.max_ref,sizeof(struct file_acl),file_acl_cmp);
}

#ifdef CONFIG_GR_SUPERDEBUG
static  void list_file_acls(void)
{
	int i;
	for(i=0;i<filedb.max_ref;i++) {
		security_alert("File ACL pos %d has inode %ld dev %d mode %x","",
			   i,filedb.db[i].inode,filedb.db[i].device,
			   filedb.db[i].mode);
	}
}

static void traverse(struct proc_acl *head) 
{
	if(!head) {
		return;
	}
	security_alert("Tree: %ld %d %x","",head->inode,head->device,
			head->mode);
	if(head->left != empty_tree) {
		security_alert("Doing left proc acl","");
		traverse(head->left);
	}
	if(head->right != empty_tree) {
		security_alert("Doing right proc acl","");
		traverse(head->right);
	}
}

static void obv_list_proc_acls(void) 
{
	int i;
	for(i=0;i<procdb.max_ref;i++) {
		security_alert("ACL pos %d has inode %ld dev %d","",i,
				procdb.db[i].inode,procdb.db[i].device);
		security_alert("ACL tree is..","");
		traverse(procdb.db[i].tree);
	}
}
#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Config file management
 * Yikes! If anyone has cleanup ideas I'd love to hear em
 * * * * * * * * * * * * * * * * * * * * * * * * * * */




static int add_pw_conf(void)
{
	struct file *filp;
	int retval;
	mm_segment_t old_fs = get_fs();
	filp = filp_open(PW_CONF, O_RDONLY, O_RDONLY);
	if( ( IS_ERR(filp) ) || (filp == NULL) || (!(filp->f_op->read)) ) {
		security_alert("Could not open config file %.1024s", 
				"config file open errors",
				PW_CONF);
		return 1;
	}
	filp->f_pos = 0;
	set_fs(KERNEL_DS);
	retval = filp->f_op->read(filp,pwent.sum,sizeof(pwent.sum)/sizeof(pwent.sum[0]),&filp->f_pos);
	set_fs(old_fs);
	if(retval != sizeof(pwent.sum)/sizeof(pwent.sum[0])) {
		security_alert("Invalid pw entry detected, got %d wanted %d",
				"password file errors",
				retval,sizeof(pwent.sum));
		retval = 1;
	} else retval = 0;
	filp_close(filp,NULL);
	return retval;
}




static int add_file_acl(char *line) 
{
	char *p, *q;
	struct nameidata filedata;
	int working = 1;
	struct file_acl *curr;

	if( (p = memscan(line,'/',strlen(line))) == line + strlen(line) ) {
		return 0;
	}
	else curr = &(filedb.db[filedb.max_ref]);

	if( (q = memscan(p,' ',strlen(p))) == p + strlen(p))
		return 0;
	*q++ = '\0';
	if(path_init(p,LOOKUP_FOLLOW|LOOKUP_POSITIVE,&filedata)) 
		if(path_walk(p,&filedata)) {
			security_alert("Unable to locate file %.1024s","",p);
			//goto out;
			return 0;
		}
	if( !(filedata.dentry) || IS_ERR(filedata.dentry) ) {
		security_alert("Error adding file acl for file %.1024s",
				"acl add errors",p);
		// goto out;
		return 0;
	}
	curr->inode = filedata.dentry->d_inode->i_ino;
	curr->device = filedata.dentry->d_inode->i_dev;
	curr->mode = 0;

	while(*q == ' ') q++;
	while(working) {	
		switch(*q) {
			case 'r' :
				curr->mode |= GR_FILE_READ;
				q++;
				break;
			case 'h' :
				curr->mode |= GR_FILE_HIDDEN;
				q++;
				break;
			case 'a' :
				curr->mode |= GR_FILE_APPEND;
				q++;
				break;
			case 'w' :
				curr->mode |= GR_FILE_WRITE;
				q++;
				break;
			case 'x':
				curr->mode |= GR_FILE_EXEC;
				q++;
				break;
				/* Note: 's' and 'a' only come into play when something is executed*/
			case 's' :
				filedb.db[filedb.max_ref].mode |= GR_FILE_REQSUM;
				q++;
				break;
			case 'u' :
				filedb.db[filedb.max_ref].mode |= GR_FILE_AUTH;
				q++;
				break;
			default:  /* if anything else, the '\0' at the end of the line will hit here and cause us to terminate, so it won't go on indefinately and start accessing memory it shouldn't*/
				working = 0;
				break;
		}
	}
#ifdef CONFIG_GR_DEBUG
	security_alert("Added File ACL for inode %ld dev %d with mode %x and max_ref %d", "",
			curr->inode,curr->device,curr->mode,filedb.max_ref);
#endif
	filedb.max_ref++;
	path_release(&filedata); 
	return 0;
}


static __inline__ struct proc_acl * avl_set(struct acl_temp *data, struct proc_acl *left, struct proc_acl *right,__u16 height) 
{
	struct proc_acl *dest = kmalloc(sizeof(struct proc_acl),GFP_KERNEL);
	if(dest == NULL) {
		security_alert("Memory allocation failure for %s, Oblivion panicking",
				"memory allocation failures",GR_VERSION);
		return procdb.db[procdb.max_ref].tree;
	}
	dest->inode = data->inode;
	dest->device = data->device;
	dest->mode = data->mode;
	dest->left = left; dest->right = right;
	dest->height = height;
	return dest;
}


static struct proc_acl * avl_tree_ins(struct proc_acl *tree, struct acl_temp *entry) 
{
	if(tree == empty_tree) {
		tree = avl_set(entry,empty_tree,empty_tree,0);
		return tree;
	}
	else if(procacl_less(entry,tree)) {
		tree->left = avl_tree_ins(tree->left,entry);
		if(heightof(tree->left) > heightof(tree->right) + 1) {
			if(procacl_less(entry,tree->left)) 
				tree = avl_rotr(tree);
			else
				tree = avl_dlrr(tree);
		}
	}
	else /* assuming no duplicate entries*/ {
		tree->right = avl_tree_ins(tree->right,entry);
		if(heightof(tree->right) > heightof(tree->left) + 1) {
			if(procacl_less(entry,tree->right)) 
				tree = avl_drlr(tree);
			else
				tree = avl_rotl(tree);
		}
	}
	if(heightof(tree->left) >= heightof(tree->right))
		tree->height = heightof(tree->left) + 1;
	else
		tree->height = heightof(tree->right) + 1;
	return tree;
}

static kernel_cap_t cap_conv(char *p)
{
	kernel_cap_t cap;
	int i;
	char *q;
	cap_t(cap) = 0;
	if(!p) return cap;
	if( (q = memscan(p,'\n',strlen(p))) != p + strlen(p))
		*q = '\0';

	while(*p == ' ') p++;
	if( (strncmp(p,"CAP",3)) != 0) 
		goto out;
	for(i=0;i<CONFIG_GR_MAX_CAP;i++) {
		if(!(strcmp(p,caplist->table[i].capname))) {
			cap_t(cap) = CAP_TO_MASK(caplist->table[i].capval);
			return cap;
		}
	}
out:
	return cap;
}


static int add_proc_acl(char *line,struct acl_subject *obv) 
{
	char *p, *q;
	struct nameidata nd;
	int working = 1;
	struct acl_temp curr;
	if( (p = memscan(line,'+',strlen(line)) ) != line + strlen(line) ) {
		cap_t(obv->file_caps) |= cap_t(cap_conv(++p));
		return 0;
	}
	else if ( (p = memscan(line,'-',strlen(line)) ) != line + strlen(line) ) {
		cap_t(obv->file_caps) &= ~cap_t(cap_conv(++p));
		return 0;
	} 

	else {
		if( (p = memscan(line,'/',strlen(line))) == line + strlen(line)) {
			security_alert("Line %.1024s missing a /","",line);
			return 0;
		}
		if ( (q = memscan(p,' ',strlen(p))) == p + strlen(p)) {
			security_alert("Missing a space in obv proc acl entry, 
					line %.1024s","",line);
			return 0;
		}
		*q++ = '\0';
		if(path_init(p,LOOKUP_FOLLOW|LOOKUP_POSITIVE,&nd))
			if(path_walk(p,&nd)) {
				security_alert("Unable to locate file %.1024s","",p);
				return 0;
			}
		if( (!nd.dentry) || (IS_ERR(nd.dentry)) || (!(nd.dentry->d_inode)) ) {
			if(nd.dentry)
				path_release(&nd);
			security_alert("Unable to locate file %.1024s","",p);
			return 0;
		}
		curr.inode = nd.dentry->d_inode->i_ino;
		curr.device = nd.dentry->d_inode->i_dev;
		curr.mode = 0;

		while(*q == ' ') q++;
		while(working) {
			switch(*q) {
				case 'r' :
					curr.mode |= GR_PROC_READ;
					q++;
					break;
				case 'w' :
					curr.mode |= GR_PROC_WRITE;
					q++;
					break;
				case 'x' :
					curr.mode |= GR_PROC_EXEC;
					q++;
					break;
				case 'a' :
					curr.mode |= GR_PROC_APPEND;
					q++;
					break;
				case 'f':
					curr.mode |= GR_PROC_ACCESS;
					q++;
					break;
				case 'o':
					curr.mode |= GR_PROC_OVERRIDE;
					q++;
					break;
				case 'u' :
					curr.mode |= GR_PROC_AUTH;
					q++;
					break;
				case 'R' :
					curr.mode |= GR_PROC_AUTH_IF_ROOT;
					q++;
					break;
				case 'i' :
					curr.mode |= GR_PROC_INHERIT;
					q++;
					break;
				case 's' :
					curr.mode |= GR_PROC_REQSUM;
					q++;
					break;
				default: 
					working = 0;
					break;
			}
		}
#ifdef CONFIG_GR_DEBUG
		security_alert("About to insert %ld %d %.1024s mode %x", "",
				curr.inode,curr.device,line,curr.mode);
#endif
		obv->tree = avl_tree_ins(obv->tree,&curr);
		path_release(&nd);
		return 0;
	}
	return 1; /* we should never make it here*/
}

static int add_pw(char *pw)
{
	struct SHA1_CTX context;
	SHA1Init(&context);
	SHA1Update(&context,pw,strlen(pw));
	SHA1Final(pwent.sum,&context);
	return 0;
}

static int add_line(char *line, int type) 
{
	static int mode = 0;
	char *p,*q;
	struct dentry *file;
	struct nameidata filed;
	static struct acl_subject *curr;
	int working;
	/* mode = 0 : we are ready for a new program acl
	 * mode = 1 : we are in the middle of reading a program acl*/
	switch(type) {
		case 2 : return add_pw(line);
		case 1 : return add_file_acl(line);
		case 0 : break; /* handled below*/
		default: security_alert("Unknown type %d passed", "",type);
			 return 1;
			 break;
	}

	if(!mode) {
		if(procdb.max_ref >= CONFIG_GR_MAX_RULESET) return 1;
		if( (p = memscan (line,'/',strlen(line))) == (line+strlen(line)) ) {
			security_alert("Could not find a '/' in line %.1024s","",line);
			return 0;
		}
		else curr = &(procdb.db[procdb.max_ref]);


		if( (q = memscan(p,' ',strlen(p)) ) != p + strlen(p) )
		*q++= '\0';

		if(path_init(p,LOOKUP_FOLLOW|LOOKUP_POSITIVE,&filed) ) 
			if(path_walk(p,&filed)) {
				security_alert("Add file error for file %.1024s"
						,"",p);
				return 1;
			}
		file = filed.dentry;
		if (IS_ERR(file) || !file || !(file->d_inode)) {
			if(file)
				path_release(&filed);
			security_alert("Could not lookup file on line %.1024s",
					"",line);
			return 1;
		}
		curr->inode = file->d_inode->i_ino;
		curr->device = file->d_inode->i_dev;
		while(*q == ' ') q++;
		curr->mode = 0;
		working = 1;
		while(working) {
			switch(*q) {
				case 'p':
					curr->mode |= GR_PROC_PROTECTED;
					q++;
					break;
				case 'h':
					curr->mode |= GR_PROC_HIDDEN;
					q++;
					break;
				default:
					working = 0;
			}
		}
		curr->file_caps = cap_bset;
		mode = 1;
#ifdef CONFIG_GR_DEBUG
		security_alert("Added program (full entry) with inode %ld dev %d 
				mode %x line %.1024s", "",
				curr->inode,curr->device,curr->mode,line); 
#endif
		path_release(&filed);
		curr->tree = empty_tree;
		return 0;
	}
	else if( (p=memscan(line,'}',strlen(line))) != (line + strlen(line)) ) {
		procdb.max_ref++;
		mode = 0;
		return 0;
	} else {
		return add_proc_acl(line,curr);
	}
}

static int init_variables(void) 
{
	caplist = kmalloc(sizeof(struct capset),GFP_KERNEL);
	if(caplist) {
		INIT_CAP_TABLE(caplist->table)
	}

	procdb.db = kmalloc(sizeof(struct acl_subject) * CONFIG_GR_MAX_RULESET,GFP_KERNEL);
	filedb.db = kmalloc(sizeof(struct file_acl) * CONFIG_GR_MAX_RULESET,GFP_KERNEL);
	procdb.max_ref = filedb.max_ref = 0;

	if( (add_proc_acl(GRADM_ACL,&auth)) )
		return 1;
	else if( (add_proc_acl(GODMODE_ACL,&god)) )
		return 1;
	else if( (add_proc_acl("/etc r",&auth)) )
		return 1;
	else if( (add_proc_acl("/lib r",&auth)) )
		return 1;

	return procdb.db && filedb.db && caplist ? 0 : 1;
}


static int conf_parse(const char *conffile,const int type)
{
	/* Purpose: open and read from oblivion config file */
	struct file *filp;
	int bytes_read = 0, end = 0, obv_fatal = 0;
	char buffer[1024];
	char *filepos = NULL,*memscanp = NULL;
	mm_segment_t old_fs = get_fs();
	filepos = buffer;
	filp = filp_open(conffile, O_RDONLY, O_RDONLY);
	if( ( IS_ERR(filp) ) || (filp == NULL) ) {
		security_alert("Could not open config file %.1024s","",conffile);

		obv_fatal = 1;
		return obv_fatal;
	} else if ( filp->f_op->read == NULL ) {
		fput(filp);
		obv_fatal = 2;
		return obv_fatal;
	}
	filp->f_pos = 0;
	/* end error checking, now for the real work */
	while ( ! end ) {
		filepos = buffer;
		/* if current task isn't from the kernel, we'll
		   have a current->addr_limit.seg that isn't 0xFFFFFFFF

		   and we can't have that */
		set_fs(KERNEL_DS);
		bytes_read = filp->f_op->read (filp,buffer,(sizeof(buffer)/sizeof(buffer[0])),&filp->f_pos);
		filp->f_pos -= bytes_read;
		set_fs(old_fs);
		if(bytes_read < (sizeof(buffer)/sizeof(buffer[0]))  ) {
			end = 1;
			filepos[bytes_read++] = '\n'; /* make sure it ends on a '\n' */
		}  
					
		while ( bytes_read > 0 && ( memscanp = memscan(filepos,'\n',bytes_read) )   != filepos + bytes_read ) {
			*memscanp++ = '\0' ;
			filp->f_pos += (int) (memscanp - filepos);
			bytes_read -= (int) (memscanp - filepos);
			/* necessary for later strlen()'s...*/
			if( (*filepos != '#') ) 
				if(add_line(filepos,type))
					return 1;
			filepos = memscanp; /* next area to search .. */
		}
	}
	filp_close(filp,NULL);
	return obv_fatal;
}

static int grsecurity_init(void) 
{
	int error = 0;
	spin_lock(&gr_lock);
	gr_status &= ~GR_DISABLED;
	gr_status |= GR_LOADING;
	if(init_variables() )
		error = -1;
	else if(add_pw_conf())
		error = -2;
	else if (conf_parse(PROC_CONF,0))
		error = -3;
	else if(conf_parse(FILE_CONF,1))
		error = -4;
	if(!error) {
#ifdef CONFIG_GR_SUPERDEBUG
		list_proc_acls();
		list_file_acls();
#endif
	        acl_sort();
		gr_status |= GR_READY; /* locked, cocked, and ready to rock*/ 
	}else {
		gr_status |= GR_DISABLED;
		security_alert("Error loading %s, trying to run kernel with oblivion disabled. To disable oblivion at startup use <kernel image name> oblivion=off from your boot loader","errors loading Oblivion",GR_VERSION);
	}
	gr_status &= ~GR_LOADING;
	spin_unlock(&gr_lock);
	return error;
}

/* * * * * * * * * * * * * * * * * * * * * * *
 * Begin Misc Section 
 * * * * * * * * * * * * * * * * * * * * * * */

/* Searching stuff, program exit cleanup, and sysctl parsing */

/* Nonrecursive binary sort, idea from libc, implementation is slightly different but the inspirationw as found there=) */

static __inline__ int do_file_acl_search(struct acl_temp *curr) 
{
	int high, low; 
	struct file_acl *pos;
	high = filedb.max_ref;
	low = 0;
	while(low < high) {
		pos = &filedb.db[((low+high)/2)];
		if(fileacl_less(pos,curr))
			low = (( (low + high) / 2) + 1 );
		else if(fileacl_eq(pos,curr)) {
			if(pos->mode & curr->mode) {
#ifdef CONFIG_GR_DEBUG
				security_alert("File ACL allow, req %x allowed %x\n source: %ld %d dest %ld %d", "",
						curr->mode,pos->mode,pos->inode,pos->device,curr->inode,curr->device);
#endif
				return GR_ALLOW;
			}
			else {
#ifdef CONFIG_GR_DEBUG
				if(!(curr->mode & GR_FILE_HIDDEN))
				security_alert("File ACL deny, req %x allowed %x\n source %ld %d dest %ld %d", "",
						curr->mode,pos->mode,pos->inode,pos->device,curr->inode,curr->device);
#endif
				return GR_DENY;
			}
		}
		else high = ( (low + high) / 2);
	}
	return GR_NOTFOUND;
}

static __inline__ int do_proc_acl_search(struct acl_subject *proc,struct proc_acl *root,struct acl_temp *curr) 
{
	/* Nonrecursively search down the Red-Black tree to see if we have an acl for this file*/
	struct proc_acl *pos = root;
	if(proc == NULL || pos == NULL) /* no entry*/ {
		return GR_DEFAULT;
	}
	// security_alert("Got called...and there's an entry (curr is %ld %d)","",curr->inode,curr->device);
	while(pos != empty_tree) {
		if(procacl_eq(curr,pos))  {
			if(curr->mode & pos->mode) {
				__u16 tmp;
#ifdef CONFIG_GR_DEBUG
				security_alert("Process %ld %d allowed for file %ld %d tried for mode %x allowed mode %x", "",
						proc->inode,proc->device,curr->inode,curr->device,curr->mode,pos->mode);
#endif
				tmp = pos->mode & GR_PROC_EXTRA;
				curr->mode |= tmp; /* Store any "extra" access rights like 
						      inherit or override*/
				return GR_ALLOW;
			} else {
#ifdef CONFIG_GR_DEBUG
				security_alert("Process %ld %d denied for file %ld %d tried for mode %x allowed mode %x", "",
						proc->inode,proc->device,curr->inode,curr->device,curr->mode,pos->mode);
#endif
				return GR_DENY;
			}
		}
		else if(procacl_less(curr,pos)) 
			pos = pos->left;
		else 
			pos=pos->right;
	}

	return GR_NOTFOUND; 
}

static __inline__ int chk_proc_acls(struct dentry * dentry, struct acl_temp *searchval, struct vfsmount *mnt)
{
	struct dentry *curr = dentry;
	int retval = GR_NOTFOUND;
	struct nameidata tmp = {NULL, NULL};
	struct vfsmount *parent;
	struct acl_subject *curracl = (struct acl_subject *) current->acl;

loop :         
	do {
		if(!curr || !curr->d_inode) break;
		searchval->inode = curr->d_inode->i_ino;
		searchval->device = curr->d_inode->i_dev;
		retval = do_proc_acl_search(curracl,curracl->tree,searchval); 
		if(curr->d_inode == curr->d_parent->d_inode) break;
		curr = curr->d_parent;
	} while( (curr) && (retval == GR_NOTFOUND) );
	if(retval == GR_NOTFOUND && curr->d_inode && curr->d_inode->i_dev != ROOT_DEV ) {
		if(tmp.mnt) {
			if(tmp.mnt == tmp.mnt->mnt_parent) goto exit;
		}
		else if(mnt == mnt->mnt_parent) goto exit;


		if(!tmp.mnt) {
			tmp.mnt = mnt->mnt_parent;
			tmp.dentry = dget(mnt->mnt_mountpoint);

		}

		else {
			parent = tmp.mnt->mnt_parent;
			dput(tmp.dentry);
			tmp.dentry = dget(tmp.mnt->mnt_mountpoint);
			mntput(tmp.mnt);
			tmp.mnt = parent;
		}
		mntget(tmp.mnt);
		curr = tmp.dentry;
		goto loop;
	}
exit:		
	if(tmp.mnt) mntput(tmp.mnt);
	if(tmp.dentry) dput(tmp.dentry);


	return retval == GR_NOTFOUND ? GR_DENY : retval;
	/* if its not in a program's ACL list it should be denied */

}

static __inline__ int chk_file_acls(struct dentry * dentry, struct acl_temp *searchval, struct vfsmount *mnt) 
{
	struct dentry *curr = dentry;
	int retval = GR_NOTFOUND;
	struct vfsmount *parent;
	struct nameidata tmp = {NULL,NULL};
loop:	
	do {
		if(!curr || !curr->d_inode) break;
		searchval->inode = curr->d_inode->i_ino;
		searchval->device = curr->d_inode->i_dev;
		retval = do_file_acl_search(searchval); 
		if(curr->d_inode == curr->d_parent->d_inode) break;
		curr = curr->d_parent;
	} while( (curr) && (retval == GR_NOTFOUND) );

	if(retval == GR_NOTFOUND && curr->d_inode && curr->d_inode->i_dev != ROOT_DEV ) {
		if(tmp.mnt) {
			if(tmp.mnt == tmp.mnt->mnt_parent) goto exit;
		}
		else if(mnt == mnt->mnt_parent) goto exit;
		if(!tmp.mnt) {
			tmp.mnt = mnt->mnt_parent;
			tmp.dentry = dget(mnt->mnt_mountpoint);

		}

		else {
			parent = tmp.mnt->mnt_parent;
			dput(tmp.dentry);
			tmp.dentry = dget(tmp.mnt->mnt_mountpoint);
			mntput(tmp.mnt);
			tmp.mnt = parent;
		}
		mntget(tmp.mnt);
		curr = tmp.dentry;
		goto loop;
	}
exit:		
	if(tmp.mnt) mntput(tmp.mnt);
	if(tmp.dentry) dput(tmp.dentry);



 	return retval;
 }


int gr_check_hidden(struct dentry *dentry, struct vfsmount *mnt) {
	/* This has to be done a little differently than a normal search.
	 * If a file "allows" hidden access it is hidden, and thus requires 
	 * a process that has hidden file access capabilities to access it.
	 * If the file acl denies hidden access/has no acl then no proc acl 
	 * check is needed*/
	struct acl_temp searchval;
	struct acl_subject *acl = (struct acl_subject *)current->acl;
	searchval.mode = GR_HIDDEN;
	if(!dentry) return GR_ALLOW;
	if(gr_status & GR_READY) {
		if( (chk_file_acls(dentry,&searchval,mnt)) == GR_ALLOW) {
			if(acl && acl->tree)
				return chk_proc_acls(dentry,&searchval,mnt);
			else return GR_DENY;
		}
	}
	return GR_ALLOW;
}

int gr_check_hidden_proc(struct acl_subject * acl) 
{
	if((gr_status & GR_READY) && acl) {
		if( acl->mode & GR_PROC_HIDDEN) 
			return 1; /* next release adds processes allowed to view hidden ones*/
	}
	return 0;
}

int gr_check_protected(struct task_struct *tsk)
{
	if(gr_status & GR_READY && tsk) {
		struct acl_subject *curracl = (struct acl_subject *) tsk->acl;
		if(curracl)
		        if(curracl->mode & GR_PROC_PROTECTED)
			        return 1;
	}
	return 0;
}


int gr_search(struct dentry *dentry, __u16 mode, struct vfsmount *mnt) 
{
	int proc_retval = GR_NOTFOUND, file_retval = GR_NOTFOUND;
	int retval = GR_DEFAULT;
	if(dentry && dentry->d_inode && mnt) {
		if(gr_status & GR_READY) {
			struct acl_subject *acl = (struct acl_subject *) current->acl;
			struct acl_temp searchval;
			searchval.mode = mode;
			if(acl && acl->tree)  {
				if( ( proc_retval = chk_proc_acls(dentry,&searchval,mnt) ) == GR_DENY) {
					if( (acl->mode & GR_AUTH) && (S_ISCHR(dentry->d_inode->i_mode)))
						retval = GR_NOTFOUND;
					/* Okay. This process is under
					 * authentication mode. Let it
					 * read/write to the tty (and pass it on
					 * to the file acls just in case
					 * Does anyone know how to gain the ino$
					 * from current->tty? It seems only the
					 * device num(major/minor) is stored*/

					else {
						retval = GR_DENY;
						goto out;
					}
				}
			}


			if( (proc_retval == GR_ALLOW) && (searchval.mode & GR_PROC_OVERRIDE)){
				retval = GR_ALLOW;
				goto out;
			}

			if( (file_retval = chk_file_acls(dentry,&searchval,mnt)) == GR_DENY) {
				retval = GR_DENY;
				goto out;
			}

			if( (proc_retval != file_retval) || (proc_retval == GR_ALLOW)) {
				/* The choice to check for proc_retval being GR_ALLOW was
				 * random. If proc_retval and file_retval are equal and
				 * either one is the value GR_ALLOW then both are GR_ALLOW
				 * */
				retval = GR_ALLOW;
				goto out;
			}
		} else retval = GR_ALLOW; /* we aren't loaded, must be early in bootup*/
	}
out: 

return retval; /* Else both returned GR_NOTFOUND..must be the default*/
}


 __inline__ int gr_copy_acl(struct task_struct *tsk) {
	 struct acl_subject * curracl = (struct acl_subject *) current->acl;
	        tsk->acl = (void *) curracl;
		        return 0;
}



static __inline__ struct acl_subject * do_find_proc(struct acl_subject *curr)  
{
	int high,low;
	struct acl_subject *pos;
	low = 0;
	high = procdb.max_ref;
	while(low < high) {
		pos = &procdb.db[((low+high)/2)];
		if(acl_less(pos,curr))
			low = ( ( (low + high) / 2) +1 );
		else if(acl_eq(pos,curr))
			return pos;
		else
			high = ( (high + low) / 2);
	}
	return NULL;
}

static int __inline__ chk_inherit(struct dentry *dentry, struct vfsmount *mnt)
{
	struct acl_temp temp;
	temp.mode = GR_PROC_INHERIT;
	if( (chk_proc_acls(dentry,&temp,mnt)) == GR_ALLOW) 
		return 1;
	return 0;
}

static int __inline__ chk_auth(struct dentry *dentry, struct vfsmount *mnt)
{
	struct acl_temp temp;
	temp.mode = GR_PROC_AUTH;
	if( (chk_proc_acls(dentry,&temp,mnt)) == GR_ALLOW)
		return 1;
	else if(!current->uid || !current->gid || !current->euid || !current->egid || !current->suid || !current->sgid || !current->fsuid || !current->fsgid) 
	{
		temp.mode = GR_PROC_AUTH_IF_ROOT;
		if( (chk_proc_acls(dentry,&temp,mnt)) == GR_ALLOW)
			return 1;
	}
		
	return 0;   
}


static int do_auth(void)
{
#define GR_AUTH_MSG "Authorize yourself to Oblivion please\n"

	char *p = GR_AUTH_MSG;   
	if(!(current->tty) || (!current->tty->driver.write)) return 1;
	current->tty->driver.write(current->tty,0,p,strlen(p));
	current->acl = (void *)&auth;
	return 0;
}

static int do_set_proc_acl(struct dentry *dentry, struct vfsmount *mnt, int mode) 
{
	struct vfsmount *parent;
	struct nameidata tmp = {NULL, NULL};
	struct acl_subject curracl, *retval = NULL;
	struct dentry *curr = dentry;


	switch(mode) {
		case GR_SET_AUTH:
			return chk_auth(dentry,mnt);
			break;
		case GR_SET_INHERIT:
			return chk_inherit(dentry,mnt);
			break;
		case GR_SET_PROC: 
			break;
		default :
			security_alert("BUG! Received unknown mode %d for set_proc_acl","",mode);
			return (unsigned long)NULL;
			break;
	}
	/* Okay now we handle GR_SET_PROC */
loop:	
	do {

		if(!curr || !curr->d_inode) break;
		curracl.inode = curr->d_inode->i_ino;
		curracl.device = curr->d_inode->i_dev;
		retval = do_find_proc(&curracl); 
		if(curr->d_inode == curr->d_parent->d_inode) break;
		curr = curr->d_parent;
	} while( (curr) && (!retval) );
	if(!retval && curr->d_inode && curr->d_inode->i_dev != ROOT_DEV) {
		if(tmp.mnt) {
			if(tmp.mnt == tmp.mnt->mnt_parent) 
				goto exit;
		}
		else if(mnt == mnt->mnt_parent) 
			goto exit;

		if(!tmp.mnt) {
			tmp.mnt = mnt->mnt_parent;
			tmp.dentry = dget(mnt->mnt_mountpoint);

		}

		else {
			parent = tmp.mnt->mnt_parent;
			dput(tmp.dentry);
			tmp.dentry = dget(tmp.mnt->mnt_mountpoint);
			mntput(tmp.mnt);
			tmp.mnt = parent;
		}
		mntget(tmp.mnt);
		curr = tmp.dentry;
		goto loop;
	}
exit:
	if(tmp.mnt) mntput(tmp.mnt);
	if(tmp.dentry) dput(tmp.dentry);

	return (unsigned long)retval;

}

static int do_inherit(void) 
{
	return 0; 
	/* Note: for copying due to fork() we use the code the oblivin code in
	 * kernel/fork.c. If we're inheriting then our current ACL
	 * (that of the program calling execve) is the ACL we inherit, 
	 * so no work needs to be done. Just return success 
	 */
}


int gr_set_proc_acl(struct dentry *dentry,struct task_struct *tsk,char *filename, struct vfsmount *mnt) 
{
	struct acl_subject *temp = NULL;
	if(gr_status & GR_READY) {
		struct acl_subject *curracl = (struct acl_subject *) current->acl;
		if(curracl) {
		if(do_set_proc_acl(dentry,mnt,GR_SET_INHERIT))
			return do_inherit();
		
		else if(do_set_proc_acl(dentry,mnt,GR_SET_AUTH))
			return do_auth();
		}
		temp = (struct acl_subject *)do_set_proc_acl(dentry,mnt,GR_SET_PROC);
	}  
        tsk->acl = (void *)temp;
	if(tsk->acl)
	{
		temp = (struct acl_subject *) tsk->acl;
		tsk->cap_permitted = tsk->cap_effective =  temp->file_caps;
	}
	return 0;
}


static int chcaps(struct admin_pw *entry)
{
	char *p = entry->extra;
	char *q;
	char *mode;
	p[GR_EXTRA_LEN-1] = '\0'; /* take no chances*/
	while( (q = memscan(p,'\n',strlen(p))) != p + strlen(p)) {
		mode = p++;
		*q++ = '\0';
		switch(*mode) {
			case '+' :
				cap_bset |= cap_t(cap_conv(p));
				break;
			case '-' :
				cap_bset &= ~cap_t(cap_conv(p));
				break;
			default:
				break;
		}
		p = q;
	}
#ifdef CONFIG_GR_DEBUG
	security_alert("cap bset is now %x","",cap_bset);
#endif
	return 0;
}

/* The following variables are needed for timer manipulation */
static struct timer_list gr_badpw;
static int failures = 0;
static int during_wait = 0;

static void gr_timer(unsigned long ignored)
{
	failures = 0;
	during_wait = 0;
	del_timer(&gr_badpw);          
}

int gr_proc_handler(ctl_table *table, int write, struct file *filp, void *buffer, size_t *lenp) 
{
	/* This is the main todo. simply I allow disabling oblivion only
	 * at bootup(after its turned on). I don't have time today to 
	 * impelement a good pw scheme. The first thing I do when I get back is to do that.
	 */
	struct admin_pw *usermode = (struct admin_pw *) buffer;
	int error = sizeof(struct admin_pw);
	if(*lenp != sizeof(struct admin_pw)){
		security_alert("Proc handler: being fed garbage %d byte send %d required","more garbage",*lenp,sizeof(struct admin_pw)); 
		return -EINVAL;
	}
	if(during_wait) return -EPERM;
	switch (usermode->mode){
		case SHUTDOWN:
			if(gr_status & GR_READY) {
				if(!(chkpw(usermode))) {
					security_alert("shutdown auth success for "
					DEFAULTSECMSG,"shutdown successes",DEFAULTSECARGS);
					spin_lock(&gr_lock);
					gr_status &= ~GR_READY;
					gr_status |= GR_DISABLED;
					spin_unlock(&gr_lock);
					break;
				} else 
					error = -EPERM;
			}
			break;
		case ENABLE:
			if(gr_disable) {
				security_alert("%s at boot time, ignoring load request","attempts to load oblivion when disabled on boot time",GR_VERSION);
				break;
			}
			else if(gr_status & GR_1ST) {
				security_alert("Loading %.32s","",GR_VERSION);
				grsecurity_init();
				spin_lock(&gr_lock);
				gr_status &= ~GR_1ST; 
				spin_unlock(&gr_lock);
				break;
			}
			else {
				security_alert("%.32s already loaded, re-enabling","",GR_VERSION);
				spin_lock(&gr_lock);
			        gr_status &= ~GR_DISABLED;
				gr_status |= GR_READY;
				spin_unlock(&gr_lock);
				break;
			}

		case CHCAPS : 
			if(gr_status & GR_1ST) {
				grsecurity_init();
			}
			if(gr_status & GR_1ST ||( gr_status & GR_READY && !(chkpw(usermode)))) {
				if(!(chcaps(usermode))) {
					security_alert("successful capability change by "
							DEFAULTSECMSG, "cap changes",
							DEFAULTSECARGS);
				}
				gr_status &= ~GR_1ST;
			}
			else
				error = -EPERM;
			break;
		case GOD:
			if(gr_status & GR_READY) {
				if(!(chkpw(usermode))) {
					security_alert("successful change to admin mode by "
							DEFAULTSECMSG,"",
							DEFAULTSECARGS);
					if(current->p_pptr)
						current->p_pptr->acl = (void *)&god;
				}
				else
					error = -EPERM;
			}
			break;
		case AUTH:
			if(gr_status & GR_READY) {
				if(!(chkpw(usermode)) ) {
					security_alert("successful authentication by "
							DEFAULTSECMSG, "",
							DEFAULTSECARGS);
					if(current->p_pptr)
						current->p_pptr->acl = NULL;
					/* The parent invoked obvadm, it is the
					 * task that we must deal with */
				}
				else
					error = -EPERM;
			}
			break;

		default:
			security_alert("Invalid value %d by " DEFAULTSECMSG, "",
					usermode->mode, DEFAULTSECARGS);
			break;
	}

	if(error < 0) {
		failures++;
		if(failures > CONFIG_GR_MAXTRIES) {
			security_alert("Maximum pw attempts reached (%d), locking "
					"password authentication","password attempt failed",CONFIG_GR_MAXTRIES);
			init_timer(&gr_badpw);
			gr_badpw.data = 0;
			gr_badpw.function = gr_timer;
			gr_badpw.expires = jiffies + CONFIG_GR_TIMEOUT*HZ;
			add_timer(&gr_badpw);
			during_wait = 1;
		}
	}

	return error;
}
